/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: common.h
 *
 *  CONTENTS:   This file contains commonly used definitions and classes
 *              It unites several include files to be used by the rest of
 *              the program (both in the front end and back end).
 *              classes: 'StringConstants', 'TraverseAtoms', 'TraverseVariables'
 *              class: 'EvalStk' : 'Stack' expanded with binary search to be
 *                                 used by grounder
 *              function template: 'quickSort' can sort anything
 *
 *  DATE:     March 29, 2004
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include "struct.h"
#include "stack.h"
#include "symtable.h"

#ifndef COMMON_H
#define COMMON_H

class StringConstants
{
	SymbolTable &symbt;		// Reference to the symbol table:
	int finalFlag;			// The flag is added for the error control.

		// Number of string constants found in the program:
	int constNum;

		// Temporary list of constants.
		//  Used before 'buildReferenceStructure' is called.
	int constList;

	// The minimal number for which there exists: Object(... , minNumber, ...)
	//  where Object is either data predicate or a weight function of any arity >=1
	// minNumber is the least number in the dataSequence that occurrs
	//  either explicitly or results from an arithmetic expression that can be
	//  evaluated during the compilation time.
	int minNumber;

		// After 'buildReferenceStructure' is called, every string constant is
		//  replaced by some unique negative number: "numberForConstant".
		//  The original string can be restored from SymbolTable using:
		//-------------------------------------------------------------------
		// symbolNumberOfConstant=constIndex[ numberForConstant - constBase ];
		//-------------------------------------------------------------------
	int constBase;
	int *constIndex;

	void error(char *nameOfMethod); // Critical error reporting method

public:
		// 1) Adding constants and registering numbers
		//       that occur in declaration sequence
	void addConstant(int symbolno);
	void registerNumber(int number)
		{ if(!finalFlag && number<minNumber) minNumber=number;}

		// 2) Bulding the array to reference the constants:
	int buildReferenceStructure(int *minNumberPtr, int *constNumPtr);

		// 3) Manipulating the constants after the
		//      array to reference the constants is built:
		// numberInPlaceOfConstant -> constantName
	char *getConstantFromNumber(int number);

		// Check whether the given is a valid number
		//  (greater than the least number in dataSequence
	int isNumber(int number) {return(number>=minNumber);}

		// symbolNumberOfConstant -> numberInPlaceOfConstant
	int getNumberForConstant(int symbolno);

		// Print a constant with a given number:
	void printConstant(FILE *out, int number);

		// Gets minimim and maximum numbers that are to be
		//  treated as string constants.
	void getConstantRange(int *min, int *max);

		// Constructor: Initializes everything to zero
	StringConstants(SymbolTable &st);

		// Destructor: Deallocates all memory
	~StringConstants();
};


/*
	This following class can be used for traversal of all
	variables inside a temporary atom list. The traversal
	functions return the poiner to the variable node and flags to
	indicate which atom the variable in and whether the variable
	occurs as a single parameter or as a part of an expression.
*/
	// Maximum nesting of the nodes:
#define TRAV_MAX_LEVELS         7
	// Maximum number of objects that can be traversed
	//  consequently in a single function call
#define TRAV_MAX_PARTS          5

// The following constants specify the type of the list that is
//  currently being traversed. The supported types are the following:
	// Expression list:
#define TRAV_EXPR               1
	// List of temporary atoms:
#define TRAV_ATOM               2
	// List of quantifiers:
#define TRAV_QUANT              3


// The following flags indicate the place where the variable occurrs:

#define TRAV_IN_HEAD            1

#define TRAV_IN_DATAPOS         2
#define TRAV_IN_DATANEG         4
#define TRAV_IN_DATA            (TRAV_IN_DATAPOS|TRAV_IN_DATANEG)
#define	TRAV_IN_PREDEFINED      8
#define TRAV_IN_CONDITION       (TRAV_IN_DATA|TRAV_IN_PREDEFINED)

#define TRAV_IN_CARD            16
#define TRAV_IN_WEIGHT          32
#define TRAV_IN_META            (TRAV_IN_CARD|TRAV_IN_WEIGHT)

#define TRAV_IN_PREDICATE       64
#define TRAV_IN_LOCALS          128
#define TRAV_IN_WEIGHTASS       256
#define TRAV_IN_RESTRICTIONS    512
#define TRAV_IN_QUANT_VAR       1024
#define TRAV_IN_QUANT_COND      2048
#define TRAV_IN_QUANT           (TRAV_IN_QUANT_VAR|TRAV_IN_QUANT_COND)
#define TRAV_IN_PROG            (TRAV_IN_PREDICATE|TRAV_IN_LOCALS|\
                                 TRAV_IN_WEIGHT|TRAV_IN_RESTRICTIONS|\
                                 TRAV_IN_QUANT)

// Class itself:
class TraverseVariables
{
	// The flag is used in the method 'next'.
	//  If it is '1', then 'start' was called right before and
	//  we need to use the current value of the part.
	// Otherwise, use the next value:
	int startFlag;

	// Structure describing the part (e.g. local variable list,
	//  argumen list, etc.) of some object (e.g. program atom).
	//  Also, several parts can be input directly by the start.
	struct Parts
	{
		// type of the part: ATOM, EXPRESSION, or QUANTIFIER:
		int token;
		// Variable that accumulates atom type to be returned
		//  by the 'next'.
		int atomType;

		// Pointer to the current value of the part:
		//  For a part in the last level, this is the pointer to
		//  the object to which the returned expression was attached.
		union
		{
			TempAtom *atom;
			Expression *exp;
			Quantifier *quant;
		};
	};

	// Levels consisting of several parts. Each level corresponds
	//  to a level of nesting of the structure being traversed:
	struct Levels
	{
		Parts *part;	// Current part.
		Parts parts[TRAV_MAX_PARTS];
	} levels[TRAV_MAX_LEVELS];

	Levels *level;		// Current level.
public:
		// Initialize the traverse:
	void start(char *controlString, ...);
	    // Control string: each character corresponds to each parameter
	    //  There should be no more than 5 characters. Allowed characters:
	    //   A  -  List of atoms from the body
	    //   H  -  List of atoms from the head
	    //   E  -  List of expressions
	    //   Q  -  Lits of quantifiers
	    //  NOTE: * symbol should never be used when this
	    //    method is called from outside the class
	    //  If * is used, then after each pointer
	    //     there should be atom type flag

	// Get the next variable:
	VarNode *next(int *atomType=0, int *paramType=0);
	TraverseVariables();	// Constructor
};


/*
This class is to be used for traversal of all
atoms in the body and head of a rule.
*/
class TraverseAtoms
{
	TempAtom dummyAtom;

	TempAtom *parts[2];     // To save pointers to the body and/or head
	int curPart;            // Current element in 'parts'

		// Current atom on the level of simple and meta atoms
	TempAtom *cur1;
		// Current atom on the level of meta-lists
	TempAtom *cur2;

		// Indicates whether the last returned atoms is in the head,
		// and/or in cardinality or weight list.
	int curSurroundType;

		//  When this flag is set, the TraverseAtoms::next will
	    //   return the meta atoms before proceeding to meta-lists
	int traverseMetaAtoms;

public:
		// Start the traversal:
	void start(TempAtom *body, TempAtom *head, int traverseMetaAtoms=0);

		// Get the next atom:
	TempAtom *next(int *surroundType=0);

	TraverseAtoms();
};


	// Evaluation stack for grounder.
template <typename type>
class EvalStack: public Stack<type>
{
public:
/*
int vectorSearch(int vSize, int *array, int vecNum)
DESCRIPTION:
	BINARY SEARCH. Searches an array of vectors for a
	particular vector(of course, the array of vectors should be sorted)
	The target vector should be located in the stack (top == last element
	of the vector). It is popped from the stack once the search is complete

PARAMETERS:
	vSize	- size of a vector
	array	- Array of of vectors we need to search. Contains 'vecNum'
	            vectors, thus its size is (vecNum*vSize).
	vecNum	- Number of vectors in the array

	RETURNS:
	    NONNEGATIVE NUMBER - Index of the first element of the vector in the
	                         array (NOT AN INDEX OF A VECTOR). It can be in
	                         the range [0, vecNum*vSize].
	    (-1)               - The vector was not found.
*/
	int vectorSearch(int vSize, type *array, int vecNum)
	{
		int i;
		int half, halfWay;
		type *start=array;
		type *element=this->top-vSize;

			// Check whether the stack contains enough elements.
		if(this->getNumOfElements()<vSize)
		{
			fprintf(stderr, "CRITICAL ERROR in "
				"EvalStack::vectorSearch\n"
				"the stack does not contain "
				"enough elements to do search.\n\n");
			exit(1);
		}

			// Check for errors in parameters
		if(!array||vSize<=0||vecNum<=0) return(-1);

		while(vecNum>1)
		{
			half=vecNum/2;      // # of the middle vector
			halfWay=half*vSize; // Index of the first element
			                    //    of the middle vector

				// Compare the vectors:
			for(i=0;i<vSize-1&&array[halfWay+i]==element[i];i++);

				// Check whether {MiddleVector} > {element}
			if(array[halfWay+i]>element[i])
				vecNum=half; // Go to first half
			else
			{
					// Go to second half:
				array+=halfWay;
				vecNum-=half;
			}
		}

			// Check whether we found what we were looking for
		for(i=0;i<vSize&&array[i]==element[i];i++);

			// Popping the vector from the stack
		this->top-=vSize;

			// Return the result:
		if(i>=vSize) return(array-start);
		else return (-1);
	}

/*
int scalarSearch(type element, type *array, int size)
DESCRIPTION:
	BINARY SEARCH. Searches an element in an array of
	integers with the given size(of course, array should be sorted).

PARAMETERS:
	element   - Element that we need to find
	array     - Array of elements we need to search
	size      - Number of elements in the array.

RETURNS:
	NONNEGATIVE NUMBER  - Index of element in the array
	(-1)                - The element was not found
*/
	static int scalarSearch(type element, type *array, int size)
	{
		int half;
		type *start=array;

			// Check for wrong input
		if(!array||size<=0) return (-1);

		while(size>1)
		{
			half=size/2;

				// Check the middle element and go to the
				// necessary half of the array.
			if(array[half]>element) size=half; // Go to first half
			else
			{
					// Go to second half.
				array+=half;
				size-=half;
			}
		}

			// Return the result
		if(*array==element) return(array-start);
		else return (-1);
	}

		// Constructor:
	EvalStack<type>(int sizeOfStack):Stack<type>(sizeOfStack){}
};



// *************************************************************************
//  QUICK SORT TEMPLATE:

// If size of an array to be sorted by 'quickSort' is less than this, then
//  we will use middle element of the array instead of random.
#define SORT_RND_MIN	30

/*
template <class Array>
void quickSort(Array &array)
DESCRIPTION:
	This is a template for sorting anything that can be represented as 
	an abstract array. This array has to support the following methods:
	
	class Array
	{
	int getNumOfElements();
		Returns the number of elements in the array

	int isLess(int i, int j);
		Given the indices of the two elements in the array, the method
		returs 1 if ARRAY[i] < ARRAY[j] or 0 otherwise.

	void swap(int i, int j);
		Swaps i-th and j-th elements of the array.
	}

PARAMETERS:
	array	-	Reference to the abstract array.
*/
template <class Array> 
void quickSort(Array &array)
{
	static StackL<int> stk(20, 2, 100);
	int start, begin, middle, end, flg;
	int num=array.getNumOfElements();
	if(num<=1) return;

		// Initializing stack.
		// At the bottom of the stack should be the last element
	stk.clear();
	stk.push(num-1);
	start=0;

		// Variables:
			// start - starting element for the splitting.
			// end - middle element obtained by a split;
			//		last element for the next split.
	do
	{
		end=stk.getElement(0);
			// Processing the left parts
			//  (actually, may process one right part at the first)
		while(1)
		{
			flg=0;
			begin=start;
				// If the given array has no more than 1 element,
				//  we need to stop the while loop.
			if(begin >= end) break;

				// Choosing the middle element:
			if(end-begin>SORT_RND_MIN)
				middle=begin + (int)((float)rand()*
					(float)(end-begin)/(float)RAND_MAX);
			else
				middle=(begin+end)/2;


			// If the array has two elements, we may need to put
			//  them into the proper places but don't
			//  need to do anything else.
			if(begin+1>=end) flg=1;

			do
			{
					// Finding the first element >=Middle
					//  (from the beginnig)
				while(array.isLess(begin, middle)) begin++;
					// Finding the first element <=Middle
					//  (from the end)
				while(array.isLess(middle, end)) end--;
				if(begin>=end) break;	// Exit in this case.

					// Checking whether the middle element
					//   has to change its position:
				if(middle==begin) middle=end;
				else if(middle==end) middle=begin;

					// Swaping the elements,
					//  and adjusting the indices.
				array.swap(begin, end);
				begin++;
				end--;
			}while(1);

				// If there were only two elements,
				//   we don't need to split more.
			if(flg || end < start) break;

				// At this point, 'end' should contain
				//  index of the middle point.
			stk.push(end);
		};

		start=stk.pop()+1;  // Setting the right part to be processed.
	}while(stk.getNumOfElements());// Do until the stack is empty.

	stk.clear();	// Clear the stack before exiting.
}
#endif
