/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: gr_rules.cpp
 *
 *  CONTENTS:  Part of the BACK END, PSGRND.
 *             Implementation of the class 'GroundRules'. The most important
 *              methods of the grounder that do propagation, one atom
 *              lookahead and output the ground theory to the output file.
 *
 *  DATE:     May 8, 2004
 *
 *  MODIFIED BY: Artur Mikitiuk, Jul 22, 2004 & Jan 15, 2005
 *
 *  MODIFIED BY: Louie G. Downey III, Sep 21, 2004
 *  (function writeTheoryTo2Files)
 ***************************************************************************/

#include "back_end.h"

// Static variable of the class:
AVLNode GroundRules::delimiters[DELIM_NUM];
// The message that is written at the end of the human-readable
//  output and explains the format of the output
char GroundRules::humanOutputFormatExplanation[]=
"\n\nNOTATION IN THE HUMAN-READABLE FORMAT:\n\n"
"1) Symbols in rules:\n"
"   \',\' means that the rule is an implication; the atoms to the left\n"
"       form the body and the atoms to the right form the head\n"
"   \':\' means that this is a horn rule; the atoms to the left\n"
"       form the body and the atom to the right is a head\n"
"1) Symbols before the rules or atoms:\n"
"   +: the atom or meta atom is true\n"
"   -: the atom or meta atom is false\n"
"   *: the horn atom is pseudo-true (forced to be true by implication)\n"
"  +*: pseudo-true during propagation, true during lookahead\n"
"   @: this is the point where the theory was found to be inconsistent\n"
"   \': added to +,-,* if the value is temporary (assigned during lookahead)\n"
"   ?: CRITICAL ERROR\n"
"2) RuleNumbers: #RN/LN\n"
"   RN: number of the ground rule in the theory\n"
"   LN: line number where original rule appears in the input file\n"
"3) Notation in the script: <RN> AT1 AT2 ...\n"
"   RN: ground rule number\n"
"   AT1, AT2 : atoms that were assigned some "
"value during propagation on the rule RN\n";

  // Values that define the place where the theory was found to be inconsistent:
    // During grounding:
#define GR_INC_GROUND           0
    // During propagation on horn rules:
#define GR_INC_HORN             1
    // During lookahead:
#define GR_INC_LA               2

/*******************************************************************************
@1 *********           MAIN METHOD                    *************************/

/*
int GroundRules::executeFinalStage(
    int executionFlag, PPnode *programPredicateList, FILE *mainOutput,
    FILE *mainOutput2, FILE *humanOutput, RuleNode *ruleSequence)
DESCRIPTION
    One of the central methods of the class.
    This method should be called after all theory was grouned. It does
    all the necessary steps to finish the grounding and output the
    ground theory (the actual steps depend on what is being enabled):
    propagation on horn atoms, compaction lookahead, assignment of numbers
    to the ground atoms, writing human-readable or solver-readable outputs

PARAMETERS
    executionFlag        - Execution and output control. Can be any
                             combination of the following flags (GROUT_...):
        GROUT_ERRORS  -  This flag should be always set if and only if there
                           were grounding errors. This is the only way to
                           notify 'GroundRules' about them
        GROUT_LEGACY  -  If this flag is set, the output will be generated
                           in the "legacy format", i.e. atoms that don't
                           affect the ground theory will be assigned numbers
                           as regular atoms instead of '-1'.
        GROUT_DIMACS  -  Generate output in the DIMACS format.
        GROUT_PBS     -  Generate output in the PBS format.

    programPredicateList - List of all program predicate in the program
                             should be taken from the class 'Rules'

    mainOutput           - Main output file. This should be opened file
                             with writing permissions; can be NULL, in
                             which case no main output will be generated

    mainOutput2          - Second main output file, needed for output in
                 the PBS format; is NULL for other output
                 formats

    humanOutput          - Output file for the human-readable output
                             works similarly to 'mainOutput'.

    ruleSequence         - Intermediate representation of the rule
                             sequence (NOT GROUND RULES). Should be
                             obtained from the class 'Rules'.

RETURNS
    TRUE    -  Produced main output successfully.
    FALSE   -  Otherwise
*/
int GroundRules::executeFinalStage(
        int executionFlag, PPnode *programPredicateList, FILE *mainOutput,
        FILE *mainOutput2, FILE *humanOutput, RuleNode *ruleSequence)
{
    int incReason=GR_INC_GROUND;
    unsigned int atomCount=0;
    unsigned int numOfAtoms=0;
    GroundTheoryStatistics stats;
    int returnValue=0;
    char *str;

        // Report the program predicates that were used but no ground atoms
        //  were generated from them:
    if(THEORY_IS_CONSISTENT) reportNotGroundedPredicates(programPredicateList);

        // Calculate the numbers of rules where horn atom occur
        //  and propagate on horn atoms  if necessary.
    if(THEORY_IS_CONSISTENT)
        if(!propagateHornAtoms(!(controlFlag&GR_DO_PROPAGATE)))
        {
                // If the theory became inconsistent,
                //  save the reason and set the flag:
            incReason=GR_INC_HORN;
            INDICATE_INCONSISTENCY;
        }

        // If the scripting is enabled, write the following line to the script:
    if(controlFlag&GR_DO_SCRIPT)
        fprintf(scriptOut, "\nPropagation is done. "
            "Atoms determined during lookahead:\n");


        // If the propagation is enabled, propagate on horn atoms
    if((controlFlag&GR_DO_PROPAGATE) && THEORY_IS_CONSISTENT)
    {
            // Do the final compaction if necessary
        if(controlFlag&GR_DO_ANY_COMPACTION)
            compactRules(GR_CM_ALLFINAL, &stats);

            // If the lookahead is enabled, perform it:
        if(controlFlag & GR_DO_LOOKAHEAD)
        {
                // Perform lookahead.
            if(!lookahead())
            {
                INDICATE_INCONSISTENCY;
                incReason=GR_INC_LA;
            }
            else  // Do compaction after the lookahead:
                if(controlFlag & GR_DO_ANY_COMPACTION)
                    compactRules(GR_CM_ALLFINAL, &stats);
        }
    }

        // If the theory is consistent, finish propagation and assign the
        //  numbers to the ground atoms:
    if(THEORY_IS_CONSISTENT)
    {
        atomCount = completePropagationAndAssignNumbers(
		executionFlag&GROUT_DIMACS || executionFlag&GROUT_PBS,
		executionFlag&GROUT_LEGACY, &stats);

            // If the legacy flag is set, we need to calculate the number of
            //  ground atoms manually: done by calling the function that writes
           //  the ground atoms to a file, without giving it a file
        if(atomCount==GR_ANUM_ULIMIT || ((executionFlag&GROUT_LEGACY) &&
            (numOfAtoms=writeAtomsToFile(0, atomCount, programPredicateList))
                    == GR_ANUM_ULIMIT))
        {  // If there are not enough integers to represent all
           //  the ground atoms, report an error and quit immediately:
            fprintf(errorStream, "ERROR: The theory contains too many different"
                " ground atoms, and there are not enough integer numbers to "
                "output such theory. Grounding is stopped, "
                "no output will be generated\n");

            return 0;
        }
    }

        // If the theory is consistent, check whether the grounder has found
        //  the solution. Otherwise, print the message that the theory is
        //  inconsistent:
    if(THEORY_IS_CONSISTENT)
    {
            // If compaction is disabled, we need to calculate the number of
            //  of rules, cardinality or weight atoms manually.
        if(!(controlFlag & GR_DO_ANY_COMPACTION))
            calculateTheoryStatistics(&stats);

            // Check whether grounder found the only solution:
        if(stats.rules==0)
            fprintf(stderr, "Grounder Found Solution To The Problem\n");
    }
    else
        fprintf(stderr, "The Theory Is Inconsistent\n");

    if(controlFlag&GR_DO_SCRIPT) fprintf(scriptOut, "\n\n");

        // Write the human-readable output if necessary:
    if(humanOutput)
    {
            // This is done because usually human-redable output and the
            //  script output are the same file. Therefore, need to help
            //  the user to separate the information:
        if(!(controlFlag&GR_DO_SCRIPT))
            fprintf(humanOutput, "Human-readable output. For explanation "
            "of the format, see the end of the file.\n");

        fprintf(humanOutput, "Printing rules and atoms:\n");

            // If the theory was found to be inconsistent, print
            //  how it was found to be this way and print the
            //  atom that caused this.
        if(!THEORY_IS_CONSISTENT)
        {
            switch(incReason)
            {
            case GR_INC_GROUND: str="propagation"; break;
            case GR_INC_HORN:   str="propagation on horn atoms"; break;
            case GR_INC_LA:     str="lookahead"; break;
            default:str="?";
            }

            fprintf(humanOutput, "The theory was found to "
                "be inconsistent during %s\n", str);

                // Print the causing atom:
            if(conflictingAtom)
            {
                fprintf(humanOutput, "The atom that causes trouble is: ");
                printAtom(humanOutput, conflictingAtom->symbolNum,
                                       conflictingAtom->atomNum);
                fprintf(humanOutput, "\n");
            }
        }

            // Write the ground theory, atom list, and the explanation
            //  of the format to the human-readable output:
        writeTheoryToFile(humanOutput,0, ruleSequence);
        writeAtomsToHRFile(humanOutput);
        fprintf(humanOutput, humanOutputFormatExplanation);
    }


        // WRITE THE OUTPUT FOR THE SOLVER:
    if(mainOutput)
    {
            // If the theory is inconsistent, all the ground rules
            //  are considered invalid. Therefore, we need to remove
            //  all the rule and add the rule: " -> ."  indicating
            //  inconsistency:
        if(!THEORY_IS_CONSISTENT)
        {
            clearTheory(&stats);
            addSimpleRule(0,0);
            stats.rules++;
        }

        // The following are the checks for various cases when the
        //  main output cannot be produced. In each of these cases,
        //  an error message is printed for the user:

        // a) If propagation is enabled, but compaction is disabled:
        if((controlFlag&GR_DO_PROPAGATE) && !(controlFlag&GR_DO_ANY_COMPACTION))
            fprintf(stderr, "ERROR: Cannot produce output for solver when"
                "\npropagation is enabled but compaction is disabled\n");
        else
        // b) If there were errors during grounding:
    if(controlFlag&GR_FLG_ERROR || (executionFlag&GROUT_ERRORS))
        fprintf(stderr, "ERROR: Cannot produce output for solver "
        "because of errors during grounding\n");
    else
        // c) If the DIMACS output format is requested but the theory
        //    contains cardinality or weight atoms:
    if((executionFlag&GROUT_DIMACS) &&
        stats.weightAtoms + stats.cardAtoms > 0)
        fprintf(stderr, "ERROR: Cannot produce DIMACS output "
            "for the theory with meta-atoms\n");
        // d) If the PBS output format is requested but not all
        //    pseudoboolean atoms form unit clauses:
    else if ((executionFlag&GROUT_PBS) && impossibleCPLEX())
        fprintf(stderr, "ERROR: To produce PBS output, all "
            "pseudoboolean atoms must form unit clauses\n");
    else
    {
        // If not legacy output, calculate the number of ground atoms
        // (otherwise, it should have already been calculated)
        if(!(executionFlag&GROUT_LEGACY))
        numOfAtoms=stats.atomsUnknown;

        // Write the first line in the file depending whether
        //  this is DIMACS output or not:
        if((executionFlag&GROUT_DIMACS)||(executionFlag&GROUT_PBS))
        fprintf(mainOutput, "p cnf %u %u\n",
	    numOfAtoms + stats.atomsTrue + stats.atomsNodiff,
            stats.rules - stats.weightAtoms - stats.cardAtoms);
        else
        fprintf(mainOutput, "p %u %u %u\n",numOfAtoms,
            stats.cardAtoms + stats.weightAtoms, stats.rules);

        // Write the ground theory to the file(s):
        if (executionFlag&GROUT_PBS)
        writeTheoryTo2Files(mainOutput, mainOutput2);
        else
        writeTheoryToFile(mainOutput, executionFlag&GROUT_DIMACS);

        // If the theory is consistent, write the ground rules
        //  to the file (NOTE: even though the theory was
        //  cleaned, the ground atoms still remain in the AVL
        //  tree, but we don't need to write any of them in
        //  the case of inconsistent theory:
        if(THEORY_IS_CONSISTENT)
        writeAtomsToFile(mainOutput,
            (executionFlag&GROUT_LEGACY)?atomCount:0,
            programPredicateList);

        returnValue=1; // Now, we can return TRUE.
    }
    }

    return returnValue;
}


/*******************************************************************************
@2 *********     GROUND ATOM ARGUMENT CONVERSION      *************************/

/*
unsigned int GroundRules::convertGroundAtomArguments(Atom *atom)
DESCRIPTION
    This function serves as a bridge between 'Grounder' and 'GroundRules'.
    It encodes the argument list of an atom provided by 'Grounder' to an
    unsigned integer that is used in all operations of
    the class 'GroundRules'

PARAMETERS
    atom    -  Atom which argument list needs to be encoded
RETURNS
    Atom Number: Unsigned Integer that corresponds to the argument
    list of the atom for the particular program predicate.
    The pair <symbolNumber, atomNumber> can encode any ground atom

CALCULATION OF ATOM NUMBER
    Assume we have P(X1, ... Xn), where Xi are numbers or
    string constants. Let the declaration of P look like this:
    "pred P(D1, ... Dn).", where Di are data types. Let Mi be the
    number of values in the domain of Di. Assume the element Xi
    has an index Yi in the sorted list of values. Then:
    params = Y[n] + M[n]* Y[n-1] + M[n]*M[n-1]* Y[n-2] + ...
    +M[n]*...*M[3]*M[2]*Y1
*/
unsigned int GroundRules::convertGroundAtomArguments(Atom *atom)
{
    unsigned int result;
    int index, i;
    Expression *exp;

        // Go over the argument list of the atom and the array of
        //  the arguments for the program predicate:
    for(result=0, i=0, exp=atom->arg ; exp ; i++, exp=exp->next)
    {
            // Initialize the variable 'index' with the index of the
            //  argument's value in the array of the argument's type
            // This index was supposed to be calculated during
            //  grounding.
            // Also, at this moment, there are two types of arguments
            //  possible:
            // EXPR_VAR:   Index is stored in 'VarNode::valIndex'
            // EXPR_INDEX: Index is stored in 'Expression::num'
        if(exp->token==EXPR_VAR) index=exp->var->valIndex;
        else
            if(exp->token==EXPR_INDEX) index=exp->num;
            else
            {
                fprintf(stderr, "CRITICAL ERROR in Grounder::"
                "convertGroundAtomArguments\nWrong expression "
                "node type in the argument list of the program"
                " predicaete\n\n");
                exit(1);
            }

            // Calculate the result based on the given formula:
        result=result*atom->pp->params[i]->valNum + (unsigned)index;
    }

    return result;
}


/*
unsigned int GroundRules::atomNumFromTail(PPnode *pp, int dataArgIndex)
DESCRIPTION
    This method works like 'GroundRules::convertGroundAtomArguments'
    in calculating Atom Number, but it uses arguments of the data
    predicate in the tail of the program predicate instead of
    arguments of the program predicate.
PARAMETERS
    pp            -  A program predicate with tail
    dataArgIndex  -  A variable indicating a vector of arguments to be
                      converted. This is the index of the first element
                      of the vector in the array of arguments of the
                      data predicate in the tail of 'pp'.
RETURNS
    Atom Number
*/
unsigned int GroundRules::atomNumFromTail(PPnode *pp, int dataArgIndex)
{
    unsigned int result;
    int i;
    int index;

        // A program predicate has to have a tail to
        //  be used with this method:
    if(!pp->tail)
    {
        fprintf(stderr, "CRITICAL ERROR in GroundRules::atomNumFromTail"
            "\nOnly program predicates with tail "
            "can be given as parameters\n\n");
        exit(1);
    }

    for(i=0,result=0;i<pp->arity; i++, dataArgIndex++)
    {
            // Find the index of the current data argument in the
            // data type of the corresponding parameter of the
            // program predicate:
        index=EvalStack<int>::scalarSearch(
            pp->tail->arguments[dataArgIndex],
            pp->params[i]->arguments, pp->params[i]->valNum);

            // If not found, cannot convert:
        if(index < 0) return UINTEGER_MAX;

            // Otherwise, convert the index:
        result=result*pp->params[i]->valNum + (unsigned)index;
    }
    return result;
}


/*
void GroundRules::printAtom(FILE *file, int symbolno, unsigned int params)
DESCRIPTION
    Given encoded version of the ground atom (i.e. integer pair
    <symbolNumber, parameters>), the method prints regular representation
    of this ground atom.

PARAMETERS
    file     - File to print information. Can be 'stdout'.
    symbolno - Number in the symbol table of the program predicate from
                which the given atom is built.
    params   - This number encodes parameters of the program predicate.
*/
void GroundRules::printAtom(FILE *file, int symbolno, unsigned int params)
{
    PPnode *pp;
    unsigned int divisor;
    int i;

        // Print the program predicate:
    fprintf(file, "%s", symbt.getName(symbolno));
    pp=(PPnode *)((symbt.getVal(symbolno))->ptr);

        // Check whether the predicate has the argument list
        //  (decide whether we should print parentheses)
    if(pp->arity>0)
    {
        fprintf(file, "(");
        divisor=pp->numOfAtoms;

            // Check for critical errors:
        if(params >= divisor)
        {
            fprintf(stderr, "CRITICAL ERROR in GroundRules::printAtom\n"
                "atomNumber was calculated incorrectly\n\n");
            exit(1);
        }


            // Print the atom:
        for(i=0;i<pp->arity;i++)
        {
            if(i) fprintf(file, ",");
            divisor/=(unsigned)(pp->params[i]->valNum);

            strConst.printConstant(file, pp->params[i]->arguments[params/divisor]);
            params%=divisor;
        }
        fprintf(file, ")");
    }
}


/***********************************************************************
************************************************************************
@3           OUTPUT METHODS                             ***************/

    // This method is for debugging purposed only. It prints the
    //  ground theory exactly as it is stored in the pages:
void GroundRules::printPageContents()
{
        // Start from the first page:
    PageDescriptor *curPage, *firstPage=curPage=input.page->next;
    int i, pageCount=1;

    do  // The loop that goes over all pages:
    {
            // Indicate the beginning of the page:
        printf("\nPAGE %d: Size=%d\n", pageCount, curPage->size);

            // Print the contents of the page: element by element:
        for(i=0;GET_DELIM_TYPE(curPage->data[i].atom)!=DELIM_PAGE;i++)
        {
            switch(GET_DELIM_TYPE(curPage->data[i].atom))
            {
            // If the current element is a delimiter, print it:
            case DELIM_PAGE:
                printf("<P>\n");
                break;
            case DELIM_HORN:
                printf("<RH, %d> ", curPage->data[i].num2);
                break;
            case DELIM_IMPLY:
                printf("<RI, %d> ", curPage->data[i].num2);
                break;
            case DELIM_HEAD:
                printf("<H, %d> ", curPage->data[i].num2);
                break;
            // For cardinality or weight atoms, print meta-limits
            //  right away (otherwise, they might be confused with
            //  atoms):
            case DELIM_WEIGHT:
                printf("<W, %d> ", curPage->data[i].num2);
                i++;
                printf(" %d %d ", curPage->data[i].num1,
                    curPage->data[i].num2);
                break;
            case DELIM_CARD:
                printf("<C, %d> ", curPage->data[i].num2);
                i++;
                printf(" %d %d ", curPage->data[i].num1,
                    curPage->data[i].num2);
                break;
            case DELIM_WVAL:
                printf("=%d ", curPage->data[i].num2);
                break;
            default:
                // If neither is the case, the current element
                //  must be atom => print it:
                printAtom(stdout,
                    curPage->data[i].atom->symbolNum,
                    curPage->data[i].atom->atomNum);
                printf(" ");
            }
        }

        printf("\n\n");
        pageCount++;
    // Go to the next page, until we get back to the first page:
    //  (pages are in circular double-linked list).
    } while((curPage=curPage->next)!=firstPage);
    printf("\n");
}


/*
void GroundRules::reportNotGroundedPredicates(PPnode *list)
DESCRIPTION
    This method reports the program predicates that were used in rules,
    but no ground atoms were generated from them
PARAMETERS
    list  -  List of all program predicates (taken from the class 'Rules')
*/
void GroundRules::reportNotGroundedPredicates(PPnode *list)
{
    PPnode *pp;
    int isFirst;

        // Check for predicates that were used, but never grounded:
    for(pp=list,isFirst=1;pp;pp=pp->next)
        if(pp->usage==PP_USED)
        {
                // If this is a first time, print the message;
                // otherwise, print the separator:
            if(isFirst)
              fprintf(errorStream, "WARNING: The following program "
                 "predicates were used in some rule, but no ground "
                 "atoms are created from these predicates: ");
            else
                fprintf(errorStream, ", ");

            isFirst=0;
                // Print the name of the predicate:
            fprintf(errorStream, symbt.getName(pp->symbolno));
        }

    if(!isFirst) fprintf(errorStream, "\n");
}


/*
int GroundRules::writeAtomsToFile(FILE *file, unsigned int startingNumber,
                                  PPnode *list)
DESCRIPTION
    This method outputs ground atoms. It can work either in
    2 modes: normal and legacy:
        - normal mode: atoms that don't affect ground theory are assigned
                        special number GR_ANUM_NODIFF.
        - legacy mode: atoms that don't affect ground theory are treated
                        like regular atoms and assigned consequent numbers.

PARAMETERS
    file            - file to output.
    startingNumber  - If ZERO, the method works in the normal mode.
                      If NOT ZERO, the method works in legacy mode and
                       assigns numbers to atoms that don't occur in
                       the theory starting from 'startingNumber'
    list            - List of all program predicates taken from the
                       class 'Rules'. This is necessary to output
                       atoms for the predicates that were used but
                       no ground atoms were generated from them
*/
//AUXILIARY MACROS:

// This macro prints the ground atom given its symbolNumber, atomNumber, and
//  the number that is assigned to the atom.
// If 'numberForAtom==GR_ANUM_NODIFF', it is written as signed integer;
//  otherwise, it is written as unsigned:
#define OUTPUT_ATOM(symbolNumber, atomNumber, numberForAtom) \
    {if(file) \
    { \
        fprintf(file, (numberForAtom==GR_ANUM_NODIFF)?"c %d ":"c %u ", \
            numberForAtom); \
        printAtom(file, symbolNumber, atomNumber); \
        fprintf(file, "\n"); \
    }}

// This macro outputs an atom that is not in the AVL tree.
// If the legacy format is disabled, it is output with the
//  number 'GR_ANUM_NODIFF'.
// If the legacy format is enabled, the macro assigns following number
//  to the atom, checks for overflow, and prints the atom with the
//  assigned number
#define OUTPUT_ATOM_NODIFF(symbolNumber, atomNumber) \
    {if(startingNumber) \
    { \
        OUTPUT_ATOM(symbolNumber, atomNumber, startingNumber) \
        (startingNumber)++; \
        if(startingNumber==GR_ANUM_ULIMIT) return GR_ANUM_ULIMIT; \
    } \
    else \
        OUTPUT_ATOM(symbolNumber, atomNumber, GR_ANUM_NODIFF) }

// The following macro outputs a list of atoms that are not in the AVL tree,
//  given the program predicate of atoms, starting atom number ('start') and
//  the first number that SHOULD NOT be printed ('limit').
// The macro also takes into account that some predicates might have predicate
//  tail, in which case only predicates with valid arguments are printed.
// NOTE: THIS MACRO REQUIRES INITIALIZATION OF THE VARIABLE 'tailCount'.
#define OUTPUT_ATOM_NODIFF_WITH_TAIL(pp, start, limit) \
    if(!pp->tail) \
    { \
            /* No tail => just output all of them */ \
        for(i=(start);i<(limit);i++) \
            OUTPUT_ATOM_NODIFF(pp->symbolno, i); \
    } \
    else \
    { \
            /* If there is a tail, then go over all possible arguments */\
        /*of the data predicate in the tail, rather than possible */\
        /*atom numbers. NOTE: The function */\
        /*[dataPredicateArguments] --> [atomNumber] */\
        /*should be increasing due to the way */\
        /*atomNumber is calculated. */\
        for(;tailCount < pp->arity*pp->tail->valNum; \
            tailCount+=pp->arity) \
        { \
            /* Get the atom number from data predicate arguments*/\
            result=atomNumFromTail(pp, tailCount); \
            /* NOTE: the method returns UINTEGER_MAX if the */\
            /*  arguments cannot be converted to the atom number*/\
            /*  (at least one of them is out of range) */\
            if(result < (limit)) \
            { \
            /*If the result is within the limits, output the atom*/\
                if(result >= (start)) \
                    OUTPUT_ATOM_NODIFF(pp->symbolno, result) \
            } \
            else \
            /* If the result exceeded the limit and this was not*/\
            /*  an error code, break the loop */\
                if(result != UINTEGER_MAX) break; \
        } \
    }
// The method itself:
int GroundRules::writeAtomsToFile(FILE *file, unsigned int startingNumber,
                                  PPnode *list)
{
    PPnode *pp;
    AVLNode *atom, *prev;
    unsigned int i;
    unsigned int result;
    int tailCount;

        // Check whether 'startingNumber' is valid.
    if(startingNumber == GR_ANUM_ULIMIT) return GR_ANUM_ULIMIT;

        // Output the ground atoms from the predicates that were used:
    for(atom=tree.first();atom;)
    {
        pp=(PPnode *)(symbt.getVal(atom->symbolNum)->ptr);
        tailCount=0;  // Initialize the tail count.

            // For a given non-horn predicate:
            //  Output all the atoms with numbers less than number of
            // the first atom with the given predicate in the AVL tree:
        if(!atom->isHorn)
            OUTPUT_ATOM_NODIFF_WITH_TAIL(pp, 0, atom->atomNum)

            // For a given predicate: output atoms with all the numbers
            //  that are in the tree.
        for(prev=atom; atom && prev->symbolNum==atom->symbolNum;
            prev=atom, atom=tree.next())
        {

                    // If the numbers of two neighbouring atoms are not
                //  consecutive and this is not a horn atom, also
                //  output atoms with all the numbers that are
                //  in between:
            if(!atom->isHorn && prev->atomNum+1 != atom->atomNum)
                OUTPUT_ATOM_NODIFF_WITH_TAIL(pp,
                    prev->atomNum+1, atom->atomNum)

                // If the atom in the AVL tree is not false,
                //  output it:
            if(atom->value != GR_FALSE)
            {
                OUTPUT_ATOM(atom->symbolNum,
                    atom->atomNum, atom->number);
                    // The assumption is that the rules are
                    //  created correctly => if the atom is not
                    //  false, there exist data predicate
                    //  tail(...) with the given arguments.
                    // Therefore, we may increase tailCount
                    //  to take this fact into account.
                tailCount+=pp->arity;
            }
        }

            // For a given non-horn predicate:
            // Output all the atoms with numbers greater than the
            //  number of the last atom in the AVL tree with the
            //  given program predicate.
        if(!prev->isHorn)
            OUTPUT_ATOM_NODIFF_WITH_TAIL(pp,
                prev->atomNum+1, pp->numOfAtoms)
    }

        // Output ground atoms for the non-horn predicates
        //  that were used, but not grounded:
    for(pp=list;pp;pp=pp->next)
        if(pp->usage == PP_USED && pp->hornMeta!=PP_HORN)
        {
            tailCount=0;
            OUTPUT_ATOM_NODIFF_WITH_TAIL(pp, 0, pp->numOfAtoms)
        }

        // If this is legacy format, return the number of atoms in the
        //  ground theory (equal to 'startingNumber-1'); otherwise return 0:
    return ((startingNumber)?(startingNumber-1):0);
}


// This macro is used in methods that write to the human-readable output file
// It writes a symbol corresponding to ground atom or ground meta-atom and
// indicating its value (e.g. TRUE, FALSE, ...).
#define PRINT_ATOM_FLAG(file, value) \
    { if(value&GR_TEMP) fprintf(file, " `"); \
        else  fprintf(file, " "); \
        switch(value&(~GR_TEMP)) \
        { \
            case GR_NORMAL: break; \
            case GR_TRUE: fprintf(file, "+"); break; \
            case GR_FALSE: fprintf(file, "-"); break; \
            case GR_PSEUDOTRUE: fprintf(file, "*"); break; \
            case GR_PSEUDOTRUE|GR_TRUE: fprintf(file, "+*");break;\
            default: fprintf(file, "?"); \
        } \
    }

/*
void GroundRules::writeTheoryToFile(FILE *file, int isDIMACS,
                                    RuleNode *ruleSequence)
DESCRIPTION
    This method writes a sequence of the ground rules to the output file
    (it doesn't write the set of ground atoms and the line identifying the
    number of atoms and rules). It can write in both solver-readable and
    human-readable formats.

PARAMETERS
    file           -  Opened file to write the output to
    isDIMACS       -  TRUE if output should be in DIMACS format
    ruleSequence   -  OPTIONAL. If absent or NULL, the output will be
                       in solver-readable format. To output in
                       human-readable format, this parameter should be
                       pointed to the rule sequence (from class 'Rules').
*/
void GroundRules::writeTheoryToFile(FILE *file, int isDIMACS,
                                    RuleNode *ruleSequence)
{
        // Start from the first page:
    PageDescriptor *curPage, *firstPage=curPage=input.page->next;

    RuleNode *curRule=ruleSequence;
    GroundTheoryElement *curElem;
    int delim, horn, num;

        // If we are currently printing cardinality list, this contains
        // the string that indicates the end of cardinality or weight
        // atom (i.e. either "]" or "}"); otherwise, the empty string
    char *closing;

        // If we output in the DIMACS format and the current atom
        //  in in the body, this is "-"; otherwise, "".
    char *minus;

    do    // Loop over all the pages:
    {
        curElem=curPage->data;  // Get data from the current page

        while(1)  // Loop over all the rules in the current page:
        {
                // Find type of the following delimiter:
            delim=GET_DELIM_TYPE(curElem->atom);

                // Stop the loop if end of page has been reached.
            if(delim==DELIM_PAGE) break;

                // At this point there is a begining of the new rule
                //  Determine whether it is a horn rule:
            if(delim==DELIM_IMPLY) horn=0;
            else if (delim==DELIM_HORN) horn=1;
                else
                {
                    fprintf(stderr, "CRITICAL ERROR detected in method "
                        "GroundRules::writeTheoryToFile\n\n");
                    exit(1);
                }

                // Reset the variables:
            closing="";
            minus="";

            if(ruleSequence)  // If human-readable output:
            {
                    // Get the number of this ground rule:
                num=getRuleNumber(curElem);

                    // NOTE, the rules that are added after the grounding
                    //  have the ground rule number equal to zero.

                    // Find the source rule that generated this ground rule:
                if(num)
                    while(curRule && num > curRule->maxGroundRule)
                        curRule=curRule->next;

                    // If succeeded in the previous step, print the rule
                    // number and the line number where the rule came from.
                    // Otherwise, assume the rule was added after grounding:
                if(curRule && num)
                    fprintf(file, "#%d\\%d ", num, curRule->lineNum);
                else
                    fprintf(file, "ADDED:");

                    // If the theory is inconsistent and this is the rule that
                    //  caused the inconsistency, indicate this by using '@':
                if(curElem==conflictingRule) fprintf(file, "@ ");
                else
                    // Otherwise, check if the rule has been eliminated,
                    //  print the corresponding symbol:
                    switch(curElem->num2)
                    {
                    case GR_NORMAL: break;
                    case GR_ELIMINATED: fprintf(file, "$ "); break;
                    default: fprintf(file, "? ");
                    }
            }
            else  // If this is output for solver and we need to output
                // in DIMACS format, print the '-' before the following
                // atoms since they are from the body:
                if(isDIMACS) minus="-";

                // Go through the atoms in the rule:
            for(curElem++ ; IN_RULE(curElem->atom) ; curElem++)
            {
                if(IS_ATOM(curElem->atom)) // If this is not delimiter:
                {
                    if(ruleSequence) // If human-readable output:
                    {
                            // Print the atom value followed by
                            //  actual representation of the atom:
                        PRINT_ATOM_FLAG(file, curElem->atom->value);

                        printAtom(file, curElem->atom->symbolNum,
                            curElem->atom->atomNum);
                    }
                    else    // Print the atom number:
                        fprintf(file, " %s%u ", minus, curElem->atom->number);
                }
                else
                {
                    // If this is human-readable output and we are about
                    // to print a meta-atom, print the value of this meta-atom:
                    if(ruleSequence &&
                        (GET_DELIM_TYPE(curElem->atom) == DELIM_CARD ||
                        GET_DELIM_TYPE(curElem->atom) == DELIM_WEIGHT))
                           PRINT_ATOM_FLAG(file, curElem->atom->value)

                    // The current element is a delimiter:
                    switch(GET_DELIM_TYPE(curElem->atom))
                    {
                    // If this is a weight value, print it:
                    case DELIM_WVAL:
                        fprintf(file, "=%d", curElem->num2);
                        break;

                    case DELIM_HEAD:  // If head delimiter:
                        // Close the last cardinality or
                        //  weight atom if there was one:
                        fprintf(file, closing);

                        // Print the separator between body and
                        // head (don't need in DIMACS format):
                        if(!isDIMACS)
                        {
                            if(horn) fprintf(file, " : ");
                            else fprintf(file, " , ");
                        }
                            // Reset the variables:
                        closing="";
                        minus="";
                        break;

                    case DELIM_CARD:
                        // Close the last cardinality or
                        //  weight atom if there was one:
                        fprintf(file, closing);

                        // Print the cardinality limits:
                        curElem++;
                        fprintf(file, "{ %d %d ", curElem->num1, curElem->num2);
                        closing=" } "; // set closing
                        break;
                    case DELIM_WEIGHT:
                        // Close the last cardinality or
                        //  weight atom if there was one:
                        fprintf(file, closing);

                        // Print the weight limits:
                        curElem++;
                        fprintf(file, "[ %d %d ", curElem->num1, curElem->num2);
                        closing=" ] "; // set closing
                        break;
                    default:
                        fprintf(stderr, "CRITICAL ERROR in method "
                            "GroundRules::writeTheoryToFile");
                        exit(1);
                    }
                }
            }
                // Finish the rule by closing the last meta-atom:
            fprintf(file, closing);
                // Print the end of rule:
            fprintf(file, (isDIMACS)?" 0\n":"\n");
        }

        // Go to the next page and continue the loop
        //  until we get back to the first page:
    }while((curPage=curPage->next)!=firstPage);
}

/*
void GroundRules::writeTheoryTo2Files(FILE *cnffile, FILE *pbfile)

DESCRIPTION
    This method is writes a sequence of the ground rules to two output files
    (it doesn't write the set of ground atoms and the line identifying the
    number of atoms and rules). It can write in both solver-readable and
    human-readable formats.

PARAMETERS
    cnffile       -  Opened file to write the output of rules in CNF
    pbfile        -  Opened file to write the output of pseudoboolean rules
             
*/
void GroundRules::writeTheoryTo2Files(FILE *cnffile, FILE *pbfile)
{
        // Start from the first page:
    PageDescriptor *curPage, *firstPage=curPage=input.page->next;

    GroundTheoryElement *curElem, *curElem2;
    int delim, horn, numCoef, goalSE=0, goalGE=0, pass;

        // If we are currently printing cardinality list, this contains
        // the string that indicates the end of cardinality or weight
        // atom (i.e. either "]" or "}"); otherwise, the empty string
    char *closing;

        // If we output in the DIMACS format and the current atom
        //  in in the body, this is "-"; otherwise, "".
    char *minus;
    
        // Used to determine whether variables will be followed by weights
        // of a specific number or weights of 1 when printing output
    bool weight;

    // Used to loop over each page three times, this is done to print the
    // greater than or equal to followed by the less than or equal to
    pass=0;
    // the number of coefficients in each page
    numCoef=0;
    
    do    // Loop over all the pages:
    {
        curElem=curPage->data;  // Get data from the current page

        while(1)  // Loop over all the rules in the current page:
        {
                // Find type of the following delimiter:
            delim=GET_DELIM_TYPE(curElem->atom);

                // Stop the loop if end of page has been reached.
            if(delim==DELIM_PAGE) break;

                // At this point there is a begining of the new rule
                //  Determine whether it is a horn rule:
            if(delim==DELIM_IMPLY) horn=0;
            else if (delim==DELIM_HORN) horn=1;
                else
                {
                    fprintf(stderr, "CRITICAL ERROR detected in method "
                        "GroundRules::writeTheoryTo2Files\n\n");
                    exit(1);
                }

                // Reset the variables:
            closing="";
            minus="-";  // This is in dimacs, !a !b -> c format

            // Copy current element for doing multiple passes over that element
            curElem2=curElem;
                // Go through the atoms in the rule:
            for(curElem++ ; IN_RULE(curElem->atom) ; curElem++)
            {
                if(IS_ATOM(curElem->atom)) // If this is not delimiter:
                {
                    if(pass==0)  // Print the atom number:
                        fprintf(cnffile, " %s%u ", minus, curElem->atom->number);
                    else if(pass==1)        // Count number of coefficient
                        numCoef++;
                    else if(pass==2 || pass==3)// print the atom number,
                    // pass 2 will be for the greater than or equal to,
                    // pass 3 will be for the less than or equal to
                    {
                        fprintf(pbfile, "%sv%d ", minus, curElem->atom->number);
                        if(!weight)
                            fprintf(pbfile, " c1\n");
                    }        
                    
                }
                else
                {

                    // The current element is a delimiter:
                    switch(GET_DELIM_TYPE(curElem->atom))
                    {
                    // If this is a weight value, print it:
                    case DELIM_WVAL:
                        if(pass>=2)  //
                            fprintf(pbfile, " c%d\n", curElem->num2);
                        break;

                    case DELIM_HEAD:  // If head delimiter:
                    // reset the minus to conform to !a !b -> c format
                       minus="";  
                        break;

                    case DELIM_CARD:
                         if(pass==0)               
                         {
                            // Obtain the cardinality limits:
                            curElem++;
                            goalSE=curElem->num1;
                            goalGE=curElem->num2;
                            pass=1;
                            weight=false;  // There is no weight values
                          }
                          else
                              curElem++; 
                          break; 
                    case DELIM_WEIGHT:
                        if(pass==0)
                        {
                            // obtain the weight limits:
                            curElem++;
                            goalSE=curElem->num1;
                            goalGE=curElem->num2;
                            pass=1;
                            weight=true;  // There are weight values
                        }
                        else
                            curElem++;
                        break;
                    default:
                        fprintf(stderr, "CRITICAL ERROR in method "
                            "GroundRules::writeTheoryTo2Files");
                        exit(1);
                    }
                }
            }
            
            if(pass==0) // Print the end of the rule 
                fprintf(cnffile, " 0\n");
            else if(pass==1) // SE & GE has been determined and the number of
            {   // coefficients has been counted
                fprintf(pbfile, "\n#PBType: GE\n#PBGoal: %d\n#NumCoef: %d\n", goalSE, numCoef);
                curElem=curElem2;
                // make another pass on the element, this time printing the coefficients
                pass++;
            }
            else if(pass==2) // still need to print SE
            {
                fprintf(pbfile, "\n#PBType: SE\n#PBGoal: %d\n#NumCoef: %d\n", goalGE, numCoef);
                curElem=curElem2;
                // make another pass on the element, this time printing the coefficients
                pass++;
            }
            else
            {  // Reset the variables after completing the three passes
                pass=0;
                numCoef=0;
            }
       }

        // Go to the next page and continue the loop
        //  until we get back to the first page:
    }while((curPage=curPage->next)!=firstPage);
}

    // This method writes lists of ground atoms to the given human-readable
    //  output file (should be opened with permission to write).
void GroundRules::writeAtomsToHRFile(FILE *file)
{
    AVLNode *atom;

        // In the following lines, traverse the AVL tree with ground
        //  atoms and print only the ones we need.
        // Print atom value before printing the atom itself:
    fprintf(file, "\nAtoms that are determined to be true: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->value==GR_TRUE)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, "\n");
        }

    fprintf(file, "\nAtoms that are determined to be false: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->value==GR_FALSE)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, "\n");
        }


    fprintf(file,"\nAtoms that are determined to be pseudo-true: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->value==GR_PSEUDOTRUE)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, "\n");
        }

    fprintf(file, "\nAtoms that are unknown and there is a rule where they occur: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->value==GR_NORMAL && atom->firstRule)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, "\n");
        }

    fprintf(file, "\nAtoms that are unknown and rules with which were eliminated during propagation: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->value==GR_NORMAL && !atom->firstRule)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, "\n");
        }

    fprintf(file, "\nHorn atoms and the number of horn rules for every one of them: \n");
    for(atom=tree.first();atom;atom=tree.next())
        if(atom->isHorn)
        {
            PRINT_ATOM_FLAG(file, atom->value);
            printAtom(file, atom->symbolNum, atom->atomNum);
            fprintf(file, " ,%d\n",atom->hornRuleNum);
        }
}


/*
void GroundRules::printNewAtomsInStack(GroundTheoryElement *rule, int oldSize)
DESCRIPTION
    Auxiliary method which is used for scripting. It prints atoms from
    the main stack that were added recently (recently is determined
    by oldSize).

PARAMETERS
    rule     -  The method is usually called when some rule is propagated.
                 In this case, the method uses this parameter to print
                 the rule number before printing atoms. Can be NULL
    oldSize  -  This variable allowes to print only atoms that were pushed
                 onto the stack during the last operation, given that the
             operation DID NOT POP any atoms. In this case, it should
             be equal to the number of elements in the stack right
             before the operation. Setting this to zero will print
             all the atoms in the stack.
*/
void GroundRules::printNewAtomsInStack(GroundTheoryElement *rule, int oldSize)
{
    int i;
    AVLNode *atom;

        // If new atoms were added to the stack:
    if(mainStk.getNumOfElements() > oldSize)
    {
            // If the rule is given, print the ground rule number
        if(rule) fprintf(scriptOut, "<%d>", getRuleNumber(rule));

            // Go over the recently added atoms in the stack
            //  starting from the deepest:
        for(i=mainStk.getNumOfElements()-1-oldSize ; i>=0 ; i--)
        {
            atom=mainStk.getElement(i);
                // Print that atom:
            PRINT_ATOM_FLAG(scriptOut, atom->value);
            printAtom(scriptOut, atom->symbolNum, atom->atomNum);
        }
        fprintf(scriptOut, "\n");
    }
}


/**** END OF OUTPUT methods *******************************************/



/***********************************************************************
************************************************************************
@4        PAGE LEVEL (INVOLVES THE WHOLE THEORY)      ***************/


/*
void GroundRules::calculateTheoryStatistics(GroundTheoryStatistics *stats)
DESCRIPTION
    This method may be used to calculate the number of rules, cardinality
    atoms, and weight atoms. Usually, the calculation is done by compaction
    but if it is disabled, one should use this method.
PARAMETER
    stats    -  Pointer to the structure that contains all the statistics
                 of the ground theory.
NOTE
    Only fields 'rules', 'cardAtoms', and 'weightAtoms' are modified
    (the ones that are caluculated). The values of the other fields
    remain without changes.
*/
void GroundRules::calculateTheoryStatistics(GroundTheoryStatistics *stats)
{
    GroundTheoryElement *cur;
    PageDescriptor *page=input.page;

        // Initialize counters:
    stats->cardAtoms=stats->weightAtoms=stats->rules=0;

    do  // Loop over all the pages
    {
            // Go over all the elements in the current page:
        for(cur=page->data;GET_DELIM_TYPE(cur->atom)!=DELIM_PAGE;cur++)
                // If we find delimiters for rules or meta-atoms,
                //  increase the corresponding count
            switch(GET_DELIM_TYPE(cur->atom))
            {
            case DELIM_HORN:
            case DELIM_IMPLY:
                stats->rules++;
                break;
            case DELIM_CARD:
                stats->cardAtoms++;
                cur++; // To avoid bumping into meta limits.
                break;
            case DELIM_WEIGHT:
                stats->weightAtoms++;
                cur++; // To avoid bumping into meta limits.
                break;
            }

        // Go to the next page until we find the one we started with:
    }while((page=page->next) !=input.page);
}

/*
int GroundRules::impossibleCPLEX()
DESCRIPTION
    This method may be used to verify whether an output in the CPLEX format
    is possible. It is possible if all pseudoboolean atoms form unit rules.
    Returns 1 if an output in CPLEX format is impossible, 0 otherwise
*/
int GroundRules::impossibleCPLEX()
{
    GroundTheoryElement *cur;
    PageDescriptor *page=input.page;
    int body=0, meta=0, simple=0;

    do  // Loop over all the pages
    {
    // Go over all the elements in the current page:
    for(cur=page->data;GET_DELIM_TYPE(cur->atom)!=DELIM_PAGE;cur++)
        if (IS_DELIM(cur->atom)) 
        switch(GET_DELIM_TYPE(cur->atom)) {
        case DELIM_HORN:
        case DELIM_IMPLY:
            body=1;
            meta=simple=0;
            break;
        case DELIM_CARD:
        case DELIM_WEIGHT:
            if (body || meta || simple)
            return 1;
            meta++;
            cur++; // To avoid bumping into meta limits.
            break;
        case DELIM_HEAD:
            body=0;
            break;
        }
        else if (!meta)
        simple++;

    // Go to the next page until we find the one we started with:
    }while((page=page->next) !=input.page);
    return 0;
}
/*
void GroundRules::compactRules(int compactionFlag,GroundTheoryStatistics *stats)
DESCRIPTION
    During propagation and lookahead, atoms and rules that should be
    eliminated are assigned special values but not physically eliminated.
    Thus, they still take memory and consume processing time when
    propagation or lookahead methods traverse them.
    The compaction DOES THE PHYSICAL ELIMINATION of atoms, meta-atom,
    and rules that were assigned value or eliminated. It uses the flags
    set up by the propagation and lookahead, IT DOES NOT CHECK WHETHER
    A RULE IN FACT SHOULD BE ELIMINATED OR ATOM ASSIGNED A PARTICULAR
    VALUE.
    This method builds (or re-builds) the links that link all the
    rules with a given ground atom. Also, for every horn atom it may
    calculate the number of horn rules that have that atom in the head.
    In addition, it may calculate ground rule statistics such as the
    number of rules, cardinality or weight atoms.

PARAMETERS
    compactionFlag    -  Can be ONE of the following:
        GR_CM_LASTRULE  - Compact only the last rule. If compaction is
                           enabled, the method with this parameter is
                           called every time a ground rule is added to
                           the theory and propagated.

        GR_CM_ALL       - Compact the whole theory and rebuild all links,
                           but don't calculate the number of horn rules
                                   this is done time after time when theory is
                           very big in order to reduce its size.

        GR_CM_ALLFINAL  - Compact the whole theory, rebuild all links,
                           and calculate the numbers of horn rules.
                           This should be done ONLY AFTER ALL RULES
                           ARE GROUNDED; otherwise, horn atoms might
                            be propagated incorrectly.

    stats             -  OPTIONAL parameter. If this parameter is given,
                          the compaction also calculates the ground
                          theory statistics.
*/
void GroundRules::compactRules(int compactionFlag,GroundTheoryStatistics *stats)
{
    AVLNode *atom;
    PageDescriptor *inPage;
    GroundTheoryElement *in, *tmp;
    RuleInputData out;
    int weight, value, trueWeight=0, unknownWeight=0;
    int continueFlag=1;
    unsigned int rules=0, cardAtoms=0, weightAtoms=0;

    // Use the variable 'weight' to initialize the variable 'hornRuleNum'
    // for every atom (if necessary).
    // If all rules are grounded to this moment, hornRuleNum should be
    //  initialized to zero and then increased for every horn rule encountered.
    // Otherwise, it will be set to UINTEGER_MAX and not changed in the future.
    weight=0;

    // Initialize the variables depending on what kind of
    //  compaction we need to do
    switch(compactionFlag)
    {
    case GR_CM_ALL:
        weight=UINTEGER_MAX;
    case GR_CM_ALLFINAL:
            // Initialize variables:
        inPage=input.page->next;
        in=inPage->data;
        out.cur=out.rule=in;
        out.page=inPage;

            // For every atom in the AVL tree, initialize variable hornRuleNum
            //  and reset the links.
        for(atom=tree.first();atom;atom=tree.next())
        {
            if(atom->isHorn) atom->hornRuleNum=weight;
            atom->firstRule=0;
        }
        break;
    case GR_CM_LASTRULE:
            // Initialize the variables:
        inPage = out.page = input.page;
        in = out.cur = out.rule = input.rule;
        break;
    default:
        fprintf(stderr, "CRITICAL ERROR in GroundRules::compactRules\n"
                "error in parameter\n\n");
        exit(1);
    }

    out.nMeta=0;
    GET_LIMIT_FOR_PAGE(out.limit, out.page); // Initialize 'out.limit'.

      // 'continueFlag' is set to FALSE once we reach the end of the last page:
    while(continueFlag) // Main loop
    {
            // Check whether we need to go to the NEXT OUTPUT PAGE:
        if(out.cur >= out.limit && GET_DELIM_TYPE(in->atom) != DELIM_PAGE)
        {
                //  Move the unfinished part of the current
                //   rule to the next page. Reset the variables:
            moveToNextPage(&out, 0);

                // Now need to repair all the links for the rule that has
                //  just been moved:
                // Since for every atom in the current rule 'AVLNode::firstRule'
                //  should point to this rule, we need to reset these
                // 'firstRule' to point to the NEW location of the rule:
            for(tmp=out.rule+1 ; tmp<out.cur ; tmp++)
            {
                    // Reset the link:
                if(IS_ATOM(tmp->atom)) tmp->atom->firstRule=out.rule;
                else // Skip the meta-limits:
                    if(NEED_TO_SKIP_ONE(tmp->atom)) tmp++;
            }
        }

            // If the current input element is an atom:
        if(IS_ATOM(in->atom))
        {
            value=in->atom->value;

                // We need to eliminate (i.e. skip and don't copy) the atom
                // if it has a value and the value is not temporary, except
                // when the value is PSEUDO-TRUE and this is a horn rule.
            if(value!=GR_NORMAL && !(value & GR_TEMP) &&
                (GET_DELIM_TYPE(out.rule->atom) != DELIM_HORN ||
                !GR_IS_PSEUDOTRUE(value)))
            {
                in++; // Go to the next atom.

                if(out.nMeta)  // If we are inside a meta-atom
                {
                    // If the next value is a weight value, read it.
                    if(GET_DELIM_TYPE(in->atom) == DELIM_WVAL)
                        {weight=in->num2; in++; }
                    else weight = 1;

                        // If the current atom is true, adjust the meta-values
                    if(GR_IS_TRUE_I(value)) trueWeight+=weight;
                }
            }
            else  // Can not eliminate the atom: should copy it to the output
            {
                    // Copy the atom:
                out.cur->atom=in->atom;

                    // Set the link:
                linkOneAtom(out.cur, out.rule);

                in++; out.cur++; // Increase the counts:

                    // If this is a meta-atom, increase 'unknownWeight'.
                    // In addition, if this is a weight atom, copy the weight
                if(out.nMeta)
                {
                    // If the next value is a weight value, read it.
                    if(GET_DELIM_TYPE(in->atom) == DELIM_WVAL)
                      {unknownWeight+=in->num2; *out.cur=*in; in++; out.cur++;}
                    else unknownWeight++;
                }
            }
        }
        else
        {
            // Finish the current meta atom if necessary:
            //  NOTE: We should not get here if the delimiter is weight value,
            //   because it is processed together with the atom it belongs to.
            if(out.nMeta)
            {
                    // Adjust the meta-limits by subtracting the weight of
                    //  the eliminated atoms with True values:
                out.rule[out.nMeta].num1-=trueWeight;
                out.rule[out.nMeta].num2-=trueWeight;

                    // Adjust the meta limits so that minimum will be at
                    //  least zero and maximum will be at most the
                    //  maxumum possible weight:
                if(out.rule[out.nMeta].num1 < 0) out.rule[out.nMeta].num1=0;
                if(out.rule[out.nMeta].num2 > unknownWeight)
                        out.rule[out.nMeta].num2=unknownWeight;

                out.nMeta=0;  // No more current meta-atom.
            }

                // Process every delimiter based on its type:
            switch(GET_DELIM_TYPE(in->atom))
            {
            case DELIM_CARD:
            case DELIM_WEIGHT:
                    // If the meta atom has non-temporary value,
                    // we should eliminate (skip) this atom.
                if(in->num2!=GR_NORMAL && !(in->num2 & GR_TEMP))
                    for(in++; IN_ATOM_LIST(in->atom); in++);
                else
                {
                        // Increase cardinality atom count or weight atom count
                    if(GET_DELIM_TYPE(in->atom)==DELIM_CARD) cardAtoms++;
                    else weightAtoms++;

                        // Copy the delimiter:
                    *out.cur = *in; in++; out.cur++;
                        // Initialize ptr. to meta limits:
                    out.nMeta = out.cur - out.rule;
                        // Copy the meta-limits:
                    *out.cur = *in; in++; out.cur++;
                        // Initialize the weight counts:
                    trueWeight = unknownWeight = 0;
                }
                break;

            case DELIM_HORN:
            case DELIM_IMPLY:
                    // Set the variable 'out.rule' to point to the new rule:
                out.rule=out.cur;

                   // If the rule has non-temporary value, we should
                   //  eliminate (skip) the rule:
                if(in->num2!=GR_NORMAL && !(in->num2 & GR_TEMP))
                    for(in++; IN_RULE(in->atom); in++);
                else
                {
                        // Increase the rule count;
                    rules++;

                        // Copy the delimiter.
                    *out.cur = *in; in++; out.cur++;
                }
                break;

            case DELIM_HEAD:
                    // Convert the horn rule to an implication if necessary:
                if(GET_DELIM_TYPE(out.rule->atom) == DELIM_HORN)
                {
                    if(!IS_ATOM(in[1].atom))
                    {
                        fprintf(stderr, "CRITICAL ERROR in "
                         "GroundRules::compact\nwrong horn rule detected\n\n");
                        exit(1);
                    }
                        // If the horn atom in the head of this horn rule is
                        //  either true or false, we need to convert the rule
                        //  to an implication.
                        // Otherwise, we may need to increase the horn rule
                        //  count for the given horn atom
                    if(GR_IS_TRUE_OR_FALSE(in[1].atom->value))
                        out.rule->atom=MAKE_DELIM(DELIM_IMPLY);
                    else
                        if(in[1].atom->hornRuleNum != UINTEGER_MAX)
                            in[1].atom->hornRuleNum++;
                }
                    // Copy the head delimiter:
                *out.cur = *in;
                in++; out.cur++;
                break;

            case DELIM_PAGE:
                    // Rule cannot span several pages. Therefore the current
                    //  rule is finished and we need to set 'out.rule'
                    //  right after the end of the rule:
                out.rule=out.cur;
                    // If this is a last page, stop the main loop:
                if(inPage == input.page) continueFlag=0;
                    // Go to the next page
                inPage=inPage->next;
                    // Set the input at the beginning of the next page's data:
                in=inPage->data;
                break;
            default:
                fprintf(stderr, "CRITICAL ERROR in the method "
                                "GroundRules::compactRules\n"
                                "Detected error in the ground rules\n\n");
                exit(1);
            }
        }
    }

    // COMPACTION IS DONE:

        //  Put DELIM_PAGE:
    out.cur->atom=MAKE_DELIM(DELIM_PAGE);
    out.cur->num2=0;

    if(compactionFlag != GR_CM_LASTRULE)
    {  // If the compaction was done over the whole theory:
            //  Release all unused memory, i.e. all pages after the
            //   current up to the last (including the last):
        if(out.page != input.page) // If 'out.page' is not the last page
        {
            do // Loop over pages
            {
                    // Release the page right after 'out.page'
                inPage=out.page->next;
                DLL_REMOVE(inPage);
                deletePage(inPage);
                    // While released page IS NOT the last one
            } while(inPage != input.page);
        }

        // Update the input rule data: copy from 'out' to 'input'
        memcpy(&input, &out, sizeof(RuleInputData));

        // Set the corresponding fields in the 'stats' structure if necessary:
        if(stats)
        {
            stats->cardAtoms=cardAtoms;
            stats->weightAtoms=weightAtoms;
            stats->rules=rules;
        }
    }
    else // Need to reset only one variable:
        input.cur=out.cur;
}

/**** END OF PAGE LEVEL    *******************************************/


/***********************************************************************
************************************************************************
@5        MEMORY MANAGEMENT METHODS                     ***************/

/*
static PageDescriptor *GroundRules::pageAllocOrExpand(
                       int minSize, PageDescriptor *page)
DESCRIPTION
    Allocate a new memory page or expand an existing page.
PARAMETERS
    minSize    -  OPTIONAL. Minimum allowed number of ground rule elements
                   in the new page allocated. The actual number of elements
                   will be:  MAX ( COMMON_PAGE_SIZE , minSize ).
    page       -  OPTIONAL: The page to expand. If not given or NULL
                   then the new page will be allocated.
                   NOTE: if 'page' is not NULL, 'minSize' has to be
                   greater than the current size of the page; errors
                   might be otherwise. NOTE: the data in the page
                   is preserved.
RETURNS
    Pointer to the created or expanded page.
*/
PageDescriptor *GroundRules::pageAllocOrExpand(int minSize,PageDescriptor *page)
{
    void *ptr;
    size_t size;

        // Find the number of elements to in the page and the
        //  actual size of the page:
    if(minSize<COMMON_PAGE_SIZE) minSize=COMMON_PAGE_SIZE;
    size=sizeof(GroundTheoryElement)*minSize + sizeof(PageDescriptor);

        // Allocate or reallocate the memory:
    if(page) ptr = realloc(page->data, size);
    else ptr = malloc(size);

    if(!ptr) // Check for errors
    {
      fprintf(stderr, "CRITICAL ERROR in GroundRules::pageAllocOrExpand\n"
        "Cannot allocate memory of the size %d\n\n", size);
        exit(1);
    }

    // NOTE: The page descriptor is stored at the end, after the data:
    page=(PageDescriptor *)((GroundTheoryElement *)ptr + minSize);

        // Initialize the page descriptor:
    page->data=(GroundTheoryElement *)ptr;
    page->size=minSize;
    page->next=page->prev=0;

    return page;
}


    // Deallocate the given page (opposite to 'pageAllocOrExpand'):
void GroundRules::deletePage(PageDescriptor *page) { free(page->data); }


/*
void GroundRules::moveToNextPage(RuleInputData *in, int needAlloc)
DESCRIPTION
    This method should be used while writing a rule to a page
    when there is no more space in the page to write. The method
    allocates the new page if necessary, copies unfinished rule
    to the new page (since rule cannot span several pages) and
    updates the valus of the pointers.

PARAMETERS
    in         -  Rule input data. The same structure that is used to
                   input a rule. Contains all the necessary information.
    needAlloc  - OPTIONAL. This determines what to do with the next page.
                  If it is missing or TRUE, then the method assumes that
                  the current page is the last page in the list and
                  have to allocate the new page. If this is FALSE,
                  the method assumes this is not the last page
                  and takes the next page in the list instead of
                  reallocating
               NOTE: This parameter should be ommitted for
                     rule input and FALSE for compaction
*/
void GroundRules::moveToNextPage(RuleInputData *in, int needAlloc)
{
    PageDescriptor *next;

        // If there is only one unfinished rule in the current page,
        //  then we should expand the current page instead
        //  of allocating another one.
    if(in->page->data == in->rule)
    {
        next=in->page->next;  // Save the pointer to the next page:

        if(next != in->page)
        {  // If there is more than one page in
           //  the circular double-linked list:
            DLL_REMOVE(in->page)   // Remove the page from the list
                // Reallocate the page:
            in->page = pageAllocOrExpand(
                in->page->size*PAGE_GROWTH_MUL, in->page);
                // Append the page to the list:
            DLL_APPEND_BEFORE(in->page, next)
        }
        else
        {  // There is only one page in the circular double-linked list.
           // CAN NOT perform operations on the list:
                // Just reallocate the page:
            in->page=pageAllocOrExpand(
                in->page->size*PAGE_GROWTH_MUL, in->page);
                // Initialize the links:
            in->page->prev=in->page->next=in->page;
        }
    }
    else
    {
        if(needAlloc) // If we need to allocate the new page:
        {
                // Allocate:
            next=pageAllocOrExpand(
                (in->cur-in->rule)*PAGE_GROWTH_MUL);
                // Append the page to the double linked
                //  list after the current page:
            DLL_APPEND_AFTER(next, in->page)
        }
        else
        {
                // Save the pointer to the next page:
            next=in->page->next;

                // If there is only one page in the list,
                //  the current page is automatically the last,
                //  which should not be in this case:
            if(next == in->page)
            {
                fprintf(stderr, "CRITICAL ERROR in "
                    "GroundRules::moveToNextPage\n");
                exit(1);
            }

              // Check whether the rule will fit into the next page:
            if(!DOES_RULE_FIT_PAGE(in->cur-in->rule, next))
            {
                    // We need to increase the size of the
                    // current page in order to fit the rule in:
                DLL_REMOVE(next); // Remove from the list
                next = pageAllocOrExpand( // Expand
                   (in->cur-in->rule)*PAGE_GROWTH_MUL, next);
                DLL_APPEND_AFTER(next, in->page) // Place back
            }
        }

        // Copy the rule from the previous page to the newly allocated:
        memcpy(next->data, in->rule,
               (in->cur - in->rule) * sizeof(GroundTheoryElement));

            // Put the endpage delimiter to the previous page:
        in->rule->atom=MAKE_DELIM(DELIM_PAGE);
        in->rule->num2=0;

            // Switch input to the newly allocated page
            //  by resetting 'in->page'
        in->page=next;
    }

        // Reset the values of the variables:
    in->cur  = in->page->data + (in->cur  - in->rule);
    in->rule = in->page->data;
    GET_LIMIT_FOR_PAGE(in->limit, in->page);
}


    // Set up the first page to start the rule input.
    // Used in constructor and 'clearTheory'.
void GroundRules::allocateAndInitializeFirstPage()
{
    input.page=pageAllocOrExpand();  // Allocate the page
        // Initialize the link for circular double-linked list:
    input.page->next=input.page->prev=input.page;

        // Initialize the other fields of 'input' structure:
    input.cur=input.page->data;
    GET_LIMIT_FOR_PAGE(input.limit, input.page);
    input.rule=input.cur;
    input.nMeta=0;

        // Place the end-of page delimiter:
    input.cur->atom=MAKE_DELIM(DELIM_PAGE);
    input.cur->num2=0;
}


    // Release all memory allocated for the pages:
void GroundRules::deleteAllPages()
{
    PageDescriptor *temp, *page;

        // Break up the circular doubly linked list.
    input.page->prev->next=0;

        // Delete all the pages
    for(page=input.page;page;temp=page,page=page->next,deletePage(temp));
}


    // Remove all the rules from the ground theory:
void GroundRules::clearTheory(GroundTheoryStatistics *stats)
{
    deleteAllPages();
    allocateAndInitializeFirstPage();
        // Set all numbers in the statistics to zero:
    memset(stats, 0, sizeof(GroundTheoryStatistics));
}

/*         END OF  MEMORY MANAGEMENT METHODS              **************/


/************************************************************************
*************************************************************************
@6           HIGH LEVEL PROPAGATION (SETS OF RULES)  ********************
SIMILARITIES BETWEEN METHODS:
    1) Don't care about page structure of the ground theory.
    2) Rules are traversed in the sequence given by nextRule links.
    2) DOES traverse AVL tree of atoms.
    2) USE next rule links
    3) Provide information regarding conflicts, if they are found.
*/


/*
unsigned int GroundRules::completePropagationAndAssignNumbers(
    int isDIMACSfmt, int isLegacyFormat, GroundTheoryStatistics *stats)
DESCRIPTION
    Purpose of this method:
    1) Assign numbers to atoms.
    2) For every horn atom A that is pseudo-true, create a rule: "->A"
    3) For every horn atom A that doesn't have corresponding horn rules,
       but is not false (occurrs when propagation is disabled), create
       a rule: "A->."
    4) Calculate numbers of true, false, and unknown atoms. Increase
       the number of rules in the theory by the number of added rules
        5) if the output in the DIMACS or PBS format is requested,
       a) for every ground atom A determined to be true, append a rule
          "->A" to the ground theory
       b) for every ground atom A that does not make any difference
          (i.e. it can be added to every model of the ground theory),
          append a rule "A->A" to the ground theory
PARAMETERS
    isDIMACSfmt - TRUE if output is requested in DIMACS or PBS format

    isLegacyFormat    - If TRUE, the output will be in legacy format,
                         i.e. the atom numbers will be assigned to
                         all atoms.
                        If FALSE, the atoms that don't occur in the
                         ground theory will be assigned 'GR_ANUM_NODIFF'
    stats             - Pointer to the structure that stores the ground
                         theory statistics. The method calculates the
                         fields: 'atomsTrue', 'atomsFalse',
                         'atomsUnknown'; updates 'rules' and leaves
                         all the other fields untouched. Can be NULL.
RETURNS
    GR_ANUM_ULIMIT    -  Failure. The number of atoms is greater than
                          the number of integers
    Otherwise         -  Success. The returned value is the number of
                          atoms that were assigned numbers. Also, it
                          is the next number that may be assigned
                          to a new atom.
*/
unsigned int GroundRules::completePropagationAndAssignNumbers(
         int isDIMACSfmt, int isLegacyFormat, GroundTheoryStatistics *stats)
{
    unsigned int atomNumber, addedRules=0;
    unsigned int atomsTrue=0, atomsFalse=0, atomsUnknown=0;
    unsigned int atomsNodiff=0;
    AVLNode *atom;

        // Go over all the ground atoms:
    for(atomNumber=1, atom=tree.first() ; atom ; atom=tree.next())
    {
        if(atom->value==GR_NORMAL) // If the atom doesn't have any value
        {
              // Checking the case where we have a horn atom A,
              //  that should has been made FALSE, but it wasn't
              //  (e.g. when propagation is disabled).
              // In this case we need to add the rule: "A->."
            if(atom->isHorn && atom->hornRuleNum==0)
            {
                addSimpleRule(atom, 0);
                addedRules++;
            }

                // If this atom occurs in at least one rule,
                //  it is considered unknown:
            if(atom->firstRule) atomsUnknown++;

                // If the atom is unknown or output is in legacy
                //  format, assign the number to the atom.
            if(atom->firstRule || isLegacyFormat)
            {
                atom->number=atomNumber++;
                    // Check for overflow:
                if(atomNumber==GR_ANUM_ULIMIT) break;
            } else if (isDIMACSfmt) {
                addSimpleRule(atom,atom);
                addedRules++;
                atomsNodiff++;
                atom->number=atomNumber++;
                if(atomNumber==GR_ANUM_ULIMIT) break;
            } else   // Otherwise, mark the atom as the one
                   //  that doesn't make any difference
                atom->number=GR_ANUM_NODIFF;
        }
        else
        {
            // If there is a horn atom A that has pseudo-true value,
            // we need to add the rule to the theory: "->A."
            if(atom->value==GR_PSEUDOTRUE)
            {
                    // Add the rule:
                addSimpleRule(0, atom);
                addedRules++;
                    // Assign the atom number:
                atom->number=atomNumber++;
                atomsUnknown++;// The atom is treated as unknown
                if(atomNumber==GR_ANUM_ULIMIT) break;
            }
            else    // Otherwise
            {
                    // Increase corresponding atom count:
                if(atom->value==GR_TRUE) {
		    atomsTrue++;
                    if (isDIMACSfmt) {
			addSimpleRule(0, atom);
			addedRules++;
			atom->number=atomNumber++;
			if(atomNumber==GR_ANUM_ULIMIT) break;
                    } else {
			    // Assign the value to the atom:
			atom->number=GR_ANUM_TRUEFALSE;
		    }
                } else if(atom->value==GR_FALSE) {
		    atomsFalse++;
		    atom->number=GR_ANUM_TRUEFALSE;
		} else {
                      fprintf(stderr, "CRITICAL ERROR in "
                      "the method GroundRules::"
                      "completePropagationAndAssignNumbers\n"
                      "Detected wrong atom value\n\n");
                      exit(1);
		}
            }
        }
    }
        // Save the numbers in the structure.
        // It is assumed that the variable 'stats->rules' has already
        //  been initialized and it will be increased here.
    if(stats)
    {
        stats->atomsFalse=atomsFalse;
        stats->atomsTrue=atomsTrue;
        stats->atomsUnknown=atomsUnknown;
        stats->rules+=addedRules;
	stats->atomsNodiff=atomsNodiff;
    }

        // Return the number that otherwise would be
        //  asssigned to the next atom:
    return atomNumber;
}


/*
int GroundRules::propagateHornAtoms(int calculateNumbersOnly)
DESCRIPTION
    For every horn atom, this method calculates the number of rules that
    have that atom in the head. If this is zero, the method assigns
    FALSE value to the horn atom and propagates (if enabled).
PARAMETERS
    calculateNumbersOnly  -  If this flag is set (TRUE) the method
        doesn't do any propagation: it just calculates the numbers
        of horn rules.
RETURNS
    TRUE  - everything is OK,
    FALSE - the theory is inconsistent.
*/
int GroundRules::propagateHornAtoms(int calculateNumbersOnly)
{
    AVLNode *atom;
    GroundTheoryElement *cur, *next;
    int isIncrease;

        // Go over all the ground atoms in the theory:
    for(atom=tree.first() ; atom ; atom=tree.next())
        if(atom->isHorn)
        {
                // Find number of the horn rules that have
                //  the given horn atom in the head:
            for(atom->hornRuleNum=0, cur=atom->firstRule ;
                cur ; cur=next)
            {
                // Check whether the given rule can possibly
                //  increase the horn rule count
                //  for the given atom:
                if(cur->num2&GR_TEMP)
                {
                    fprintf(stderr, "CRITICAL ERROR in "
                    "GroundRules::propagateHornAtoms\n\n");
                    exit(1);
                }

                if(GET_DELIM_TYPE(cur->atom) == DELIM_HORN &&
                        !(cur->num2&GR_ELIMINATED))
                    isIncrease=1;
                else isIncrease=0;

                for(next=0,cur++;IN_RULE(cur->atom);cur++)
                {
                    // Skip cardinality and weight limits:
                    if(NEED_TO_SKIP_ONE(cur->atom))
                    { cur++; continue; }
                        // Look for the next rule:
                    if(cur->atom==atom && cur->nextRule)
                        next=cur->nextRule;
                    // If the atom in the head of this horn
                    //  rule is the given atom, increase
                    //  the horn rule count.
                    if(GET_DELIM_TYPE(cur->atom)==DELIM_HEAD
                        && isIncrease && cur[1].atom==atom)
                        atom->hornRuleNum++;
                }
            }

                // If there are no rules that have the given
                //  horn atom in the head, then propagate.
            if(!atom->hornRuleNum && !calculateNumbersOnly)
            {
                  // If the script is enabled and the
                  //  atom will be propagated:
                if(controlFlag&GR_DO_SCRIPT &&
                   (atom->value==GR_NORMAL ||
                   atom->value==GR_PSEUDOTRUE))
                {
                    // Print the atom to the script:
                    fprintf(scriptOut, "HORN_NO_RULES: %c",
                     (atom->value==GR_PSEUDOTRUE)?'*':'-');
                    printAtom(scriptOut, atom->symbolNum,
                              atom->atomNum);

                    // If the atom causes the conflict,
                    //  print this also to the script:
                    if(atom->value==GR_PSEUDOTRUE)
                        fprintf(scriptOut, " CONFLICT");
                    fprintf(scriptOut, "\n");
                }


                if(atom->value == GR_NORMAL)
                {
                    // Assign FALSE value to the atom and
                    //  push it for further propagation:
                    atom->value = GR_FALSE;
                    mainStk.push(atom);
                }
                else  // Value is PSEUDO-TRUE => conflict
                    if(atom->value == GR_PSEUDOTRUE)
                    {
                        conflictingAtom=atom;
                        return 0;
                    }
            }
        }

        // If we only need to calculate the numbers of horn rule,
        //  return TRUE, since the conflict could not be found
    if(calculateNumbersOnly) return 1;
        // Otherwise, finish propagation:
    else return finishPropagation(0);
}


/*
int GroundRules::lookahead()
DESCRIPTION
    This method performs one-step lookahead on the ground theory.
RETURNS
    TRUE  - everything is OK.
    FALSE - the theory is inconsistent
*/
int GroundRules::lookahead()
{
    AVLNode *atom;

        // The following variable is FALSE by default but becomes TRUE
        //  once some atom was assigned a value during lookahead.
    int madeChange;

    mainStk.clear(); // Clean the main stack.

    do
    {
            // Go over all the ground atoms:
        for(madeChange=0, atom=tree.first();atom ;atom=tree.next())
                // Lookahead should be done only for atoms that don't have any
                //  assigned value. Also, it doesn't make sense to propagate
                //  on atoms that don't appear in any rule.
            if(atom->value == GR_NORMAL && atom->firstRule)
            {
                    // Try to assign the True value:
                atom->value=(atom->isHorn)? (GR_PSEUDOTRUE | GR_TEMP):
                                        (GR_TRUE | GR_TEMP);

                mainStk.push(atom); // Push the atom for propagation

                    // If the theory became inconsistent, assign the false value
                    //  and propagate.
                if(!finishPropagation(1))
                {
                        // If we need to do scripting,
                        //  repeate the previous step again:
                    if(controlFlag&GR_DO_SCRIPT)
                    {
                            // Set this flag so that propagation method will
                            //  output propagated atom despite the lookahead
                            //  flag (by default, if the lookahead flag is set,
                            //  the propagated atoms are not being output).
                        controlFlag|=GR_SCRIPT_LA;

                            // Repeat previous step:
                        atom->value=(atom->isHorn)? (GR_PSEUDOTRUE | GR_TEMP):
                                        (GR_TRUE | GR_TEMP);
                        mainStk.push(atom);

                            // Print the pushed atom with its value:
                        fprintf(scriptOut, "TRY:");
                        printNewAtomsInStack(0, 0);

                        finishPropagation(1); // Propagate

                            // Clear the flag:
                        controlFlag&=~GR_SCRIPT_LA;

                            // Indicate that the atom will be
                            //  assigned the opposite value:
                        fprintf(scriptOut, "ASSIGN:-");
                        printAtom(scriptOut, atom->symbolNum, atom->atomNum);
                        fprintf(scriptOut, "\n");
                    }

                        // Assing the value to the atom and push for propagation
                    madeChange=1;
                    atom->value=GR_FALSE;
                    mainStk.push(atom);

                        // If now the theory is also inconsistent, return FALSE.
                    if(!finishPropagation(0)) return 0;
                }
                else // True value didn't work, try assigning False value:
                {
                        // Now, try to assign the false value:
                    atom->value=GR_FALSE | GR_TEMP;

                    mainStk.push(atom);
                    if(!finishPropagation(1))
                    {
                            // If we need to do scripting,
                            //  repeate the previous step again:
                            // Everything is similar to "TRUE" case:
                        if(controlFlag&GR_DO_SCRIPT)
                        {
                            controlFlag|=GR_SCRIPT_LA;
                            atom->value=GR_FALSE | GR_TEMP;
                            mainStk.push(atom);

                            fprintf(scriptOut, "TRY:");
                            printNewAtomsInStack(0, 0);

                            finishPropagation(1);

                            controlFlag&=~GR_SCRIPT_LA;
                            fprintf(scriptOut, "ASSIGN:%c",
                                (atom->isHorn)?'*':'+');
                            printAtom(scriptOut,atom->symbolNum,atom->atomNum);
                            fprintf(scriptOut, "\n");
                        }

                            // Assign the value to the atom and push it:
                        madeChange=1;
                        atom->value=(atom->isHorn)? GR_PSEUDOTRUE : GR_TRUE;
                        mainStk.push(atom);

                            // If the theory is inconsistent, return FALSE.
                        if(!finishPropagation(0)) return 0;
                    }
                }
            }
        // If some value was assigned to at least one atom,
    //  repeat the lookahead again:
    }while(madeChange);

    return 1;
}

/*        END OF HIGH LEVEL                        *********************/


/************************************************************************
*************************************************************************
@7      MIDDLE LEVEL PROPAGATION (SETS OF RULES)     ********************
SIMILARITIES BETWEEN METHODS:
    1) Don't care about page structure of the ground theory.
    2) Rules are traversed in the sequence given by nextRule links.
    2) Does NOT traverse AVL tree of atoms.
    2) USE next rule links
    3) Provide information regarding conflicts, if they are found.
*/

    // Add links to the atoms in the rule. Normally, this is done by
    // compaction, but this method is used when compaction is disabled
void GroundRules::linkRule(GroundTheoryElement *rule)
{
    GroundTheoryElement *cur;

        // Go over all the atoms in the rule:
    for(cur=rule+1 ; IN_RULE(cur->atom) ; cur++)
    {
            // Link every atom:
        if(IS_ATOM(cur->atom)) linkOneAtom(cur, rule);
        else    // Skip meta-limits:
            if(NEED_TO_SKIP_ONE(cur->atom)) cur++;
    }
}


/*
int GroundRules::finishPropagation(int isLookahead)
DESCRIPTION
    The central method for propagation and lookahead. It does the actual
    propagation on atoms starting from those that are in the main stack.
EXAMPLE OF USAGE
    1) Assign FALSE value to atom A:
    2) mainStk.push(A);
    2) finishPropagation(0).
PARAMETERS
    isLookahead    -  FALSE means to do regular propagation
                   -  TRUE:
                       1) Similar to the regular propagation, but all the
                          assigned values are temporary
                       2) Perform cleaning by returning all the atoms with
                          temporary values back to their previous state
RETURNS
    TRUE  -  Everything is OK: the theory is consistent
    FALSE -  The theory is inconsistent
NOTE: If isLookahead==TRUE, the returned value reflects the state between
       steps (1) and (2), i.e. before the temporary values are removed.
*/
int GroundRules::finishPropagation(int isLookahead)
{
    AVLNode *curAtom;
    GroundTheoryElement *curRule, *nextRule;
    int isConflict=0;// Becomes TRUE once theory is found to be inconsistent
    int oldNumOfElements;

    clearanceStk.clear(); // Make sure clearance stack is empty.

    // STEP 1. THE PROPAGATION
        // While there are some atoms that were assigned values and
        //  no inconsistency is detected:
    while(mainStk.getNumOfElements() && !isConflict)
    {
        curAtom=mainStk.pop();  // Pop the current atom.
            // If lookahead, save the atom for the further clearance:
        if(isLookahead) clearanceStk.push(curAtom);

            // Traverse all the rules where the atom occurrs:
        for(curRule=curAtom->firstRule ; curRule ; curRule=nextRule)
        {
                // Save the number of elements in the stack:
                //  (need if we do script)
            oldNumOfElements=mainStk.getNumOfElements();
                // Propagate the current rule:
            nextRule=propagateOneRule(mainStk, curRule,
                curAtom, isLookahead);

               // If we need to do script, which happens if it is
               //  enabled and either this is not lookahead or
               //  script is specially requested during lookahead
               //  via the flag 'GR_SCRIPT_LA':
            if( (controlFlag&GR_DO_SCRIPT) && (!isLookahead ||
                (controlFlag&GR_SCRIPT_LA)))
            {
                    // If there is a conflict, print this.
                    // Otherwise, print the atoms that
                    //  were assigned values:
                if(nextRule==curRule)
                    fprintf(scriptOut, "<%d>:CONFLICT\n",
                        getRuleNumber(curRule));
                else
                    printNewAtomsInStack(curRule,
                        oldNumOfElements);
            }

            if(nextRule == curRule)  // If there is a conflict:
            {
                    // Set the flag to stop the outer loop:
                isConflict=1;
                    // If this is not a lookahead, save the
                    //  information about conflict into the
                    //  corresponding variables to be reported
                    //  in the hunan-readable output:
                if(!isLookahead)
                {
                    conflictingRule=curRule;
                    conflictingAtom=curAtom;
                }
                break; // Break the inner loop.
            }
        }
    }

    // STEP 2. CLEANING
        // NOTE: If there was lookahead, all the atoms that have been
        //       popped from the 'mainStk' are now in 'clearanceStk'
        // Cleaning requires to traverse the same rules as were traversed
        //  during propagation.

        // If lookahead, Clean the atoms that remain in the 'mainStk':
    if(isLookahead)
    {
        while(mainStk.getNumOfElements()) // While stack is not empty
        {
                // Pop the atom from the main stack and clean it:
            curAtom=mainStk.pop();
            GR_CLEAR_ATOM(curAtom);
        }
    }
    else  // Not a lookahead, just clean the stack:
        mainStk.clear();

        // Reset the value of the atoms and rules affected by the lookahead
        //  (in case of propagation, the clearanceStk is empty and the loop
        //   will not be executed anyway)
    while(clearanceStk.getNumOfElements())
    {
        curAtom=clearanceStk.pop(); // Pop the next atom

            // Clear the value of the atom
        GR_CLEAR_ATOM(curAtom);

            // Clear all the rules where the atom occurrs.
        for(curRule=curAtom->firstRule; curRule;
            curRule=clearOneRule(curRule, curAtom));
    }

        // Return TRUE or FALSE depending on whether
        //  the theory was (is) inconsistent:
    if(isConflict) return 0; else return 1;
}


/*
int GroundRules::propagateRule(GroundTheoryElement *rule)
DESCRIPTION
    High-level propagation. The method starts propagation from a rule.
    If the propagation is enabled, this method should be used for
    every new rule added to the ground theory.
    Uses 'propagateOneRule' and 'finishPropagation' methods.
PARAMETERS
    rule    -  Rule to start propagation from.
RETURNS
    TRUE - everything is OK.
    FALSE - the theory is inconsistent.
*/
int GroundRules::propagateRule(GroundTheoryElement *rule)
{
    mainStk.clear();  // Clean the main stack.

      // Propagate the given rule:
    if(rule == propagateOneRule(mainStk, rule, 0, 0))
    { // If the theory is inconsistent:
          // If we do script, write the information about inconsistency:
        if(controlFlag&GR_DO_SCRIPT)
            fprintf(scriptOut, "<%d>: CONFLICT\n",
                getRuleNumber(rule));

          // Save the rule that caused the conflict and return FALSE:
        conflictingRule=rule; return 0;
    }

      // If we do scripting, print the atoms that were assigned values
      //  the result of propagation on the 'rule':
    if(controlFlag&GR_DO_SCRIPT)
        printNewAtomsInStack(rule, 0);

      // Finish the propagation:
    return(finishPropagation(0));
}


/*
void GroundRules::addSimpleRule(AVLNode *A, AVLNode *B)
DESCRIPTION
    Given two ground atoms A and B, the method creates a simple
    implication "A->B." and adds it to the ground theory.
    This might be used when horn atoms have PSEUDO-TRUE value;
    horn atoms have to be FALSE, but propagation is disabled;
    or when we need to output inconsistent rule "->."
PARAMETERS
    A    -  Atom to be in the body of the implication.
             Can be NULL which means the body is empty.
    B    -  Atom to be in the head of the implication.
             Can be NULL which means the head is empty.
*/
void GroundRules::addSimpleRule(AVLNode *A, AVLNode *B)
{
        // This method cannot be called if we are in the process of
        //  ground rule input. The previously input rule has to
        //  be finished:
    if(state!=4)
    {
        fprintf(stderr, "CRITICAL ERROR in GroundRules::addSimpleRule"
            "\n\nTrying to create a new rule but have not "
            "finished input of the previous rule\n\n");
        exit(1);
    }

    input.rule=input.cur;  // Indicate that the rule is started
        // Initialize 'input.limit':
    GET_LIMIT_FOR_PAGE(input.limit, input.page);

        // If we don't have enough space for 4 rule elements
        //  (<rule> A <head> B) go to the next page.
        // NOTE: 'input.cur + 0 >= input.limit'
        //  would check for one available element
    if(input.cur + 3 >= input.limit) moveToNextPage(&input);

        // Output the beginning of the rule:
    input.cur->atom=MAKE_DELIM(DELIM_IMPLY);
    input.cur->num2=0;
    input.cur++;

        // Output the first atom:
    if(A)
    {
        input.cur->atom=A;
        linkOneAtom(input.cur, input.rule); // Link 'A'.
        input.cur++;
    }

        // Output the head delimiter:
    input.cur->atom=MAKE_DELIM(DELIM_HEAD);
    input.cur->num2=0;
    input.cur++;

        // Output the second atom:
    if(B)
    {
        input.cur->atom=B;
        linkOneAtom(input.cur, input.rule); // Link 'B'.
        input.cur++;
    }

        // Place the endpage delimiter:
    input.cur->atom=MAKE_DELIM(DELIM_PAGE);
    input.cur->num2=0;

        // Adjust the rule start variable:
    input.rule=input.cur;
}

/*      END OF MIDDLE LEVEL                  *******************************/



/****************************************************************************
*****************************************************************************
@8        LOW LEVEL PROPAGATION (RULES)      ********************************
SIMILARITIES BETWEEN METHODS:
    1) Don't care about page structure of the ground theory.
    2) LOOK FOR the next rule link, but DON'T USE it.
    3) Deal with atoms in the rule
    4) Modify the values of Rules, Meta-Atoms, or Atoms, but don't
        modify physical structure of a rule.
    5) DON'T traverse atom AVL tree.
    6) All the methods are static
****************************************************************************/

/*
GroundTheoryElement *GroundRules::clearOneRule(GroundTheoryElement *rule,
                                               AVLNode *targetAtom)
DESCRIPTION
    The method is used during lookahead to clear the
    temporary values IN ONE RULE.
    It looks through the rule and resets the temporary values of
    meta-atoms or temporary eliminate flag back to 'GR_NORMAL'.
    If this is a horn rule eliminated during lookahead, the method
    increases 'AVLNode::hornRuleNum' for the horn atom in the head
    of this rule. This method also looks for the 'nextRule' link.
PARAMETERS
    rule        -  The rule to clear
    targetAtom  -  This method is called for every rule of 'targetAtom'
                    to be cleared. This parameter is necessary so that
                    the method can return the next rule to clean
                    (the next rule where 'targetAtom' occurs).
RETURNS
    The pointer to the next rule where the 'targetAtom' occurs or
    NULL if this was the last rule.
*/
GroundTheoryElement *GroundRules::clearOneRule(GroundTheoryElement *rule,
                                               AVLNode *targetAtom)
{
    GroundTheoryElement *nextRule=0;
        // This variable is TRUE => This is a horn rule and we need to
        //  increase the horn rule count for the atom in the head.
    int hornIncrease=0;
        // Clear the rule delimiter:
        // If this rule is a horn rule and it has to be restored from
        //  being eliminated, we need to increase horn rule count for
        //  the horn atom in the head of this rule.
    if(rule->num2 & GR_TEMP)
    {
        if(GET_DELIM_TYPE(rule->atom) == DELIM_HORN) hornIncrease=1;

        rule->num2=GR_NORMAL;  // Restore the rule value:
    }

        // Clear all meta-atoms in the rule. Find the next rule.
        //  Increase horn rule count if necessary.
    for(rule++; IN_RULE(rule->atom); rule++)
    {
            // Looking for the next rule
        if(rule->atom==targetAtom && rule->nextRule)
            nextRule=rule->nextRule;

        if(!IN_ATOM_LIST(rule->atom))
        {
                // Clearing the meta atoms
            if(GET_DELIM_TYPE(rule->atom) == DELIM_CARD ||
                GET_DELIM_TYPE(rule->atom) == DELIM_WEIGHT)
            {
                if(rule->num2 & GR_TEMP) rule->num2=GR_NORMAL;
                rule++; // Skip the meta-limit
                continue;
            }

                // Increae the horn rule count if necessary.
            if(hornIncrease&&GET_DELIM_TYPE(rule->atom)==DELIM_HEAD)
            {
                    // Make sure the horn rule has exactly
                    // one horn atom in the head:
                if(!IS_ATOM(rule[1].atom) ||
                    !rule[1].atom->isHorn)
                {
                     fprintf(stderr, "CRITICAL ERROR in"
                        "GroundRules::clearOneRule\n"
                        "Wrong horn rule detected\n\n");
                    exit(1);
                }

                    // Increasing the horn rule count.
                rule[1].atom->hornRuleNum++;
            }
        }
    }

    return nextRule;
}


/*
GroundTheoryElement *GroundRules::propagateMetaAtom
        (StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *&cur,
        AVLNode *targetAtom, int forcedValue)
DESCRIPTION
    This method propagates the given cardinality or weight atom.
    It can be used for two functions:
      1) ANALYSIS: Find the VALUE OF THE META-ATOM based on the values
          of the ground atoms inside it.
      2) PROPAGATION: Force meta-atom to have some value, when the
          meta-atom is the only atom in the rule and it has to be either
          True or False.
          NOTE: forcing doesn't succeed in every case. It may happen
          that the atom will still be unknown.
    The function assigns values only; it DOES NOT PHYSICALLY MODIFY
    the meta-atom. The method also looks for the 'nextRule; link.
    NOTE: This method is static.
IMPORTANT NOTE
    This function retains values in some variables between its call.
    Need to call it with the second parameter referencing zero to reset
    the variables. This is necessary when the rule is PHYSICALLY
    removed(or changed) or some atoms change their values.
    This better be done before or after propagating every rule.

PARAMETERS
    mainAtomStk  -  Main propagation stack ('mainStk'). Need to give the
                     reference since the method is static.
    cur          -  Reference to the pointer that points to the beginning
                     of the meta-atom. Upon return this will point to the
                     element right after the meta-atom.
    targetAtom   -  Atom for which we need to find the next rule.
    forcedValue  -  Can be one of the following: GR_NORMAL, GR_TEMP,
                     GR_TRUE, GR_FALSE, GR_TRUE|GR_TEMP, GR_FALSE|GR_TEMP.
                     1) Flag GR_TEMP should be present if and only if the
                        method is used during lookahead
                     2) If neither GR_TRUE nor GR_FALSE is present, the
                        method finds the value of the meta-atom based on
                        the ground atoms in the meta-list
                     3) If either of GR_TRUE or GR_FALSE is present, the
                        method forces the meta-atom to be either true
                        or false.
RETURNS
    Pointer to the beginning of the next rule or NULL if the pointer could
    not be found inside this meta-atom.
*/
GroundTheoryElement *GroundRules::propagateMetaAtom(
        StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *&cur,
        AVLNode *targetAtom, int forcedValue)
{
      // Static variables. Their values are used in the next call of the
      //  method if the method was called two times with the same value
      //  of 'cur'.
      // NOTE: value of 'cur' is retained in the variable 'prev'.
      //  If 'cur==0', the method should reset the variable 'prev'
      //  and don't do anything else.
    static GroundTheoryElement *prev=0;
    static int min, max, unknownWeight;

      // If 'cur' is zero, we just need to reset the method:
    if(!cur) {prev=0; return 0;}

      // Local variables:
    GroundTheoryElement *cur1, *metaAtom;
    GroundTheoryElement *nextRule=0;  // Link to the next rule to be found
    int trueWeight, weight;
    short atomValue, hornAtomValue;
    short isWeight; // TRUE for weight atoms, FALSE for cardinality atoms
    short isChange;
    char T_T, T_F, F_T, F_F;

      // Find whether this is cardinality or weight atom:
    if(GET_DELIM_TYPE(cur->atom)==DELIM_CARD) isWeight=0;
        else if(GET_DELIM_TYPE(cur->atom) == DELIM_WEIGHT) isWeight=1;
            else // No other case is acceptable
            {
                fprintf(stderr, "CRITICAL ERROR in "
                    "GroundRules::propagateMetaAtom\n"
                    "The given pointer does not "
                    "point to a meta atom\n\n");
                exit(1);
            }

      // Checking whether the value of the given meta atom has already
      // been determined. If yes, there is nothing else to do; just skip
      // this meta atom. During this procedure we have to look for a
      // link to a next rule.
    if(cur->num2!=GR_NORMAL)
    {
        for(cur+=2;IN_ATOM_LIST(cur->atom);cur+=1+isWeight)
            if(cur->atom==targetAtom && cur->nextRule)
            { nextRule=cur->nextRule;  break; }

        return(nextRule);
    }

    // ANALYSIS PHASE:
    // GO THROUGHT THE META-LIST, DETERMINE IF META-ATOM
    //  ATOM IS TRUE OR FALSE. IF YES, ASSIGN THE VALUE:

    // NOTE: If the previous call of the function was done with the same
    //        we don't really need to do this again, since the value is
    //        already assigned and all the necessary values were saved
    //        in the static variables.
    if(cur!=prev)
    {  // If this time the method is called with the different atom:
        prev=cur; // Save the atom for the next call.

          // Initialize the meta-limits:
        cur++;
        min=cur->num1;
        max=cur->num2;

          // Go over the meta-list and calculate the
        for(cur++, unknownWeight=0, trueWeight=0, weight=1 ;
            IN_ATOM_LIST(cur->atom) ; cur+=1+isWeight)
        {
              // Find the weight of the current atom (in cardinality
              //  atom 'weight' will always be equal to one).
            if(isWeight) weight=cur[1].num2;

              // Looking for a link to the next rule:
            if(cur->atom==targetAtom && cur->nextRule)
                nextRule=cur->nextRule;

              // NOTE: atoms should be considered true as they are
              //  true in implications (i.e. pseudo-true is
              //  cosidered true). If the meta atom occurs in a
              //  horn rule, it should NOT contain horn atoms.
            if(cur->atom->value==GR_NORMAL)
                unknownWeight+=weight;
            else if(GR_IS_TRUE_I(cur->atom->value))
                    trueWeight+=weight;
        }

          // Update 'min' and 'max' to take into account the values
          //  of the true atoms (false atoms don't matter) so that
          //  these will be the limits for unknown atoms only:
        min-=trueWeight;
        if(min<0) min=0;  // "min<0" doesn't make sense: the same as 0
        max-=trueWeight;

          // In these cases the atom is always false no matter what are
          //  the values of the atoms in the meta-list.
        if(max<min || unknownWeight<min)
        {
              // Assing the false value to the meta-atom and put
              //  the temporary flag if it is lookahead.
            prev->num2=GR_FALSE + (forcedValue&GR_TEMP);
              // Once the atom value is determined, there is
              //  nothing else we can do with the atom:
            return(nextRule);
        }

          // In these cases the atom is always true no matter what are
          //  the values of the atoms in the meta-list:
        if(min==0 && unknownWeight<=max)
        {
              // The same as in FALSE case:
            prev->num2=GR_TRUE + (forcedValue&GR_TEMP);
            return(nextRule);
        }
    }

      // Check whether we need to do the PROPAGATION phase:
    if(!(forcedValue&(~GR_TEMP))) return (nextRule);

      // Reset the function state: we won't need the saved value
      //  after this stage:
    metaAtom=prev;
    prev=0;

    // PROPAGATION PHASE:

      //  NOTE: at this moment variables prev, min, max, unknownWeight
      //         are assumed to be known.
      //  NOTE: 'nextRule' is not calculated if the method was called the
      //         second time with the same atom:
      //  NOTE: any meta atom without unknown atoms is known whether
      //         true or false and should not appear here.
      //  NOTE: See documentation regarding propagation on weigh atoms
      //         for more details on algorithms used in this stage.

    if(GET_DELIM_TYPE(metaAtom->atom)==DELIM_CARD)
    {  // PROPAGATE CARDINALITY ATOM:
        atomValue=GR_NORMAL;

          // NOTE: at this moment, there should be no contradictions in
          //  the cardinality atom. i.e. if its value does not depend on
          //  the values of the atoms in the cardinality list, this
          //  should have been determined on the previous stage.
        if(metaAtom->num2!=GR_NORMAL)
        {
            fprintf(stderr, "CRITICAL ERROR in "
                "GroundRules::propagateMetaAtom\n\n");
            exit(1);
        }

          // If the cardinality atom is forced to be true
        if(GR_IS_TRUE_I(forcedValue))
        {
              // Condition 4: F --> F.
            if( unknownWeight <= min ) atomValue=GR_TRUE;

              // Condition 3: T --> F, check for contradictions
            if( max == 0 ) atomValue=GR_FALSE;
        }
        else
        {
              // Condition 2: F --> T
            if( min == 0 && unknownWeight - 1 <= max )
                atomValue=GR_TRUE;

              // Condition 1: T --> T; check for contradictions:
            if( min <=1 && unknownWeight <= max )
                atomValue=GR_FALSE;
        }

          // If any of these conditions was true, then the
          //  cardinality atom can be propagated and the
          //  values of all atoms can be determined.
        if(atomValue!=GR_NORMAL)
        {
              // Determine the value that will
              //  be assigned to the horn atoms:
            hornAtomValue=(atomValue==GR_TRUE)?
                GR_PSEUDOTRUE:atomValue;

              // Setting the temporary flag if necessary:
            atomValue+=(forcedValue&GR_TEMP);
            hornAtomValue+=(forcedValue&GR_TEMP);

              // Cardinality atom propagated =>
              //  assigned the forced value.
            metaAtom->num2=forcedValue;

              // Assign the necessary values to the atoms:
            for(cur1=metaAtom+2; IN_ATOM_LIST(cur1->atom); cur1++)
                if(cur1->atom->value==GR_NORMAL)
                {
                      // Assign the proper value to atom:
                    cur1->atom->value=(cur1->atom->isHorn)?
                        hornAtomValue:atomValue;
                      // Push the atom onto stack for
                      //  subsequent propagation:
                    mainAtomStk.push(cur1->atom);
                }
        }

          // We are done with the cardinality atom:
        return nextRule;
    }

    // PROPAGATE WEIGHT ATOM:
    do  // Loop while there there is at least one
        //  atom that was assigned value:
    {
          // Go over the weight list:
        for(isChange=0,cur1=metaAtom+2; IN_ATOM_LIST(cur1->atom); cur1+=2)
        {
              // Need to consider only unknown atoms:
            if(cur1->atom->value!=GR_NORMAL) continue;

            atomValue=GR_NORMAL;
                // Read the weight of the atom:
            weight=cur1[1].num2;

                // Calculating the values of the conditions:

            T_T = (min - weight <= 0 && unknownWeight <= max); // 1
            F_T = (min == 0 && unknownWeight - weight <= max); // 2
            T_F = (max - weight < 0);                          // 3
            F_F = (unknownWeight - weight < min);              // 4

              // Check whether the weight atom will have some value
              //  regardless of the forced value.
            if(T_T && F_T)
            {
                metaAtom->num2=GR_TRUE + (forcedValue&GR_TEMP);
                return nextRule;
            }

            if(T_F && F_F)
            {
                metaAtom->num2=GR_FALSE + (forcedValue&GR_TEMP);
                return nextRule;
            }

              // Force the current atom:
            if(GR_IS_TRUE(forcedValue))
            {
                if(F_F) atomValue=GR_TRUE;
                if(T_F) atomValue=GR_FALSE;
            }
            else
            {
                if(F_T) atomValue=GR_TRUE;
                if(T_T) atomValue=GR_FALSE;
            }

              // If the atom is forced to have some value:
            if(atomValue!=GR_NORMAL)
            {
                  // Update the values of the min,
                  //  max, and unknownWeight:
                unknownWeight-=weight;
                if(atomValue==GR_TRUE)
                { min-=weight; max-=weight; }

                  // If the atom is a horn atom and is forced
                  // to be true, its value becomes pseudo-true.
                if(atomValue==GR_TRUE && cur1->atom->isHorn)
                    atomValue=GR_PSEUDOTRUE;

                  // Assign the value to the atom, set flag for
                  //  the next iteration of the weight atom
                  //  propagation, and push the
                  //  atom to the stack.
                cur1->atom->value = atomValue +
                    (forcedValue&GR_TEMP);
                isChange=1;
                mainAtomStk.push(cur1->atom);
            }
        }
      // Do the next iteration if at least one atom was assigned a value:
    }while(isChange);

    return nextRule;
}


/*
GroundTheoryElement *GroundRules::propagateOneRule(
        StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *rule,
        AVLNode *targetAtom, int isLookahead)
DESCRIPTION
    Propagates a given rule (uses 'propagateMetaAtom'). The method does not
    physically modify the rule but rather assign values to the atoms,
    meta-atoms, or the rule.
    NOTE: This method should be static.
PARAMETERS
    mainAtomStk  -  Main propagation stack ('mainStk'). Need to give the
                     reference since the method is static.
    rule         -  Pointer to the beginning of the rule to propagate
    targetAtom   -  Atom for which we need to find the next rule.
    isLookahead  -  TRUE if the method is called during lookahead;
                    FALSE otherwise
RETURNS
    NULL - there is no next rule for the given atom.
    THE SAME POINTER AS GIVEN IN THE PARAMETER - the theory is inconsistent
    Other - the next rule to propagate.
*/
GroundTheoryElement *GroundRules::propagateOneRule(
        StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *rule,
        AVLNode *targetAtom, int isLookahead)
{
    int isHorn;         // Indicates whether this is a horn rule.
      // The following flag is set to TRUE, once it is detected that the
      //  rule has to be eliminated. After this is set, the method may
      //  go through the atoms faster.
    int isEliminated=0;
      // Pointer to the last unknown program atom or meta-atom in the rule:
    GroundTheoryElement *last=0;
      // Pointer to the head delimiter:
    GroundTheoryElement *head=0;
    int numPseudoTrue=0;   // The number of PSEUDO-TRUE atoms

      // The number of UNKNOWN atoms. NOTE: after we make sure that
      //  the horn rule is not reduced to implication, the number of
      //  PSEUDO-TRUE atoms is added to this number
    int numOfAtoms=0;

      // Non-important local variables:
    GroundTheoryElement *cur, *metaAtom, *temp;
    int toForce;
    GroundTheoryElement *nextRule=0;

      // Determine the type of the rule:
    if(GET_DELIM_TYPE(rule->atom)==DELIM_IMPLY) isHorn=0;
    else if(GET_DELIM_TYPE(rule->atom)==DELIM_HORN) isHorn=1;
        else
        {
            fprintf(stderr, "CRITICAL ERROR in the method GroundRules"
                "::propagateOneRule\nWrong parameter given\n\n");
            exit(1);
        }

      // If the rule is eliminated, do nothing except find the next rule:
    if(rule->num2&GR_ELIMINATED)
    {
        for(cur=rule+1; IN_RULE(cur->atom); cur++)
        {
                // Skip meta-limits
            if(NEED_TO_SKIP_ONE(cur->atom)) { cur++; continue; }
                // Check for the next rule:
            if(cur->atom==targetAtom && cur->nextRule)
                nextRule=cur->nextRule;
        }

        return(nextRule);
    }

        // Reset the state of the method propagateMetaAtom:
    cur=0; propagateMetaAtom(mainAtomStk, cur, 0, 0);

    if(isLookahead) isLookahead=GR_TEMP;

    // 1) PROCESSING ATOMS IN THE BODY:
    //   A) Processing simple atoms in the body.
    for(cur=rule+1;IN_ATOM_LIST(cur->atom) && !isEliminated;cur++)
    {
          // Looking for the next rule:
        if(cur->atom==targetAtom&&cur->nextRule) nextRule=cur->nextRule;

          // NOTE: The atom that has both flags GR_TRUE and
          //  GR_PSEUDOTRUE on, is considered to be true and
          //  therefore doesn't matter here.
        switch(cur->atom->value&(~GR_TEMP))
        {
        case GR_NORMAL:     numOfAtoms++;    last=cur; break;
        case GR_PSEUDOTRUE: numPseudoTrue++;           break;
        case GR_FALSE:      isEliminated=1;
        }
    }

    //   B) Processing meta atoms in the body
    while(GET_DELIM_TYPE(cur->atom) != DELIM_HEAD && !isEliminated)
    {
        metaAtom=cur;  // Save the pointer to the meta-atom.
          // Find the value of the meta-atom:
        temp=propagateMetaAtom(mainAtomStk,cur,targetAtom,isLookahead);
        if(temp) nextRule=temp; // Save the next rule if found

          // If the meta-atom doesn't have any value, increment the
          //  number of unknown atoms and save this as last atom;
          // otherwise, if it is False, eliminate the rule:
        if(metaAtom->num2==GR_NORMAL) {numOfAtoms++; last=metaAtom;}
        else if(GR_IS_FALSE(metaAtom->num2)) isEliminated=1;
    }

      // Making the pointer 'head' point to the beginning of the head:
    if(!isEliminated) { head=cur; cur++; }

    // 2) PROCESSING ATOMS IN THE HEAD:
    if(isHorn && !isEliminated)
    {  // Process the horn atom in the head:

          // Make sure that the atom in the head is in fact horn atom:
        if(!IS_ATOM(cur->atom))
        {
            fprintf(stderr, "CRITICAL ERROR in the method GroundRules::"
                "propagateOneRule\nWrong horn rule detected(1)!\n\n");
            exit(1);
        }

            // Look for the next rule:
        if(cur->atom==targetAtom&&cur->nextRule) nextRule=cur->nextRule;

        if(GR_IS_TRUE(cur->atom->value)) // If this horn atom is true
            isEliminated=1; // The rule should be eliminated:
        else
            if(GR_IS_FALSE(cur->atom->value))
                //The head of the horn rule is false => the horn
                // rule should be considered as an implication:
                isHorn=0;
            else // We are dealing with a horn rule:
            {
                  // In this case, pseudo-true atoms
                  //  should be regarded as normal
                numOfAtoms+=numPseudoTrue+1;
                last=cur;
            }

            // Let the 'cur' point to the next atom, which should
            //  be either a rule delimiter or a page delimiter.
        cur++;
    }
    else  // Processing the head of an implication:
    {
        // A) Processing simple atoms in the head.
        for(;IN_ATOM_LIST(cur->atom) && !isEliminated;cur++)
        {
              // Look for the next rule:
            if(cur->atom==targetAtom && cur->nextRule)
                nextRule=cur->nextRule;

              // Don't need to consider pseudo-true atoms
              //  separately, since we know this is an implication
            if(cur->atom->value==GR_NORMAL)
            { numOfAtoms++; last=cur; }
            else
                if(GR_IS_TRUE_I(cur->atom->value)) isEliminated=1;
        }

        //  B) Processing meta atoms in the head
        while(IN_RULE(cur->atom) && !isEliminated)
        {
            metaAtom=cur;  // Save the pointer to the meta-atom
            temp=propagateMetaAtom(mainAtomStk, cur, // propagate
                targetAtom, isLookahead);
            if(temp) nextRule=temp; // Save the next rule

              // If the meta-atom doesn't have any value, increment
              //  the number of unknown atoms, save the last atom;
              // otherwise, if it is True,  eliminate the rule:
            if(metaAtom->num2==GR_NORMAL)
            {numOfAtoms++; last=metaAtom;}
            else if(GR_IS_TRUE(metaAtom->num2)) isEliminated=1;
        }
    }

    // 3) RULE ELIMINATION:
    if(isEliminated) // If the rule is eliminated
    {
          // Set the rule flag accordingly:
        rule->num2=GR_ELIMINATED + isLookahead;

          // Traverse the remaining part of the rule
          //  in order to find 'head' and 'nextRule'
        for(; IN_RULE(cur->atom); cur++)
        {
              // Skip the meta-limits:
            if(NEED_TO_SKIP_ONE(cur->atom)) { cur++; continue; }
              // Look for the head:
            if(GET_DELIM_TYPE(cur->atom) == DELIM_HEAD) head=cur;
              // Look for the next rule:
            if(cur->atom==targetAtom && cur->nextRule)
                nextRule=cur->nextRule;
        }

        if(isHorn)  // If the horn rule is to be eliminated:
        {
	      /* 
              // Check for consistency:
            if(!head || !IS_ATOM(head[1].atom))
            {
                fprintf(stderr, "CRITICAL ERROR in the method "
                    "GroundRules::propagateOneRule\n"
                    "Wrong horn rule detected(2)!\n\n");
		if (!head)
		    fprintf(stderr, "!head\n");
		else
		    fprintf(stderr, "!IS_ATOM(head[1].atom) %d\n",
			GET_DELIM_TYPE(head[1].atom));
                exit(1);
            }
	    */

            head++; // Now head will point to the horn atom:

              // Since this horn rule is eliminated,
              //  decrease horn rule count
            head->atom->hornRuleNum--;

              // If there are no more horn rules for the given atom
              //  and the atom has not been assigned a value,
              // then it should be false
            if(head->atom->hornRuleNum==0)
            {
                if(head->atom->value==GR_NORMAL)
                {
                      // Assign False value to the atom and
                      //  push it into the stack for
                      //  further propagation
                    head->atom->value=GR_FALSE+isLookahead;
                    mainAtomStk.push(head->atom);
                }
                else
                      // If the value of the atom is pseudo-
                      // true (GR_TRUE flag has to be down),
                      // the theory is inconsistent:
                    if((!GR_IS_TRUE(head->atom->value)) &&
                       GR_IS_PSEUDOTRUE(head->atom->value))
                        return rule;

              // Do nothing if the value is already True or False.
            }
        }

          // Nothing more needs to be done for the eliminated rule:
        return nextRule;
    }

    //  3) PROPAGATING THE RULE (we get here if the rule is NOT eliminated)
      //    If there are no unknown atoms, then the theory is inconsistent:
    if(numOfAtoms==0) return rule;

      // If there is only one unknown atom, we can propagate:
    if(numOfAtoms==1)
    {
        // At this moment, the 'last' pointer should point
        //  to this atom.
        // Checking whether this atom is in the head or in the body:
        //  Determine which value should be forced to this atom:
        if(last<head) toForce=GR_FALSE; else toForce=GR_TRUE;

        if(IS_ATOM(last->atom)) // We propagate a simple atom
        {
              // If 'last' points to the horn atom, this is
              //  implication, and the atom is forced to be
              //  true, it becomes pseudo-true:
            if(last->atom->isHorn && !isHorn && toForce==GR_TRUE)
                toForce=GR_PSEUDOTRUE;

              // Force the atom to have the given value:
              // NOTE: if this is lookahead, atom is forced to
              //  have GR_TRUE value, and it has already been
              //  assigned GR_PSEUDOTRUE during propagation,
              //  then we should keep this pseudo-true flag so
              //  that it will remain after clearing:
            if(isLookahead && GR_IS_PSEUDOTRUE(last->atom->value) &&
                    !(last->atom->value&GR_TEMP))
                last->atom->value|=toForce + isLookahead;
            else
                last->atom->value=toForce + isLookahead;

              //  Push the atom into the stack:
            mainAtomStk.push(last->atom);

              // After this, the rule should be
              //  automatically eliminated:
            rule->num2=GR_ELIMINATED + isLookahead;
        }
        else // we propagate meta-atom:
        {
            cur=last;
            toForce+=isLookahead;  // Include the lookahead flag
              // Force the atom to have required value:
            propagateMetaAtom(mainAtomStk,cur,targetAtom,toForce);

              // If the resulting value of the meta atom is the same
              //  as forced, then rule may be eliminated:
            if(last->num2==toForce)
                rule->num2=GR_ELIMINATED + isLookahead;
            else  // Otherwise, if the resulting value of meta-atom
                  // is other than GR_NORMAL,
                  //  then the theory is inconsistent.
                if(last->num2!=GR_NORMAL) return rule;
        }

          // If this was a horn rule, we need to adjust the hourn rule
          //  count for the atom in the head (no matter what its value)
        if(isHorn)
        {
              // Again, check for consistency:
            if(!head || !IS_ATOM(head[1].atom))
            {
                fprintf(stderr, "CRITICAL ERROR in the method "
                    "GroundRules::propagateOneRule\n"
                    "Wrong horn rule detected(3)!\n\n");
                exit(1);
            }
              // Since this horn rule is eliminated,
              //  decrease horn rule count
            head[1].atom->hornRuleNum--;

            // NOTE: Even if hornRuleNum is now 0, no propagation
            //  can be done for this atom, because it has already
            //  been assigned some value (otherwise,
            //  we wouldn't be in this place)
        }
    }

      // Return the next rule link:
    return nextRule;
}


    // Find the number of the given ground rule. That number should be
    //  the second number in the ground rule element that contains
    //  the head delimiter:
    // NOTE: should be static method.
int GroundRules::getRuleNumber(GroundTheoryElement *rule)
{
        // Find the head delimiter and return the rule number
        //  that should be stored there:
    for(rule++;IN_RULE(rule->atom);rule++)
        if(NEED_TO_SKIP_ONE(rule->atom)) rule++;
        else
            if(GET_DELIM_TYPE(rule->atom)==DELIM_HEAD)
                return(rule->num2);
    return 0;
}

/*             END OF LOW LEVEL                  *********************/


/**********************************************************************
***********************************************************************
@9           INPUT INTERFACE                    **********************/

/*
void GroundRules::startRule(int horn, int lineNum, int groundRuleNum)
DESCRIPTION
    This method should be called by the grounder to start
    the rule input.
PARAMETERS
    horn           -  TRUE for horn rules, FALSE for implications
    lineNum        -  Line number of the source rule in the rule file
    groundRuleNum  -  Number of this ground rule. These numbers are
                       consequently assigned by the grounder.
*/
void GroundRules::startRule(int horn, int lineNum, int groundRuleNum)
{
        // Make sure the method is called in time:
    if(state<4 || !isHeadI) inpError("startRule");

        // Save the parameters for the use by the other input methods:
    isHorn=horn;
    curLineNum=lineNum;
    curGroundRuleNum=groundRuleNum;

        // Set the new values of the state variables:
    state=0;
    isHeadI=0;

    input.rule=input.cur; // Indicate that the new rule is started:
        // Initialize 'input.limit':
    GET_LIMIT_FOR_PAGE(input.limit, input.page);

        // Check whether we can write the rule delimiter.
        //  If not, get more memory:
    if(input.cur >= input.limit) moveToNextPage(&input);

        // Write the rule delimiter and increase the count:
    input.cur->atom=MAKE_DELIM((isHorn)?DELIM_HORN:DELIM_IMPLY);
    input.cur->nextRule=GR_NORMAL;
    input.cur++;

        // Clear the stacks used to eliminate repetitions:
    mainStk.clear();
    clearanceStk.clear();
}


/*
void GroundRules::addAtom(Atom *atom, int weight)
DESCRIPTION
    This function outputs ground atom or weight assignment to the
    ground theory (works in any case: body, head, inside, or
    outside of meta atom).
PARAMETERS
    atom      -  Atom to be added
    weight    -  Weight of the atom. If currently expected atom is
                  not weight assignment, this parameter is ignored
*/
void GroundRules::addAtom(Atom *atom, int weight)
{
    int isError;
    PPnode *pp = atom->pp;

        // Check for timely call:
    if(state>=3) inpError("addAtom");

        // Check whether we can write the atom and get more
        //  memory if not:
    if(input.cur >= input.limit) moveToNextPage(&input);

        // Insert ground atom into the AVL tree, write the atom
        //  to the theory. For this we also need to calculate the
        //  number corresponding to the set of argument of the
        //  program predicate using the
        //  method 'convertGroundAtomArguments'
    input.cur->atom=tree.insert(pp->symbolno,
        convertGroundAtomArguments(atom));
    input.cur->nextRule=0;

        // Indicate that the given predicate appears in the
        //  AVL tree for  ground atoms:
    pp->usage=PP_GROUNDED;

    if(pp->hornMeta == PP_HORN) // If 'pp' is a horn predicate
    {
        // Indicate that this the ground atom is a horn atom:
        input.cur->atom->isHorn=1;

        // Initialize the horn rule count to the maximum possible
        // number so that a horn atom won't become false when there
        // are no more rules containing the atom in the head. This
        // variable will be re-initialized at the end of propagation.
        input.cur->atom->hornRuleNum=UINTEGER_MAX;
    }


        // Check for critical error with negative weight:
    if(state==2 && weight <=0)
    {
        fprintf(stderr, "CRITICAL ERROR in the method "
            "GroundRules::addAtom\n"
            "Negative weight is given as a parameter\n\n");
        exit(1);
    }

    // CHECK FOR REPETITIONS:

        // If there is no repetition in the given part of the rule:
    if(!input.cur->atom->number)
    {
            // Save the current atom in the main stack:
            //  This is necessary step in the process of 
            // eliminating repetitions:
        mainStk.push(input.cur->atom);    


        if(state==2)   // If we are inside weight atom:
        {
                // Mark that the atom occurred in the current 
                // weight list and save its weight:
            input.cur->atom->number=weight;


            input.cur++; // Go to the next element to input weight.

            // Write the weight value (don't need to check for available
            //  memory because one check allowes to output two atoms.
            input.cur->atom=MAKE_DELIM(DELIM_WVAL);
            input.cur->num2=weight;
        }
        else // Mark that the atom occurred in the given part of the rule:
            input.cur->atom->number=1;

        input.cur++;    // Go to the next atom.
    }
    else
    {
            // If this is a weight atom, check whether the
            //  weight of the current atom is the same as the
            //  weight of the previous ocurrence.
            // If not, it means error.
        if(state==2 &&
            ((signed)input.cur->atom->number) != weight)
        { isError=1; controlFlag|=GR_FLG_ERROR; }
        else isError=0;

            // Print the message:
            //  Check whether we need to suppress the message:
        if(state!=0 && numOfRepetitionMsg < MAX_META_REPETITION_MSG)
        {
            // We still need to print the message:
            if(isError)
            {
                // Print the error message:
                fprintf(errorStream,
                   "ERROR: rule #%d(line %d): Detected "
                   "repetition in weight list where repeated "
                   "atom have different weights: ",
                    curGroundRuleNum, curLineNum);

                 printAtom(errorStream, input.cur->atom->symbolNum,
                     input.cur->atom->atomNum);

                 fprintf(errorStream, "=%d,=%d\n",
                     input.cur->atom->number, weight);
            }
            else
            {
                    // Print the warning message:
                fprintf(errorStream,
                 "WARNING: rule #%d(line %d): "
                 "Repetition in %s list: ", curGroundRuleNum,
                  curLineNum,
                  (state==2)?"weight":"cardinality");

                printAtom(errorStream, input.cur->atom->symbolNum,
                    input.cur->atom->atomNum);
                fprintf(errorStream, "\n");
            }
            numOfRepetitionMsg++;// Increase the message count.

            // If there was the last message to be printed and all
            //  messages will be suppressed, inform the user about this:
            if(numOfRepetitionMsg == MAX_META_REPETITION_MSG)
                fprintf(errorStream,
                    "TOO MANY MESSAGES about repetitions: "
                    "all subsequent messages will be suppressed\n");
        }
    }
}


/*
void GroundRules::startMeta(int min, int max, int isWeight)
DESCRIPTION
    This method should be called by the grounder to start
    output of cardinality or weight atom.
PARAMETERS
    min       -  Minimum limit of the meta-atom
    max       -  Maximum limit of the meta-atom
    isWeight  -  TRUE for weight atoms, FALSE for cardinality atoms
*/
void GroundRules::startMeta(int min, int max, int isWeight)
{
        // Check for timely call:
    if(state != 3) inpError("startMeta");

    state=1+isWeight; // Change the state

        // Check whether we have enough memory and get more if needed:
    if(input.cur >= input.limit) moveToNextPage(&input);

        // Save the beginning of the meta-atom:
    input.nMeta=input.cur-input.rule;

        //  Write meta-atom delimiter:
    input.cur->atom=MAKE_DELIM((isWeight)?DELIM_WEIGHT:DELIM_CARD);
    input.cur->nextRule=GR_NORMAL;
    input.cur++;

        // Write the meta-limits:
    input.cur->num1=min;
    input.cur->num2=max;
    input.cur++;
}


// This method should be called to finish output of simple atoms 
//  before meta-atoms, or to finish input of a meta-atom:
void GroundRules::endAtomList()
{
    AVLNode *atom;

    switch(state)
    {
    case 0: // If we are finishing a list of simple atoms:
        if(isHeadI) // If the list is from the head
        {
            // NOTE: clearanceStk contains all the atoms from the
            //       body of the rule outside meta-atom.
            // If any of these atoms is in the head, the rule can
            //  be eliminated (A,B->A|C is tautology):
            while(clearanceStk.getNumOfElements())
                if(clearanceStk.pop()->number)
                    input.rule->num2=GR_ELIMINATED;
        
            // Reset 'atom->number' to zero for all atoms in 
            //  mainStk (mainStk contains all the atoms from 
            //  the head outside the meta-atoms)
            while(mainStk.getNumOfElements())
                mainStk.pop()->number=0;
        }
        else // The list is from the body:
        {
            // Reset 'atom->number' to zero for all atoms in
            //  the mainStk and push them to clearanceStk.
            // (mainStk contains all the atoms from the head
            //  outside the meta-atoms).
            while(mainStk.getNumOfElements())
            {
                atom=mainStk.pop();
                atom->number=0;
                clearanceStk.push(atom);
            }
        }
        break;
    case 1:
    case 2:
        // If we are finishing cardinality or weight atom,
        //  just reset the 'atom->number' field.
        while(mainStk.getNumOfElements())
            mainStk.pop()->number=0;
        break;
    default: // Otherwise, the call is out of sequence:
        inpError("endAtomList");
    }

    state=3; // Switch to the new state
}


    // This method should be called when the body of the rule was completely
    // written to the theory and we need to start writing the head:
void GroundRules::switchToHead()
{
        // Check for timely call:
    if(isHeadI || state!=3)
        inpError("switchToHead");

        // Go to the new state:
    isHeadI=1; state=0;

        // Check for memory to write the head delimiter:
    if(input.cur >= input.limit) moveToNextPage(&input);

        // Write the head delimiter:
    input.cur->atom=MAKE_DELIM(DELIM_HEAD);
    input.cur->num2=curGroundRuleNum;
    input.cur++;
}


/*
int GroundRules::endRule()
DESCRIPTION
    This method should be called after writing all atoms for the given
    rule. It does propagation, and compaction of a rule (or sometimes
    of the whole theory).
RETURNS
    TRUE - everything is OK
    FALSE - the theory is inconsistent.
*/
int GroundRules::endRule()
{
    int size;
    PageDescriptor *page;

        // Check for timely call:
    if(state!=3 || !isHeadI) inpError("endRule");
    state=4; // Go to the next state

        // Put the endpage delimiter at the end of the rule:
    input.cur->atom=MAKE_DELIM(DELIM_PAGE);
    input.cur->num2=0;


        // Propagate the rule if necessary:
    if(controlFlag&GR_DO_PROPAGATE)
        if(!propagateRule(input.rule)) INDICATE_INCONSISTENCY;

        // Compact the current rule and establish the links:
    if(THEORY_IS_CONSISTENT && (controlFlag&GR_DO_ANY_COMPACTION))
        compactRules(GR_CM_LASTRULE);
    else // Just need to create links:
        linkRule(input.rule);

        // Find the size of the rule after this step;
    size=input.cur-input.rule;

        // The size cannot be negative:
    if(size <0)
    {
        fprintf(stderr, "CRITICAL ERROR in GroundRules::endRule\n\n");
        exit(1);
    }

        // Adjust the variable indicating when we need to do the next
        //  memory compaction of the whole theory:
    tillNextCompaction-=size;

        // If memory compaction is enabled and we need to do it now:
    if((controlFlag&GR_DO_M_COMPACTION) && tillNextCompaction <=0)
    {
        compactRules(GR_CM_ALL);  // Do the compaction

            // Calculate the total number of
            //  the occupied elements in all pages
        page=input.page; tillNextCompaction=0;
        do
        { tillNextCompaction+=(double)page->size;}
        while((page=page->next) != input.page);

            // If the next compaction is expected to be at the
            //  point that is less than GR_MIN_COMPACTION_SIZE,
            //  adjust 'tillNextCompaction' so that the next
            //  compaction point will be at GR_MIN_COMPACTION_SIZE.
        if(tillNextCompaction * GR_COMPACTION_MULTIPLIER <
            GR_MIN_COMPACTION_SIZE)
            tillNextCompaction = GR_MIN_COMPACTION_SIZE -
                tillNextCompaction;
        else
            tillNextCompaction *=(GR_COMPACTION_MULTIPLIER-1);
    }

        // Re-initialize the variable for the current rule:
    input.rule=input.cur;
        // Initialize the 'input.limit'.
    GET_LIMIT_FOR_PAGE(input.limit, input.page);

        // Return TRUE if the theory is consistent and FALSE otherwise:
    if(THEORY_IS_CONSISTENT) return 1; else return 0;
}


    // This method reports critical error when one of the input methods,
    //  indicated in the parameter, is called out-of-sequence:
void GroundRules::inpError(char *method)
{
    fprintf(stderr, "CRITICAL ERROR in the method GroundRules::%s\n"
        "Method call is out of sequence\n\n", method);
    exit(1);
}
/*          END OF INPUT INTERFACE                **************************/
