/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: lexan.h
 *
 *  CONTENTS:  LEXICAL ANALYZER : FRONT END, PSGRND.
 *             Supplies tokens to the parser generated by "bison"
 *              class 'LexicalAnalyzer'
 *
 *  DATE:     March 21, 2004
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <malloc.h>

    // Include the file with functions: 'open', 'close', 'seek'.
    //  This is "io.h" for Windows and "unistd.h" for Unix
#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

    // The constant 'O_BINARY' is defined in Windows but not in Unix
#ifndef O_BINARY
#define O_BINARY 0
#endif

    // Include the common header:
#include "common.h"

// *************************************************************
//  Token Numbers:
#define ASSIGN          '='
#define ASTERISK        '*'
#define COLON           ':'
#define COMMA           ','
#define GREATER         '>'
#define LEFTBRACE       '{'
#define LEFTBRACKET     '['
#define LEFTPARENTH     '('
#define LESS            '<'
#define MINUS           '-'
#define NEWLINE         '\n'
#define OR              '|'
#define PERIOD          '.'
#define PLUS            '+'
#define RIGHTBRACE      '}'
#define RIGHTBRACKET    ']'
#define RIGHTPARENTH    ')'
#define SEMICOLON       ';'
#define SLASH           '/'
#define UNDERSCORE      '_'
#define ENDMARKER       0


//*********************************************************
// CONSTANTS DEFINIG SIZES OF BUFFERS

// Size of the output buffer.
#define OLD_TOKEN_BUF_SIZE      4

// Constants defining file buffer and file loading.
#define FILE_BUFFER_SIZE        8192


// ***************************************************************
//  CONSTANTS, WHICH NEED CHANGE ONLY WHEN CHANGING TOKENS

// Maximal length of a keyword.
#define MAX_KW_LEN              6

// Constants defining the STATE TABLE:
	// Dimensions of the table:
#define STT_WIDTH		15
#define STT_HEIGHT		13
	// Number of final states(they are not included in the table).
#define STT_FIN_NUM		15


// Constants for printing tokens:
#define PR_NORMAL		0
#define PR_NUM			1
#define PR_ID			2
#define PR_NONE			(PR_NUM|PR_ID)


//*********************************************************
//*********************************************************
// CLASS DEFINITIONS
//*********************************************************
//*********************************************************

// The following structure defines an entry in the letter table
//  The entry corresponds to every character symbol.
struct LTEntry
{
	unsigned char colInd;
	int spec;
};


// Structure to hold information regarding input files:
struct InputFile
{
	int file;
	int isNew;
	char *name;
	InputFile *next;
};

// Structure to save several previous tokens. Necessary for
//  error reporting.
struct OldToken
{
	int token;
	int number;
	int lineNum;
	char *fileName;
};

// MAIN CLASS:
class LexicalAnalyzer
{
		// STATIC TABLES OF THE LEXICAL ANALYZER:
	static char keywords[][MAX_KW_LEN+1]; // Keywords
	static int kwTokens[];                // Token numbers for the keywords

		// Lexical analyzer state table:
	static char StateTable[STT_HEIGHT][STT_WIDTH];

	static LTEntry LetterTable[256];

	static int finalToToken[];      // Array to convert FinalState -> Token
	static char finalTokens[][6];   // Names of tokens corresponding to
	                                //   each final state:
		// Information regarding whether the lexical analyzer has to
		//  backtrack after the final state or not:
	static char FinalTable[];

	static unsigned char oneSymbolTokens[];	// Names of one sybol tokens
		// Token numbers for the one symbol tokens
	static int oneSymbolTokenNums[];

		// TypeOfIdentifier -> Token for the identifier.
		// Tokens corresponding to the identifier types. If the lexical
		//  analyzer encounters an identifier, it outputs the token:
		//  IDTypes[typeOfIdentifier].
	static int IDTypes[];

		// Buffer that contains the previously scanned tokens for the
		//  lexical analyzer:
	static OldToken oldTokens[];


		// OBJECT VARIABLES OF THE LEXICAL ANALYZER:
	SymbolTable &st;	// Reference to the symbol table to work with
	int lineCount;		// Line counter.
	int curOldToken;	// Index in the
	int errorCount;		// # of lexical errors.
		// Flag, indicating whether the method
		//  'LexicalAnalyzer::interruptInput have been called'
	int isInputInterrupt;


		// Variables to input to the scanner:
	unsigned char *inBuffer;// Pointer to the input file buffer
	int inBufFilled;        // # of occupied elements in the buffer
		// inBuffer[inBufStart] is the beginning of the current token
	int inBufStartToken;

		// Variables to deal with the list of the opened files:
		//  Head and tail pointers for a
		//    single-linked list of opened files:
	InputFile *firstFile, *lastFile;
		//  Pointer to the file that is currenty scanned:
	InputFile *curFile;

		// The following numbers indicate that integers between
		//    MinTerminal and MaxTerminal should be
		//     considered as Terminals
	int MinTerminal;
	int MaxTerminal;

		// Stream where lexical error messages will be written:
	FILE *errorStream;

		//	PRIVATE METHODS OF THE LEXICAL ANALYZER:
	void InitLetterTable(); // Initializes the letter table.
	int error(char *cause); // Print the lexical error message

		//Load the next portion of input to be analyzed by the scanner:
	int readMoreFromFile(int *current);

public: // PUBLIC CONSTANTS:
	const int EndMarker;	// End-of-input token.

	// PUBLIC METHODS OF THE LEXICAL ANALYZER:
		// MAIN METHOD OF THE CLASS. Returns token numbers for the yacc.
	int scan(int *number);

		// Open the file in the lexical analyzer for processing.
		//  If this method is called several times, several files
		// will be opened and then processed in the same order as
		//   function calls:
	int addFile(char *fileName);
		// Closes all the opened files (if any) and releases the
		//  memory that was used to store the file names:
	void closeAllFiles();
		// Stop all the input (usually in case of error)
	void interruptInput();

		// This method returns either the line number of one of
		//  the previous tokens or (-1). In addition, it may return
		//  the name of the file corresponding to that line number.
	int getLineNum(int offset, char **fileName=0);

		// Return the last token scanned by the scanner:
	int getLastToken(int *number);

		// Print the name of the current file
		// (last token came from this file).
	void reportCurrentFileName(FILE *file);

		// File name points to the memory allocated with the nodes of
		//  the structure 'InputFile'. When files are closed, the nodes
		//  and memory allocated for the file names is released. In this
		//  case all the references obtained via 'getLineNum' become
		//  invalid. To avoid errors, any function which deals with
		//  file names, should first call 'isFileNameValid'.
		// If 'isFileName' returns TRUE, then one can dereference
		//  the pointers; otherwise, not.
	int isFileNameValid() {return (firstFile!=0);}

			// Convert token back to the string representation.
	const char *tokenToString(int token, int number,
					int genericLevel=PR_NORMAL);

		// Retrieve parameters from the analyzer
	int getTerminalRange(int *min=0, int *max=0);
		// Return number of lexical errors:
	int getErrorCount(){return errorCount;}
		// Check whether the method 'interruptInput' has been called
	int isInputInterrupted() {return isInputInterrupt;}

		// Constructor and destructor.
	LexicalAnalyzer(SymbolTable &symbolTable);
	~LexicalAnalyzer();
};
