/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: rules.h
 *
 *  CONTENTS:  Main include file for FRONT END, PSGRND.
 *             Classes: 'Rules' - main class of the FRONT END
 *                      'QuantBuilder' - used by class 'Rules'
 *             Auxiliary classes: 'IntegerArray', 'DWVectorArray' needed
 *               for method 'quickSort' defined in "common.h"
 *
 *  DATE:     March 24, 2004
 *
 *  MODIFIED BY: Artur Mikitiuk, January 6, 2005
 * 
 ***************************************************************************/

#include "lexan.h"

// ***************************************************************************
/*
ATOM_ADD(ret, list, atom)
DESCRIPTION:
	Action that adds a temporary atom in a list of temporary atoms:
PARAMTERS:
	ret    - The variable to store the result
	list   - List of temporary atoms.
	atom   - The new atom to be added to the list.
*/
#define ATOM_ADD(ret, list, atom) \
{if(list) { list->last = list->last->next = atom; ret=list; } else ret=atom; }


// This class is used by the class 'Rules' to build quantifiers given data
//  restrictions and local variables.
// This class is implemented in the file "RULESAUX.CPP".
class QuantBuilder
{
		// Dummy nodes for the main lists of the class
		//  (all are circular double-linked).
	TempAtom conditionList;   // List of conditions
	TempAtom dataAtomList;    // List of data predicate quantifiers
	Quantifier quantVarList;  // List of quantified variable nodes

		// Stack needed to count only distinct variables:
	Stack<VarNode *> repeatStk;

		// External references to memory sources of the nodes.
	MemNodeN &tempSrc;
	MemNode<Expression> &expSrc;
	MemNode<Quantifier> &quantSrc;
	MemNode<Atom> &atomSrc;

		// Internal cost-related methods:
	void updateCosts(Expression **firstPtr, Expression **lastPtr);
	float conditionCost(TempAtom *atom, VarNode **minVarPtr);
	void moveDivModToEnd(Quantifier *quant);

		// Internal output and conversion methods:
	Quantifier *removeBestVariable();
	Quantifier *removeBestQuantifier(VarNode *var);
	void removeVariable(VarNode *var);
	void createCondition(TempAtom *tempAtom, 
		Expression **firstPtr, Expression **lastPtr);
public:
		// MAIN METHOD OF THE CLASS. Does all the work.
	Quantifier *build(TempAtom *conditions, Quantifier *quantList);

		// CONSTRUCTOR: Initialize the constants and the circular 
		// double-linked lists 'dataAtomList', 'conditionList',
		// and 'quantVarList'
	QuantBuilder(MemNodeN &tempSource, MemNode<Expression> &expSource,
		 MemNode<Quantifier> &quantSource, MemNode<Atom> &atomSource):
		 	repeatStk(16), tempSrc(tempSource),expSrc(expSource),
		 	quantSrc(quantSource), atomSrc(atomSource)
		{
			dataAtomList.next=dataAtomList.prev=&dataAtomList;
			conditionList.next=conditionList.prev=&conditionList;
			quantVarList.next=quantVarList.prev=&quantVarList;
		}
};


	// This method checks whether parser has a lookahead token
extern int isParserLookahead();


// Main class to hold all the structures:
//  NOTE: Below words "error messages" refer to syntax errors, semantic erros,
//        or execution control messages, but NOT TO critical error messages.
// This class is implemented in the three files:
//   1) "RULESACT.CPP": Contains all the aciton handlers (i.e. methods called
//                       from the file "grammar.y") and only them.
//                     Neither of these actions prints error messages.
//   2) "RULESAUX.CPP": Contains auxiliary methods (mostly called from the
//                        action handlers).
//                     Neither of them prints error messages.
//   3) "RULESERR.CPP": Printing error or debug messages.
class Rules
{// PRIVATE DATA FIELDS AND METHODS
	// [1] REFERENCES TO THE OTHER OBJECTS:
	SymbolTable &symbt;
	StringConstants &strConst;
	LexicalAnalyzer &lexan;

	// [2] NODE SOURCES:
	//     (a) Temporary nodes for list handlers and temporary atoms:
	MemNodeN tempSrc;
	//     (b) Value nodes to store arguments (and results) of program
	//         predicates and (weight functions). This is used only
	//         while parsing "dataSequence". Once the declaration
	//         sequence starts, arguments and results are converted
	//         into arrays and all such nodes are released.
	MemNode<Value> valSrc;
	//     (c) Memory for structures in "dataSequence" and
	//         "declarationSequence" after arguments are converted
	//         to arrays. Once used, these structures won't be
	//         released until the end of the program:
	MemAlloc permMSrc;
	//     (d) Nodes for the rest of the structures.
	MemNode<Expression> expSrc;
	MemNode<VarNode> varSrc;
	MemNode<Quantifier> quantSrc;
	MemNode<Atom> atomSrc;
	MemNode<MetaAtom> metaSrc;
	MemNode<RuleNode> ruleSrc;

	// [3] OBJECTS TO TRAVERSE ATOMS AND PROGRAM 
	//      VARIABLES AND TO BUILD QUANTIFIERS:
	TraverseAtoms goAtoms;
	TraverseVariables goVars;
	QuantBuilder quantBuilder;


	// [4] INSTANCE VARIABLES:
		// Structure that contains pointers for single-linked lists of
		//  data predicates+weight functions, program variables and
		//  program predicates.
	struct Links
	{
		WFnode *data;
		VarNode *var;
		PPnode *pred;
	} first, last;

	FILE *errorStream;        // Stream for syntax and semantic errors
	int syntaxErrorCount;     // # of syntax errors
	int semanticErrorCount;   // # of semantic errors
	int warningCount;         // # of warnings
	int dataRepetitionCount;  // # of messages regarding repetitions in data

		// This flag indicates whether we finished parsing data sequence
		//  (i.e. whether we currently are in the declaration sequence).
	int isDataFinished;

	int tempVarCount;    // Total Number of allocated temporary variables

		// Dummy node that contains no arguments. Used instead of
		// regular data type to recover from semantic errors:
	DPnode DPError;

		// Seqence of rules in the theory.
		//  Creating this list is the goal of this class.
	RuleNode *ruleSequence;

	// [5] SEMANTIC ERROR HANDLER:
	int semanticError(int code, ...);

	// [6] METHODS TO RELEASE MEMORY ALLOCATED FOR THE NODES: "RULESAUX.CPP"
	void deleteExpression(ExpressionHandler *eh, int leaveErrorHandler=0);
	void deleteExpression(Expression *exp);
	void deleteQuantifierList(Quantifier *list);
	void deleteTempAtomList(TempAtom *list);
	void deleteMetaList(MetaAtom *list);
	void deleteAtomList(Atom *list);
	void deleteRule(RuleNode *rule);

	// [7] AUXILIARY METHODS: "RULESAUX.CPP"
	  //  Check for repetitions in data predicates or weight functions:
	void checkRepetitionsInData(WFnode *wn, Value *oldList);
	  //  Create arguments for data atom, program atom, or weight assignment
	  //  Convert from the list of 'ExpressionHandler' nodes into
	  //    'Expression' nodes and check for semantic errors.
	Expression *createArguments(ExpressionHandler *argList, int symbolno);
	  //  Compile-time evaluator:
	int expEvaluate(int left, int right, int Operation, int *result);



	// [8] FINISH PARSING A RULE, BUILD QUANTIFIERS: "RULESAUX.CPP"
	//      All these methods are called from 'ruleNew' or some method
	//       that is called from the 'ruleNew'.
	//     (a) Auxiliary methods:
	        //  Allocate a new temporary variable
	VarNode *newTempVar(DPnode *type);

	        //  Given an expression and the expected type of the result,
	        //   convert it into assignment expression:
	Expression *createAssignment(Expression *exp, DPnode *type);

		// Given a list of quantifiers, this method determines the
		//  maximum number of possible combinations of values
		//   that the quantifiers can assign to the variables:
		//  (the number is based on the possible numbers of values
		//   of the variables: the CONDITIONS are
		//   NOT TAKEN INTO ACCOUNT)
	int maxPossibleValues(Quantifier *quant);


	//     (b) Primary steps (given in the sequence the methods called
	//            from the 'ruleNew'):
	//      (b1) Pre-processing:
	  //      Check for data atoms like: "dataType(variable)", where the
	  //       data type of the 'variable' is 'dataType'
	void reportUselessData(TempAtom *list);
	  // Check for horn atoms in meta-lists in horn rules:
	int checkHornRule(TempAtom *antecedent, TempAtom *consequent);
	  // Remove 'ATOM_TRUE' or 'ATOM_FALSE' atoms that could
	  //  appear as a recovery from errors or warnings.
	int removeTrueFalseAtoms(TempAtom **list, int isHead);

	//      (b2) Prepare building quantifiers:
	  // Check consistency of the local variable lists:
	void checkLocalVarLists(TempAtom *antecedent, TempAtom *consequent);
	  // Check for conflicts where some variable has to be global and
	  //  local at the same time:
	int determineLocalGlobal(TempAtom *body, TempAtom *head);
	  // Replace underscores with temporary variables:
	void processUnderscores(TempAtom *body, TempAtom *head);
	  // Given a body and a head, create a condition list from data atoms and
	  //  and predefined atoms. Create a list of global variables.
	Quantifier *createConditions(TempAtom **body, TempAtom **head,
	                             TempAtom **condList);
	  // Convert existential atoms in the body to cardinality atoms:
	void convertBodyExistentialAtoms(TempAtom **body);

	//      (b3) After building quantifiers:
	int convertTempAtomList(TempAtom *tempList, Atom **simpleAtomList,
	                        MetaAtom **cardList, MetaAtom **weightList);


	// [9] WARNINGS AT THE END OF PARSING: "RULESERR.CPP"
	void reportUnusedVariables();
	void reportUnusedProgramPredicates();

public:// PUBLIC METHODS:
	// ACTION HANDLERS: "RULESACT.CPP"
	//  [1] Parsing expressions:
	ExpressionHandler *expErrorID(int symbolno);
	ExpressionHandler *expNew(int num, int type);
	ExpressionHandler *expMerge(ExpressionHandler *HLeft,
	                            ExpressionHandler *HRight, int Operation);
	ExpressionHandler *weightCallNew(int symbolno,
	                                 ExpressionHandler *argList);

	// [2] Parsing 'dataSequence':
		// (a) Low level: arguments:
	Value *valueNew(int type, int v);
	VListHandler *valueAdd(VListHandler *vlh, Value *val);
	VListHandler *valueNewRange(ExpressionHandler *min,
	                            ExpressionHandler *max);
		// (b) High level:
	void dataNew(int symbolno, VListHandler *vlh, ExpressionHandler *eh);

	// [3] Parsing 'declarationSequence':
		// (a) Program Predicate Declaration:
	TempList *dataIdAdd(TempList *List, int symbolno);
	void programPredNew(int symbolno, TempList *params=0, int tail=-1);
		// (b) Program Variables:
	VarNode * varAdd(VarNode *List, int symbolno);
	void varDeclarationNew(int data, VarNode *vars);

	// [4] Parsing the rule sequence:
		// (a) Arguments:
	ExpressionHandler *argAdd(ExpressionHandler *argList,
	                          ExpressionHandler *newExp);
	ExpressionHandler *argUndMerge(ExpressionHandler *ALeft,
	                               ExpressionHandler *ARight);
		// (b) Local Variables:
	Quantifier *localVarAdd(Quantifier *varList, int newVar);
		// (c) Atoms from arguments:
	TempAtom *dataAtomNew(int symbolno, ExpressionHandler *argList);
	TempAtom *predAtomNew(ExpressionHandler *HLeft,
	                      ExpressionHandler *HRight, int RelOp);
	TempAtom *programAtomNew(
	                  int symbolno, ExpressionHandler *argList,
	                  Quantifier *localVariables, TempAtom *restrictions);
		// (d) Atoms from atoms:
	TempAtom *weightAssignmentFromSimple(
	                      TempAtom *simpleAtom, ExpressionHandler *eh,
	                      Quantifier *localVariables, TempAtom *RR);
	TempAtom *programAtomFromSimple(TempAtom *simpleAtom,
	                                Quantifier *localVariables,
	                                TempAtom *restrictions);
	TempAtom *metaAtomNew(TempAtom *cList, ExpressionHandler *ehMin,
	                      ExpressionHandler *ehMax, int atomType);
		// (e) RULES:
	void ruleNew(TempAtom *antecedent, TempAtom *consequent, int horn);

	// [5] Global scale:
		// This action checks for errors before parsing the rule
		//  sequence.  If errors are found, it prints corresponding
		//  message and returns the value to stop parser
		//  from parsing rules.
	int checkErrorsBeforeRules();
		// This action is called after data sequence and it
		//  rebuilds string constants, data predicates,
		//  and weight functions:
	void rebuild();
		// This action is called right before quitting the parser.
	int finishParse();


	// FUNCTIONS FOR DEBUGGING PURPOSES: "RULESERR.CPP":
	void test(); // Main debugging method.

		// Print the objects from the data sequence
		//   and declaration sequence:
	void printStringConstants();
	void printDataPredicates();
	void printWeights();
	void printVariables();
	void printProgramPredicates();

		// Methods for printing rules:
	void printTempRule(TempAtom *antecedent, TempAtom *consequent,int horn);
	void printTempAtom(TempAtom *atom);
	void printObjectWithArgs(char *name, int arity,
	                         Expression *arg, DPnode **Types);
	void printExpression(Expression *exp);
	void printQuantifierList(Quantifier *list);
	void printProgramAtomList(Atom *list, char *separator);
	void printMetaAtomList(MetaAtom *list, char *separator);
	void printRule(RuleNode *rule);
	void printTheory(); // PRINT THE WHOLE RULE SEQUENCE

	// ADDITIONAL INTERFACE METHODS:
		// To be used by the output classes:
	RuleNode *getRules() {return ruleSequence;}
	PPnode *getProgramPredicateList() {return first.pred;}

		// Used by yacc syntax error function to report syntax errors
		//  to the same place as semantic errors:
	FILE *getErrorStream() {return errorStream;}
		// Used by yacc syntax error to notify the class that the
		//  syntax error has occurred.
	void registerSyntaxError();
		// Return the number of warnings:
	int getNumberOfWarnings() {return warningCount;}


	// CONSTRUCTOR:
	Rules(SymbolTable &st, StringConstants &oc, LexicalAnalyzer &la):
		symbt(st),strConst(oc),lexan(la),
		tempSrc(sizeof(Handler), 20, (float)1.5, 100),
		valSrc(100, (float)2.,1000), permMSrc(10000, (float)2., 400000),
		expSrc(100, (float)1.2, 400), varSrc(20, (float)1.5, 64),
		quantSrc(50, (float)1.2, 200),atomSrc(100, (float)1.5, 200),
		metaSrc(20, (float)1.5, 100), ruleSrc(20, (float)1.5, 64),
		quantBuilder(tempSrc, expSrc, quantSrc, atomSrc)
	{
		memset(&first, 0, sizeof(first));
		memset(&last, 0, sizeof(last));

		ruleSequence=0;
		isDataFinished=0;
		errorStream=stderr;

			// Initialize temporary variable counter
		tempVarCount=0;

			// Initialize the dummy data predicate node:
		DPError.symbolno=-1;
		DPError.arity=0;
		DPError.valNum=0;
		DPError.arguments=0;

			// Initialize error counts:
		syntaxErrorCount = semanticErrorCount = 0;
		warningCount = dataRepetitionCount=0;
	}
};


// Defining and declaring class integerArray, necessary to sort values
//  for the data predicates with arity one.
class IntegerArray
{
	int *A;    // Pointer to the actual array in memory
	int N;     // Number of elements in the array
public:
		// Methods for the 'quickSort' (details in "common.h")
	void swap(int i, int j) {int temp=A[i]; A[i]=A[j]; A[j]=temp;}
	int isLess(int i, int j) {return(A[i] < A[j]);}
	int getNumOfElements() {return N;}
		// Associate the object with the physical array.
	void insert(int *array, int size) {A=array; N=size;}
		// Simple constructor
	IntegerArray() {A=0; N=0;}
};



// Defining and declaring class DWVectorArray, necessary to sort
//  values for the data predicates with the arity more than one
//  or for weight functions.
class DWVectorArray
{
	int *values;     // Array of arguments (array of vectors of 'vSize')
	                 //     SIZE: vSize*num.
	int *weights;    // Array of weight function results (can be NULL)
	                 //     SIZE: num.
	int vSize;       // Size of one vector == arity of data
	                 //   predicate or weight function
	int num;         // Number of vectors.
public:
		// Associate the object with arguments (and weight results)
		//  of some data predicate (weight function).
	void insert(int *vectorData, int vectorSize,
	            int vectorNum,   int *weightData=0)
	{values=vectorData; vSize=vectorSize;
	 num=vectorNum; weights=weightData;}

		// Methods for 'quickSort':
	void swap(int a, int b)
	{
		int i, temp;
			// Swapping the weights if we are dealing
			//  with a weight function.
		if(weights)
		{
			temp=weights[a];
			weights[a]=weights[b];
			weights[b]=temp;
		}

			// Swapping the vectors:
		for(a*=vSize, b*=vSize, i=0;i<vSize;i++)
		{
			temp=values[a+i];
			values[a+i]=values[b+i];
			values[b+i]=temp;
		}
	}

	int isLess(int a, int b)
	{
		int i;
			// Either find the first non-equal element
			//  or go to the last one.
		for(a*=vSize, b*=vSize, i=0;
			values[a+i] == values[b+i] && i<vSize-1; i++);

			// The relationship between vectors is determined by
			//  the first non-equal elements or the last ones.
		return(values[a+i] < values[b+i]);
	}

	int getNumOfElements() {return num;}

		// Simple constructor:
	DWVectorArray() {values=0; vSize=1; num=0;}
};
