/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: rulesaux.cpp
 *
 *  CONTENTS:  Part of the FRONT END, PSGRND. Consists of 3 parts:
 *             1) Auxiliary methods for the action handlers in 'Rules'.
 *             2) Auxiliary methods to release data structures in 'Rules'
 *             3) Implementation of the class 'QuantBuilder'
 *
 *  DATE:     March 24, 2004
 *
 *  MODIFIED BY: Artur Mikitiuk, January 6, 2005
 *               (function QuantBuilder::moveDivModToEnd)
 *
 ***************************************************************************/

#include <math.h>
#include "rules.h"


/****************************************************************************
*****************************************************************************
******    AUXILIARY METHODS FOR THE ACTION HANDLERS in the class 'Rules
*****************************************************************************
****************************************************************************/

/*
void Rules::reportUselessData(TempAtom *list)
DESCRIPTION
	This method checks for the data atoms like "dataType(varableOfDataType)"
	That were used in the old format for local variables, but don't make
	any difference in the new format. When such atom is found, it reports
	a warning.

PARAMETERS
	list  -  List of temporary atoms from the body or head of a rule
*/
void Rules::reportUselessData(TempAtom *list)
{
	for(;list;list=list->next)
	switch(list->type)
	{
	case ATOM_DATA:
		  // If this is a data atom with the data predicate of
		  //  arity one and its (the only) argument is a variable and
		  //  the data type of the variable is equal to the data
		  //  atom, report the warning
		if(list->data.dp->arity==1 && list->data.arg->token==EXPR_VAR &&
			list->data.arg->var->type == list->data.dp)
			semanticError(SW_USELESS_DATA_ATOM, list->data.dp,
				list->data.arg->var);
		break;

	case ATOM_PROG:
		  // If this is a program atom,
		  // check the list of data restrictions
		reportUselessData(list->prog.restrictions);
		break;
	case ATOM_CARD:
	case ATOM_WEIGHT:
		  // If this is a meta-atom, check the meta-list
		reportUselessData(list->meta.list);
		break;
	}
}


/*
int Rules::convertTempAtomList(TempAtom *tempList, Atom **simpleList,
                               MetaAtom **cardList, MetaAtom **weightList)
DESCRIPTION
	This method converts temporary body or head to permanent.
	Done by allocating permanent data structures instead of temporary,
	copying the corresponding fields and then deallocating the
	temporary structures.

PARAMETERS
	tempList    -  Temporary atom list: body or head
	simpleList  -  Pointer to the variable that should receive pointer
	                to the list of permanent program atoms
	cardList    -  Pointer to the variable that should receive pointer
	                to the list of permanent cardinality atoms
	weightList  -  The same as 'cardList', but for weight atoms.
RETURNS
	The estimated maximum number of ground atoms that can be generated
	in the given part of the rule.
*/

// The following macro copies data for a program atom from a temporary
//  structure to a permanent one
#define COPY_PROG_TEMP_TO_PERM(source, target) \
	{ target->pp=source->prog.pp; \
	target->arg=source->prog.arg; \
	target->quant=source->prog.quant; \
	target->weight=source->prog.weight; \
	target->next=0;}

//  The following copies data for a cardinality or weight atom from a
//   temporary structure to a permanent
#define COPY_META_TEMP_TO_PERM(source, target) \
    {target->min=source->meta.min; \
	target->max=source->meta.max; \
	target->next=0; }

int Rules::convertTempAtomList(TempAtom *tempList, Atom **simpleList,
                               MetaAtom **cardList, MetaAtom **weightList)
{
		// Count of the maximum possible ground
		// atoms that can be generated from the rule
	int maxGenAtoms=0;
	TempAtom *temp, *tempMeta;
	Atom **metaListPtr;       // Need for processing meta-lists
	Atom *curAtom=0;          // Last atom in the program atom list
	Atom *curAtomInList=0;    // Last atom in the meta-list
	MetaAtom *curCard=0, *curWeight=0;  // Last cardinality or weight atom

	if(!simpleList || !cardList || !weightList)
	{
		fprintf(stderr,"CRITICAL ERROR in method "
			"Rules::convertTempAtomList\nwrong parameters\n\n");
		exit(1);
	}

		// Initialize the lists
	*simpleList=0;
	*cardList=*weightList=0;

		// Go over all the atoms in the given list.
		//  Once atom processed, deallocate it
	for( ; tempList ; temp=tempList, tempList=tempList->next,
		tempSrc.Delete(temp))
	{
			 // If the current is a meta-atom, this will point
			 //   to a meta-list. Otherwise, it is zero.
		metaListPtr=0;
		switch(tempList->type)
		{
		case ATOM_PROG:
				// Allocate the new permanent atom node and
				//  add it to the end of the simple atom list
			if(*simpleList) curAtom=curAtom->next=atomSrc.New();
			else curAtom=*simpleList=atomSrc.New();

				// Copy fields from the temporary
				//  structure to the permanent
			COPY_PROG_TEMP_TO_PERM(tempList, curAtom);
				// Increase the maximum atom count:
			maxGenAtoms+=maxPossibleValues(curAtom->quant);
			break;

		case ATOM_CARD:
				// Allocate the new meta-atom node and add it
				//  to the end of the cardinality atom list
			if(*cardList) curCard=curCard->next=metaSrc.New();
			else curCard=*cardList=metaSrc.New();

			COPY_META_TEMP_TO_PERM(tempList, curCard) // Copy
				// Need to convert the meta-list (will be
				//  done after the 'switch'). After that, the
				// pointer to the meta-list will be written
				//  to the place pointed by 'metaListPtr'
			metaListPtr=&(curCard->list);
			break;

		case ATOM_WEIGHT:
				// Allocate the new meta-atom node and add it
				//  to the end of the weight atom list
			if(*weightList) curWeight=curWeight->next=metaSrc.New();
			else curWeight=*weightList=metaSrc.New();

			COPY_META_TEMP_TO_PERM(tempList, curWeight) // Copy
			metaListPtr=&(curWeight->list); // Process meta-list
			break;
		}

		if(metaListPtr)
		{
				// Initialize the pointer to the meta-list
				// (which is pointed by 'metaListPtr')
			*metaListPtr=0;

			// At this moment, the atom pointed by 'tempList'
			//  should be current meta-atom which meta-list
			//  needs to be converted

			    // Go over the atoms in the meta list.
			    //  Release them after processing:
			for(tempMeta=tempList->meta.list;tempMeta;temp=tempMeta,
				tempMeta=tempMeta->next,tempSrc.Delete(temp))
			{
				  // Allocate the new permanent atom node and
				  //  add this to the meta-list
				if(*metaListPtr) curAtomInList =
					curAtomInList->next = atomSrc.New();
				else curAtomInList=*metaListPtr=atomSrc.New();

					// Copy the fields
				COPY_PROG_TEMP_TO_PERM(tempMeta, curAtomInList);
					//  Increase the count:
				maxGenAtoms += maxPossibleValues(
					curAtomInList->quant);
			}
		}
	}

	return(maxGenAtoms);
}


/*
int Rules::checkHornRule(TempAtom *antecedent, TempAtom *consequent)
DESCRIPTION
	This method should be called for every horn rule. It checks
	whether restrictions on horn predicates are not violated.
	Horn predicate is the predicate that occurrs in a head of
	some horn rule. RESTRICTION: horn predicate CAN NOT occur
	inside cardinality or weight atom in any horn rule.
	See documentation for more details

PARAMETERS
	antecedent   -  The body of the horn rule
	consequent   -  The head of the horn rule
RETURNS
	TRUE  - Everything is OK
	FALSE - Restriction is violated
*/

int Rules::checkHornRule(TempAtom *antecedent, TempAtom *consequent)
{
	TempAtom *atom, *clAtom;
	PPnode *pp;

		// Make sure that the horn rule has exactly one atom in the
		//  head; otherwise, the rule is already invalid
	if(!consequent || consequent->type!=ATOM_PROG || consequent->next)
	{
		semanticError(SE_BAD_HORN_RULE, antecedent, consequent);
		return 0;
	}

		// Analyze the program predicate of that head atom
	pp=consequent->prog.pp;

		// If the given predicate has already caused an error,
		// then quit and don't print any messages
	if(pp->hornMeta==PP_WASTED) return 0;

		// If program predicate from the head appeared in a meta atom
		//  of a horn rule, then ERROR. Give an error message
	if(pp->hornMeta==PP_META)
	{
		semanticError(SE_HORN_ATOM_CONFLICT, pp);
		pp->hornMeta=PP_WASTED;
		return 0;
	}
		// Mark it as a horn predicate
	pp->hornMeta=PP_HORN;

		// Check all meta atoms in this horn rule:
	for(atom=antecedent;atom;atom=atom->next)
		if(atom->type==ATOM_CARD || atom->type==ATOM_WEIGHT)
			for(clAtom=atom->meta.list;clAtom;clAtom=clAtom->next)
			{
				if(clAtom->type!=ATOM_PROG) continue;

				pp=clAtom->prog.pp;
					// The predicate has already
					//  caused an error => QUIT
				if(pp->hornMeta==PP_WASTED) return 0;

					// If the predicate has already appeared
					// in the head of some horn rule,
					// then ERROR. Give a message.
				if(pp->hornMeta==PP_HORN)
				{
					semanticError(SE_HORN_ATOM_CONFLICT,pp);
					pp->hornMeta=PP_WASTED;
					return 0;
				}

					// Mark it as a meta-atom predicate
				pp->hornMeta=PP_META;
			}

		// No errors found
	return 1;
}


/*
int Rules::maxPossibleValues(Quantifier *quant)
DESCRIPTION
	This method gives the maximum possible number of atoms that will
	result from instantiating the given list of existential quantifiers.

PARAMETERS
	quant   -  Possibly empty list of existential quantifiers.
*/
int Rules::maxPossibleValues(Quantifier *quant)
{
	int count=1;

	for(;quant;quant=quant->next)
		if(quant->token==QUANT_VAR) count*=quant->var->type->valNum;
		else if(quant->token==QUANT_DP) count*=quant->dp->valNum;

	return(count);
}


/*
void Rules::convertBodyExistentialAtoms(TempAtom **body)
DESCRIPTION
	This method finds existential atoms (<atom>) (i.e. atoms that
	have local variables) in the body of a rule and replaces them
	with cardinality atoms: 1 {<atom>} .

PARAMETERS
	body  -  Pointer to the list of temporary atoms that represent
	         the body of a rule.
*/
void Rules::convertBodyExistentialAtoms(TempAtom **body)
{
	TempAtom dummyAtom, *cur,*prev, *atom;
	Expression *exp;
	dummyAtom.next=*body; // Append the body to the dummy atom

		// Do the following for every program atom in
		// the body that has local quantifiers
	for(prev=&dummyAtom;(cur=prev->next);prev=prev->next)
		if(cur->type==ATOM_PROG && cur->prog.quant)
		{
				// Allocate the new temporary atom node and
				// initialize it as cardinality atom:
				//   1 { <cur> } INTEGER_MAX.
			atom=(TempAtom *)tempSrc.New();
			atom->type=ATOM_CARD;
			atom->meta.list=cur;
				// Allocate new expression and
				//  initialize 'min' limit:
			exp=expSrc.New();
			exp->next=0;
			exp->token=EXPR_NUM;
			exp->num=1;
			atom->meta.min=exp;
				// Allocate new expression and
				//  initialize 'max' limit:
			exp=expSrc.New();
			exp->next=0;
			exp->token=EXPR_NUM;
			exp->num=INTEGER_MAX;
			atom->meta.max=exp;

				// Put 'atom' in place of 'cur' in the list
			prev->next=atom;
			atom->next=cur->next;
				// Set the 'next' link of 'cur' to NULL,
				//  since 'cur' is now in the meta-list
			cur->next=0;
		}

	*body=dummyAtom.next; // Save the resulting list of atoms in the body
}


/*
Quantifier *Rules::createConditions(TempAtom **bodyPtr, TempAtom **headPtr,
                                    TempAtom **condList)
DESCRIPTION
	This method does the following things:
	1) Create a list of global variables
	2) Remove data and predefined atoms from the body and the head and
	    add them to the condition list (NOTE: atoms from the head
	    have to be negated)
	3) If some program predicate has a tail,
	    create a data atom from this tail
	4) Convert expressions in the arguments of program predicates into
	   assignments
	5) If some atom in the body has a data restriction that doesn't have
	    quantified variables, move this restriction to the condition list

PARAMETERS
	bodyPtr   -  Pointer to the variable that points to the body of a rule
	headPtr   -  Pointer to the variable that points to the head of a rule
	condList  -  Pointer to the variable that receives pointer to the
	              newly created condition list
RETURNS
	The list of global quantifiers
*/
Quantifier *Rules::createConditions(TempAtom **bodyPtr, TempAtom **headPtr,
                                    TempAtom **condList)
{
    TempAtom *cur, *prev, *atom;
    TempAtom dummyAtom;
    Quantifier *quant, *globalVarList;
    VarNode *var;
    Expression *exp, *dupl, *exp1;
    Expression dummyExp;
    PPnode *pp;
    int i, localFlg;

        // Constructing a list of all global variables:
    for(globalVarList=0,goVars.start("AA", *bodyPtr, *headPtr) ;
        (var=goVars.next()) ; )
    {
        if(var->valIndex==VAR_GLOBAL)
        {
                // This is done to eliminate repetitions
            var->valIndex=VAR_FREE;

                // Create a quantified variable node
            quant=quantSrc.New();
            quant->prev=quant->next=0;
            quant->conditions=0;
            quant->token=QUANT_VAR;
            quant->var=var;

                // Append the node to the list.
            quant->next=globalVarList;
            globalVarList=quant;
        }
    }

        // Initialize the list of conditions
    *condList=0;

        // Add condition atoms from the body to the list of conditions
    dummyAtom.next=*bodyPtr;  // List of atoms in the body will be modified
    for(prev=&dummyAtom;(cur=prev->next);)
        if(cur->type==ATOM_DATA || cur->type==ATOM_DATANEG ||
            cur->type==ATOM_CONDITION)
        {
                // Remove the atom from the list
            prev->next=cur->next;

            cur->next=*condList;
            *condList=cur;
        }
        else    // Go to the next atom
            prev=cur;

        // After the loop is done, save the remaining body.
    *bodyPtr=dummyAtom.next;


        // Add condition atom from the head to
        // the list of conditions negating them:
    dummyAtom.next=*headPtr;  // List of atoms in the head will be modified
    for(prev=&dummyAtom;(cur=prev->next);)
    {
            // If current atom is a condition atom, negate it.
        switch(cur->type)
        {
        case ATOM_DATA:
            cur->type=ATOM_DATANEG;
            break;
        case ATOM_DATANEG:
            cur->type=ATOM_DATA;
            break;
        case ATOM_CONDITION:
                // Negate the relational operation
            exp=cur->cond.last;
            switch(exp->token)
            {
            case EXPR_GREATER:   exp->token=EXPR_NOTGREATER; break;
            case EXPR_NOTGREATER:exp->token=EXPR_GREATER;    break;
            case EXPR_LESS:      exp->token=EXPR_NOTLESS;    break;
            case EXPR_NOTLESS:   exp->token=EXPR_LESS;       break;
            case EXPR_EQUAL:     exp->token=EXPR_NOTEQUAL;   break;
            case EXPR_NOTEQUAL:  exp->token=EXPR_EQUAL;      break;
            default:
                fprintf(stderr, "CRITICAL ERROR in method "
                        "Rules::createConditions \n"
                        "Last operand of a predefined atom is not relational"
                        "\n\n");
                    exit(1);
            }
            break;
        default:
                // The current atom is not condition atom. Go to the next atom:
            prev=cur;
            continue;
        }

        // We can be here only if the current atom was condition atom
        //  and has been negated.

            // Remove the atom from the list
        prev->next=cur->next;

            // Put the atom in the condition list:
        cur->next=*condList;
        *condList=cur;
    }
    *headPtr=dummyAtom.next;    // Save the remaining head


        // Convert all expressions to assignments.
        // Convert predicate tails to data atoms.
    for(goAtoms.start(*bodyPtr, *headPtr); (atom=goAtoms.next());)
        if((pp=atom->prog.pp)->tail)
        {
            // THE PREDICATE HAS A TAIL. WE NEED TO CONSTRUCT A DATA
            //  ATOM OUT OF THE TAIL:
                // Initialize the variables to duplicate argument list.
            dummyExp.next=0; dupl=&dummyExp;

                // Duplicate the argument list and convert expressions
                //  into assignments
            for(i=0,exp=atom->prog.arg ; exp ; exp=exp->next,i++)
            {
                    // Append new node to the duplicate list
                dupl=dupl->next=expSrc.New();

                if(exp->token!=EXPR_EXPR)
                    // Simply duplicate nodes that are other than expressions
                    memcpy(dupl, exp, sizeof(Expression));
                else
                {

                        // Node in the duplicate list should contain temporary
                        //  variable only.
                    dupl->token=EXPR_VAR;

                        // Convert expression into assignment (in the original
                        // argument list). Make reference to the resulting
                        // temporary variable in the duplicate list.
                    dupl->var=(createAssignment(exp->exp, pp->params[i]))->var;
                }

                        // If the argument is an index, it needs to be changed
                        //  to the actual number.
                if(exp->token==EXPR_INDEX)
                {
                    exp->token=EXPR_NUM;
                    exp->num=pp->params[i]->arguments[exp->num];
                }

                dupl->next=0;
            }

                // Create data atom 'cur' out of the predicate tail with
                //  parameters from parameters almost like in program predicate,
                //  but expressions converted into assignments.
            cur=(TempAtom *)tempSrc.New();
            cur->next=0;
            cur->type=ATOM_DATA;
            cur->data.arg=atom->prog.arg;
            cur->data.dp=pp->tail;

                // Put this data atom 'cur' in the list of data
                //  restrictions for the current predicate:
            cur->next=atom->prog.restrictions;
            atom->prog.restrictions=cur;

                // Assign duplicate argument list to the program cur:
            atom->prog.arg=dummyExp.next;
        }
        else
        {
            if(atom->type!=ATOM_PROG)
            {
                fprintf(stderr, "CRITICAL ERROR: logical error in "
                    "Rules::createConditions\n\n");
                exit(1);
            }

            for(i=0,exp=atom->prog.arg;exp;exp=exp->next,i++)
                if(exp->token==EXPR_EXPR)
                {
                        // Allocate new temporary cur node, convert expression
                        //  into assignment and put this assignment into
                        //  the condition node:
                    cur=(TempAtom *)tempSrc.New();
                    cur->next=0;

                    cur->type=ATOM_CONDITION;
                    cur->cond.first=exp->exp;
                    cur->cond.last=createAssignment(exp->exp, pp->params[i]);

                        //Put the temporary variable in the place of expression
                    exp->token=EXPR_VAR;
                    exp->var=cur->cond.last->var;

                        // Put the resulting cur in the list of data
                        //  restrictions for the current predicate:
                    cur->next=atom->prog.restrictions;
                    atom->prog.restrictions=cur;
                }
        }

        // For every atom in the body outside meta-atoms:
        //  if any of its data restrictions doesn't contain quantified
        //   variables, move it to the 'condList'
    for(atom=*bodyPtr;atom;atom=atom->next)
        if(atom->type==ATOM_PROG)
        {
                // Scan the list of data restrictions:
            dummyAtom.next=atom->prog.restrictions;

            for(prev=&dummyAtom;(cur=prev->next);)
            {
                    // Initialize 'exp' with corresponding expression
                    //  depending on type of atom.
                if(cur->type==ATOM_CONDITION) exp=cur->cond.first;
                else
                    if(cur->type==ATOM_DATA || cur->type==ATOM_DATANEG)
                        exp=cur->data.arg;
                    else
                    {
                        fprintf(stderr, "CRITICAL ERROR in "
                                "Rules::createConditions\n"
                                "inappropriate atom in data restrictions\n\n");
                        exit(1);
                    }

                    // Check for local variables in the expression:
                for(localFlg=0;exp&&!localFlg;exp=exp->next)
                    if(exp->token==EXPR_EXPR)
                    {
                        for(exp1=exp->exp;exp1;exp1=exp1->next)
                            if(exp1->token==EXPR_VAR &&
                               exp1->var->valIndex==VAR_LOCAL)
                            { localFlg=1; break; }
                    }
                    else
                        if(exp->token==EXPR_VAR &&
                           exp->var->valIndex==VAR_LOCAL) localFlg=1;


                if(!localFlg)
                {      // If local variable is not found,
                       // Remove the current atom from the restrictions list.
                    prev->next=cur->next;

                        // Put it into condition list.
                    cur->next=*condList;
                    *condList=cur;
                }
                else    // Move to the next atom.
                    prev=cur;
            }

            atom->prog.restrictions=dummyAtom.next;
        }

        // Return the list of global variables
    return(globalVarList);
}


/*
Expression *Rules::createAssignment(Expression *exp, DPnode *type)
DESCRIPTION:
	This method converts arithmetic expression to the assignment to a new
	temporary variable by appending assignment node to the end of the list
	for the expression. For proper operation, the expression should
	evaluate to exactly one number.

PARAMETERS:
	exp     -  Source expression.
	type    -  Type of the resulting temporary variable. The type of any
	            variable inside the given expression should be the same
	            as this parameter.

RETURNS: The pointer to the assignment node that has been appended.
          'var' field of this node points to the new temporary variable.
*/
Expression *Rules::createAssignment(Expression *exp, DPnode *type)
{
		// Find the last element of the expression:
	if(!exp) return 0;
	while(exp->next) exp=exp->next;

		// Append new node to the end and make it to be assignment
		//  to the new temporary variable.
	exp=exp->next=expSrc.New();
	exp->next=0;
	exp->token=EXPR_ASSIGN;
	exp->var=newTempVar(type);

		// Return the expression node that contains the
		//  newly created temporary variable
	return(exp);
}


/*
void Rules::checkLocalVarLists(TempAtom *body, TempAtom *head)
DESCRIPTION
	This method is used to check repetitions in the local variable list and
	enforce condition that any variable that appears in the local variable
	list should also appear in the program predicate or data restrictions
	of the existential atom

PARAMETERS
	body    - The list of temporary atoms from the body
	head    - The list of temporary atoms from the head
*/
void Rules::checkLocalVarLists(TempAtom *body, TempAtom *head)
{
	TempAtom *atom;
	VarNode *var;
	Quantifier dummyQuant;
	Quantifier *quant, *tmp;
	int errCode;

	    // Do the loop for every atom in the body or
	    //  head that has a local variable list:
	for(goAtoms.start(body, head) ; (atom=goAtoms.next()) ; )
		if(atom->type == ATOM_PROG && atom->prog.quant)
		{
				// Mark all variables in argument list and
				//  data restrictions as VAR_FIRST.
			for(goVars.start("EA",atom->prog.arg,
				   atom->prog.restrictions);(var=goVars.next());)
				var->valIndex=VAR_FIRST;

			    // "Dummy Expression" makes it easier to delete an
			    //  element from the list of expressions: can do it
			    //  the same way when the element is in the
			    //  beginning or in the middle of the list
			dummyQuant.next=atom->prog.quant;

			    // For every variable inside local variable list:
			    //  If VAR_FREE: warning - unused local variable;
			    //    remove it from the list.
			    //  If VAR_FIRST: mark as VAR_SECOND
			    //  If VAR_SECOND: warning - repetition in local
			    //    variable list; remove it from the list.
			for(quant=&dummyQuant;quant->next;)
			{
				var=quant->next->var;

				switch(var->valIndex)
				{
				case VAR_FREE:
					errCode=SW_LOCALVAR_UNUSED;
					break;
				case VAR_FIRST:
					var->valIndex=VAR_SECOND;
					quant=quant->next;
					continue; //loop
				case VAR_SECOND:
					errCode=SW_LOCALVAR_REPEAT;
					break;
				default:
				     fprintf(stderr, "CRITICAL ERROR in method"
				             " Rules::checkLocalVarLists\n\n");
				     exit(1);
				}

				    // Report an error and remove from the list
				semanticError(errCode, atom, var);
				tmp=quant->next;
				quant->next=tmp->next;
				quantSrc.Delete(tmp);
			}

				// Save new, possibly empty, local variable list
			atom->prog.quant=dummyQuant.next;

				// Restore all variable types to be VAR_FREE.
			for(goVars.start("EA",atom->prog.arg,
				  atom->prog.restrictions);(var=goVars.next());)
				var->valIndex=VAR_FREE;
		}
}


/*
int Rules::determineLocalGlobal(TempAtom *body, TempAtom *head)
DESCRIPTION
	This method checks for errors when some variable should be global
	and local at the same time
PARAMETERS
	body    -  Body of a rule to check
	head    -  Head of a rule to check
RETURNS
	TRUE    -  Everything is OK
	FALSE   -  Error
*/
int Rules::determineLocalGlobal(TempAtom *body, TempAtom *head)
{
	TempAtom *atom;
	VarNode *var;
	Quantifier *quant;
	int correctFlag=1;

		// For every program atom in the body or head including the
		//  atoms inside meta-atoms:
	for(goAtoms.start(body, head, 1);(atom=goAtoms.next());)
		if(atom->type==ATOM_PROG)
		{	// If atom is the program atom:

			    // For every local variable: VAR_FREE -> VAR_FIRST
			for(quant=atom->prog.quant;quant;quant=quant->next)
				if(quant->var->valIndex==VAR_FREE)
					quant->var->valIndex=VAR_FIRST;

			    // For the rest of variables: VAR_FREE -> VAR_GLOBAL
			for(goVars.start("EA",atom->prog.arg,
				  atom->prog.restrictions);(var=goVars.next());)
				if(var->valIndex==VAR_FREE)
					var->valIndex=VAR_GLOBAL;

			    // For every local variable: VAR_FIRST -> VAR_FREE
			for(quant=atom->prog.quant;quant;quant=quant->next)
				if(quant->var->valIndex==VAR_FIRST)
					quant->var->valIndex=VAR_FREE;

			  // RESULT OF THESE STEPS:
			  //  Variables in the program atom that don't occur in
			  //  the local variable list are marked as VAR_GLOBAL
		}
		else
		{	// For the rest of atoms:
			  //  MARK ALL THE VARIABLES AS VAR_GLOBAL
			switch(atom->type)
			{
			case ATOM_CARD:
			case ATOM_WEIGHT:
				goVars.start("EE", atom->meta.min,
					atom->meta.max);
				break;
			case ATOM_CONDITION:
				goVars.start("E",atom->cond.first);
				break;
			case ATOM_DATA:
				goVars.start("E",atom->data.arg);
				break;
			default:
				fprintf(stderr, "CRITICAL ERROR in the method "
					"Rules::checkGlobalLocalOverlap\n\n");
				exit(1);
			}

			while((var=goVars.next())) var->valIndex=VAR_GLOBAL;
		}

	    // Go over the same atoms again and check the local variable lists.
	    // If some of these variables is marked as VAR_GLOBAL, then
	    //  there is a conflict.
	    // If not, then mark the variable as VAR_LOCAL.
	for(goAtoms.start(body, head);(atom=goAtoms.next());)
		if(atom->type==ATOM_PROG)
			for(quant=atom->prog.quant;quant;quant=quant->next)
			{
				var=quant->var;

				switch(var->valIndex)
				{
				case VAR_GLOBAL:
					semanticError(SE_GLOBAL_LOCAL_MIX,
						atom, var);
					    // Indicate that the method
					    //  should return FALSE
					correctFlag=0;
					break;
				case VAR_FREE:
					var->valIndex=VAR_LOCAL;
				case VAR_LOCAL:
					break;
				default:
				 fprintf(stderr, "CRITICAL ERROR in the method "
				         "Rules::checkGlobalLocalOverlap\n\n");
				 exit(1);
				}
			}

	return(correctFlag);
}


/*
void processUnderscores(TempAtom *body, TempAtom *head)
DESCRIPTION
	The function replaces all underscores in all existential atoms with
	temporary variables of the proper type and places these variables
	into local variable list.
PARAMETERS:
	body    - The list of temporary atoms from the body
	head    - The list of temporary atoms from the head
*/
void Rules::processUnderscores(TempAtom *body, TempAtom *head)
{
      TempAtom *atom;
      Expression *exp;
      Quantifier *quant;
      VarNode *var;
      int argCount;

        // For every underscore in the argument list of every existential atom:
      for(goAtoms.start(body, head);(atom=goAtoms.next());)
          if(atom->type==ATOM_PROG)
              for(argCount=0,exp=atom->prog.arg;exp;exp=exp->next, argCount++)
                  if(exp->token==EXPR_UNDERSCORE)
                  {
                              // Create the new temporary local variable
                        var=newTempVar(atom->prog.pp->params[argCount]);
                        var->valIndex=VAR_LOCAL;
                        exp->token=EXPR_VAR;
                        exp->var=var;

                              // Create quantifier node with this variable
                        quant=quantSrc.New();
                        quant->token=QUANT_VAR;
                        quant->conditions=0;
                        quant->prev=quant->next=0;
                        quant->var=var;

                              // Append this node to the Local Variable List
                        quant->next=atom->prog.quant;
                        atom->prog.quant=quant;
                  }
}


/*
VarNode *Rules::newTempVar(DPnode *type)
DESCRIPTION
	The method allocates new temporary variable of a type given as a
	parameter. The variable will be automatically marked as VAR_DEPENDENT.
*/
VarNode *Rules::newTempVar(DPnode *type)
{
	VarNode *var;

	tempVarCount++;	// Increase temporary variable counter
	var=varSrc.New();

	var->symbolno=-tempVarCount;
	var->type=type;
	var->valIndex=VAR_DEPENDENT;

	return(var);
}


/*
int Rules::removeTrueFalseAtoms(TempAtom **list, int isHead)
DESCRIPTION
	Compile-time propagation. It looks for atoms with types
	ATOM_TRUE or ATOM_FALSE and either removes the atom or removes the rule.
	Mainly needed to recover from errors or warnings.
	(ATOM_TRUE, isHead==0) || (ATOM_FALSE, isHead=1) => Remove the ATOM.
	(ATOM_TRUE, isHead==1) || (ATOM_FALSE, isHead=0) => Remove the RULE.

PARAMETERS
	list      -  List of atoms from the body or the head of the rule
	isHead    -  TRUE if the list is from head, FALSE otherwise.

RETURNS
	TRUE    - Keep the rule
	FALSE   - Delete the rule.
*/
#define DELETE_ATOM {prev->next=cur->next; tempSrc.Delete(cur); cur=prev;}
#define DELETE_RULE {deleteTempAtomList(start.next); *list=0; return 0;}

int Rules::removeTrueFalseAtoms(TempAtom **list, int isHead)
{
    TempAtom start;       // Dummy node.
    TempAtom *cur, *prev; // Pointers

    start.next=*list;     // Append the list to the dummy node.

    if(isHead) isHead=1;  // Make sure isHead is either 0 or 1.

    for(prev=&start,cur=*list;cur;prev=cur,cur=cur->next)
    {
            // If True atom is in the body or false atom
            //  is in the head, REMOVE THE ATOM
        if(cur->type == isHead*ATOM_FALSE+(1-isHead)*ATOM_TRUE)
            DELETE_ATOM
        else
                // If a False atom in the body or a true
                //  atom in the head, REMOVE THE RULE
            if(cur->type == isHead*ATOM_TRUE+(1-isHead)*ATOM_FALSE)
                DELETE_RULE
            else
            {
                // If the current atom might have data restrictions,
                //  remove True/False atoms from data restrictions as if
                //   we were removing them from a body of some rule, i.e.
                //  the function will return FALSE if at least one atom
                //  there is false.
                if(cur->type==ATOM_PROG)
                    if(!removeTrueFalseAtoms(&(cur->prog.restrictions), 0))
                        // Now, the existential/weight atom is now FALSE atom.
                        //  We need to remove it accordingly.
                        if(isHead) DELETE_ATOM else DELETE_RULE

                    // If the current atom is metaatom, remove True/False atoms
                    //  From metalist as if it was a head of a rule.
                    //  We SHOULD NOT encounter a TRUE atom in metalist.
                if(cur->type==ATOM_CARD||cur->type==ATOM_WEIGHT)
                    if(!removeTrueFalseAtoms(&(cur->meta.list),1))
                    {
                        fprintf(stderr, "CRTICICAL ERROR: TRUE atom cannot"
                            " appear inside a cardinality atom\n");
                        exit(1);
                    }

            }
    }

        // Save the resulting list that is appended to the dummy node
    *list=start.next;

        // Find the value for the 'last' link in the first node of the list
        //  (so that the resulting list would be just like it was after
        //    parsing):
    if(*list)
    {
        for(cur=*list;cur->next;cur=cur->next);
        (*list)->last=cur;
    }

    return 1;
}


/*
Expression *Rules::createArguments(ExpressionHandler *argList, int symbolno)
DESCRIPTION
	Creates argument lists for data atom, program, or weight function.
	The method converts the argument list to the compatible format and
	checks for errors

PARAMETERS
	argList   -  List of arguments as it was reduced by the actions
	symbolno  -  Symbol number of the object for which the argument
	              list has to be constructed. Can be program predicate,
	              data predicate or a weight function.
RETURNS
	Created argument list, or NULL in case of error
	NOTE: If the program predicate doesn't have any arguments (arity==0),
	the method returns NULL even in the case of success.
*/
Expression *Rules::createArguments(ExpressionHandler *argList, int symbolno)
{
	PPnode *pp=0; WFnode *wn=0;
	DPnode *argType;
	int errFlag=0;
	int i, num;
	int arity=0;
	ExpressionHandler *temp;
	Expression *exp, *expList;

		// Depending on what object we are creating arguments for
		//  (program predicate, data predicate, or weight function),
		//  need to find the object's arity and initialize corresponding
		//  variable with the pointer to the corresponding node.
	switch(symbt.getType(symbolno))
	{
	case TYPE_WEIGHT_FUN:
	case TYPE_DATA_PRED:
		wn=(WFnode *)((symbt.getVal(symbolno))->ptr);
		if(wn) arity=wn->arity; else errFlag=1;
		break;

	case TYPE_PROGRAM_PRED:
		pp=(PPnode *)((symbt.getVal(symbolno))->ptr);
		if(pp) arity=pp->arity; else errFlag=1;
		break;

	default:
	  fprintf(stderr,"CRITICAL ERROR in the method Rules::createArguments\n"
	      "expected data predicate, program predicate or weight function\n"
	      "given identifier with the type %d\n", symbt.getType(symbolno));
	   exit(3);
	}

		// Make sure there are no weight functions in the arguments and
		// the number of arguments is equal to the arity of the object
	if(!errFlag)
	{
		for(temp=argList,i=0;temp;temp=temp->next,i++)
			if(temp->type==ET_ERROR) errFlag=1;
			else
				if(temp->type&ET_WEIGHT)
				{
					semanticError(SE_ARGUMENT_WEIGHT,
						symbolno, temp);
					errFlag=1;
				}

		if(i!=arity)
		{
			semanticError(SE_PARAM_NUM, symbolno, argList);
			errFlag=1;
		}
	}

		// Checking arguments that are numbers. Need to make sure
		//  that these numbers won't be confused with string
		//  constants during future processing.
	if(!errFlag)
	{
		for(temp=argList;temp;temp=temp->next)
			if(temp->type==ET_NUM &&
				!strConst.isNumber(temp->first->num))
			{
				semanticError(SW_NUM_OUT_RANGE,
				              temp->first->num);
				errFlag=1;
			}
	}

	  // If we constructing arguments for the program predicate, then for
	  // every argument do the following:
	  //  1) If argument is a number, check whether the number is inside the
	  //      range of the corresponding type. Convert the number to index.
	  //  2) If argument contains variables, make sure that the variables of
	  //      the right type.
	if(!errFlag && pp)
	{
		for(temp=argList, i=pp->arity-1 ; temp ; temp=temp->next, i--)
		{
				// Find the type of the current argument
			argType=pp->params[i];

			switch(temp->type)
			{
			case ET_STRING:
			case ET_NUM:
				// If the argument is a legal number
				// (won't be confused with a string constant)
				// or a constant, convert this argument to
				//  the index in the 'pp->argument' array:
				exp=temp->first;
				exp->token=EXPR_INDEX;
				num=exp->num;

					// The index is found by
					//  performing binary search:
				if((exp->num=EvalStack<int>::scalarSearch(num,
				      argType->arguments, argType->valNum)) < 0)
				{
					semanticError(SW_PP_OUT_RANGE,pp,num,i);
					errFlag=1;
				}
				break;

			case ET_VAR:
				    // If the argument is an underscore,
				    //  don't check for type mismatch.
				if(temp->first->token == EXPR_UNDERSCORE) break;
			case ET_VAR|ET_NUM:
			case ET_STRVAR:
			case ET_MIXVAR:
				// The argument is either single variable
				// (including one that can take string values)
				//  or an expression. Need to check whether
				// the types of all variables are the same as
				// the type of the argument:
				for(exp=temp->first;exp;exp=exp->next)
					if(exp->token==EXPR_VAR &&
					   exp->var->type!=argType )
					{
						semanticError(SE_PP_VAR_TYPE,
						              pp, argList,i);
						errFlag=1;
						break;
					}
				break;
			default:
				fprintf(stderr, "CRITICAL ERROR in "
					"Rules::createArguments\n"
					"Incorrect argument type during "
					"parsing program predicate\n");
				exit(1);
			}
		}
	}


		// If the error occurred on any of the previous steps,
		//  release all the memory and return zero.
	if(errFlag)
	{
		for( ; argList ; temp=argList, argList=argList->next,
		    deleteExpression(temp));

		return(0);
	}

	    // Remove all expression handler nodes from the argument list and
	    //  reverse order of the arguments.
	    //  If some argument is an expression, place expression node with
	    //  token == EXPR_EXPR instead of the expression handler node:
	    // expList - the resulting list.
	for(expList=exp=0 ; argList ; temp=argList,
	    argList=argList->next, tempSrc.Delete(temp))
	{
		    // If there is only one node attached to the given handler:
		if(argList->first==argList->last)
		{
			    // Add this node to the 'expList'
			argList->first->next=expList;
			expList=argList->first;
		}
		else
		{
			    // If there are more than one node, allocate the new
			    //  expression node and put it in place of handler
			exp=expSrc.New();
			exp->token=EXPR_EXPR;
			exp->exp=argList->first;
			    // Add the resulting node to the 'expList'
			exp->next=expList;
			expList=exp;
		}
	}

	return(expList);
}



/*
void Rules::checkRepetitionsInData(WFnode *wNode, Value *oldList)
DESCRIPTION
	This method is called while converting arguments of program
	predicates or weight functions from lists to array.
	After the array is built and sorted, this method checks 
	whether there are any repetitions of the elements in 
	the array. If found, it reports a warning (for data predicates)
	or an error (for a weight function) and eliminates these repetitions

PARAMETERS
	WFnode  -  Pointer to either data predicate node or weight function
	            node (treated based on the type of the wNode->symbolno)
	oldList -  Old argument list of the predicate. It is necessary because
	            it stores the information regarding file names and line
	            numbers where the error occured.
*/
void Rules::checkRepetitionsInData(WFnode *wNode, Value *oldList)
{
    int in, out, j;
    int arity, limit;
    int *args;
    short isWeight, isEqual;
    int lastNotEq;
    Value *cur, *badList, *temp;

    isWeight=(symbt.getType(wNode->symbolno) == TYPE_WEIGHT_FUN);
    args=wNode->arguments;
    arity=wNode->arity;
    limit=arity*(wNode->valNum-1);

        // Go over the arguments, check for repetitions and compact the
        // array of arguments by eliminating the repetitions:
        // NOTATION: [[N]] - means vector with number of elements equal
        //    to arity starting from the element args[N].
    for(in=0,out=0; in < limit ; in+=arity, out+=arity)
    {
        if(isWeight) wNode->results[out/arity]=wNode->results[in/arity];

            // Copy the [[in]] to [[out]] and check whether
            //     [[in]] == [[in+arity]]
            //  ('isEqual' indicates that they are equal).
        for(isEqual=1,j=0;j<arity;j++)
        {
            args[out+j]=args[in+j];
            if(args[in+j]!=args[in+j+arity]) isEqual=0;
        }

            // If there is a repetition
        if(isEqual)
        {
            dataRepetitionCount++;

                // Check whether we need to report it
            if(dataRepetitionCount > MAX_DATA_REPETITION_MSG)
                    // If the messages should be suppressed, just call the
                    //  'semanticError' method to take this error into account
                semanticError(SE_DATA_REPETITION, wNode,0,0);
            else
            {// Analyze the error and print the message:
                // We detected repetitions of [[in]].
                // Now we need to search through the old list and find
                //  corresponding vectors there. This will let us to report
                //  the file names and the line numbers where this repetition
                //  has occurred.
                //  NOTE: the variable 'lastNotEq' contains the number
                //   (in the list) of the last node that was not equal to the
                //   coresponding element in the [[in]].
                for(j=0, lastNotEq=-1, cur=oldList, badList=0 ; cur ;
                    cur=cur->next, j++)
                {
                        // Check whether the current value in the list is equal
                        // to the corresponding value in the [in] vector.
                        //  if not, then set lastNotEq to the current number.
                    if(cur->num!=args[in+(j%arity)]) lastNotEq=j;

                        // If we at the last element in a vector:
                        // Check whether the vector is equal to [[in]]
                        //  If the last element that was not equal not less
                        //  than the size of vector away from the current
                        //  element, then the last vector is equal to the [[in]]
                    if(!((j+1)%arity) && lastNotEq + arity <= j)
                    {
                            // Allocate the new node to carry information about
                            // the line numbers and file names. Copy information
                            //  to the node and add the node to the list:
                        temp=valSrc.New();
                        memcpy(temp, cur, sizeof(Value));
                        temp->next=badList; badList=temp;
                    }
                }

                    // Report the semantic error or warning:
                semanticError(SE_DATA_REPETITION, wNode, in, badList);

                if(dataRepetitionCount == MAX_DATA_REPETITION_MSG)
                    semanticError(SM_SUPPRESS_REPEAT_MSG);


                    // For every in such that [[in]]==[[in + arity]]: in+=arity.
                for(in+=arity;in<limit;in+=arity)
                {
                        // Compare all the elements of [[in]] and [[in+arity]].
                    for(j=0;j<arity && args[in+j]==args[in+j+arity];j++);
                        // If comparison failed before we reached the end of the
                        //  vectors, they are not equal => stop the loop.
                    if(j<arity) break;
                }
            }
        }
    }

        // Copy the last vector to [[out]] (limit was initialized
        //  the way that the last element was left over)
        //  Copy the weight function result:
    if(isWeight && in<limit+arity)
        wNode->results[out/arity]=wNode->results[in/arity];
        //  Copy the arguments
    for(;in<limit+arity;in++,out++) args[out]=args[in];

        // Save the new number of vectors:
    wNode->valNum=out/arity;
}


/*
int Rules::expEvaluate(int left, int right, int Operation, int *result)
DESCRIPTION
	Compile-time expression evaluation

PARAMETERS
	arg1       - First argument (number)
	arg2       - Second argument (number)
	Operation  - Operation code
	result     - Pointer to the variable where the result is to be stored.

RETURNS
	TRUE    -    Everything is OK
	FALSE   -    Error (Division by zero).
*/
int Rules::expEvaluate(int left, int right, int Operation, int *result)
{
	switch(Operation)
	{
	case EXPR_PLUS:      *result = left+right; break;
	case EXPR_MINUS:     *result = left-right; break;
	case EXPR_MUL:       *result = left*right; break;
	case EXPR_DIV:
		if(right==0)
		{
			semanticError(SE_DIV_ZERO, EXPR_DIV);
			return 0;
		}
		*result = left/right;
		break;
	case EXPR_MOD:
		if(right==0)
		{
			semanticError(SE_DIV_ZERO, EXPR_MOD);
			return 0;
		}
		*result = left%right;
		break;
	case EXPR_MIN:       *result = (left>right)?right:left; break;
	case EXPR_MAX:       *result = (left<right)?right:left; break;
	case EXPR_ABS:       *result = (left>0)?left:-left; break;
	case EXPR_UMINUS:    *result = -left; break;
	default:
		fprintf(stderr, "CRITICAL ERROR in Rules::expEvaluate\n"
			"unsupported operation #%d\n\n", Operation);
		exit(1);
	}

	return 1;
}




/****************************************************************************
*****************************************************************************
******    METHODS TO RELEASE MEMORY USED FOR DATA STRUCTURES
*****************************************************************************
****************************************************************************/

// NOTE: The methods in this section DO NOT release memory physically.
//  They just return it to the corresponding instances of 'MemNode'
//  to be reused later by the program

	// Releases memory allocated for a permanent rule
void Rules::deleteRule(RuleNode *rule)
{
	deleteAtomList(rule->body);
	deleteAtomList(rule->head);
	deleteQuantifierList(rule->quant);

	deleteMetaList(rule->cardBody);
	deleteMetaList(rule->cardHead);
	deleteMetaList(rule->weightBody);
	deleteMetaList(rule->weightHead);

	ruleSrc.Delete(rule);
}


	// Releases memory allocated for a permanent list of
	// cardinality or weight atoms
void Rules::deleteMetaList(MetaAtom *list)
{
	MetaAtom *temp;

	for( ; list ; temp=list, list=list->next, metaSrc.Delete(temp))
	{
		deleteExpression(list->min);
		deleteAtomList(list->list);
		deleteExpression(list->max);
	}
}


	// Releases memory allocated for a list of permanent atoms
void Rules::deleteAtomList(Atom *list)
{
	Atom *temp;

	for(;list;temp=list,list=list->next,atomSrc.Delete(temp))
	{
		deleteExpression(list->arg);
		deleteQuantifierList(list->quant);
		deleteExpression(list->weight);
	}
}


/*
void Rules::deleteExpression(ExpressionHandler *eh, int leaveErrorHandler)
DESCRIPTION
	Releases memory allocated for an arithmetic expression, given
	the expression handler
PARAMETERS
	eh                 -  Expression to deallocate
	leaveErrorHandler  -  Optional flag. If this is TRUE, the expression is
	                       not deallocated completely. The method leaves the
                               expression handler set as error expression.
*/
void Rules::deleteExpression(ExpressionHandler *eh, int leaveErrorHandler)
{
	if(!eh) return;

	    // Releasing every node of the expression
	deleteExpression(eh->first);

	    // Either make error expression or deallocate expression handler:
	if(leaveErrorHandler)
	{
		eh->type=ET_ERROR;
		eh->first=eh->last=0;
	}
	else
		tempSrc.Delete(eh);
}


	// Release memory allocated for an arithmetic expression given
	// the list of expression nodes
void Rules::deleteExpression(Expression *exp)
{
	Expression *temp;

		// Go over the list:
	for(temp=exp ; exp ; temp=exp, exp=exp->next, expSrc.Delete(temp))
		if(exp->token==EXPR_EXPR)
				// Release the attached expression
			deleteExpression(exp->exp);
		else
			if(exp->token==EXPR_DPOSITIVE || exp->token==EXPR_DNEGATIVE)
			{
				// In this case, we need to release argument
				// list and deallocate the atom node
				deleteExpression(exp->atom->arg);
				atomSrc.Delete(exp->atom);
			}
}


	// Release memory allocated for a list of quantifiers
void Rules::deleteQuantifierList(Quantifier *list)
{
	Quantifier *temp;

	for(;list;temp=list,list=list->next,quantSrc.Delete(temp))
	{
			// If this is data predicate quantifiers,
			//  release its argument list:
		if(list->token!=QUANT_VAR && list->token!=QUANT_EMPTY)
			deleteExpression(list->arg);
			// Release condition list:
		deleteExpression(list->conditions);
	}
}


	// Release the memory allocated for a list of temporary atoms:
void Rules::deleteTempAtomList(TempAtom *list)
{
	TempAtom *atom;

		// Go over the list of atoms and release the atom nodes after
		//  they are processed
	for(atom=list ; atom ; list=atom, atom=atom->next, tempSrc.Delete(list))
		switch(atom->type)
		{
		case ATOM_TRUE:
		case ATOM_FALSE:
			break;

		case ATOM_CONDITION:
			deleteExpression(atom->cond.first);
			break;

		case ATOM_DATA:
		case ATOM_DATANEG:
			deleteExpression(atom->data.arg);
			break;

		case ATOM_PROG:
			// For the program atom: release argument list,
			//  local variables, weight expression, and
			//  data restrictions:
			deleteExpression(atom->prog.weight);
			deleteTempAtomList(atom->prog.restrictions);
			deleteQuantifierList(atom->prog.quant);
			deleteExpression(atom->prog.arg);
			break;
		case ATOM_WEIGHT:
		case ATOM_CARD:
			// For meta-atom release meta-list and meta-limits:
			deleteTempAtomList(atom->meta.list);
			deleteExpression(atom->meta.min);
			deleteExpression(atom->meta.max);
			break;
		default:
			fprintf(stderr, "CRITICAL ERROR: Unknown type of a"
				" temporary atom: %d",atom->type);
			exit(1);
		}
}




/****************************************************************************
*****************************************************************************
******   IMPLEMENTATION OF 'QuantBuilder'
*****************************************************************************
****************************************************************************/

/*
Quantifier *QuantBuilder::build(TempAtom *conditions, Quantifier *quantList)
DESCRIPTION
	THE PRIMARY METHOD OF THE CLASS. Given the list of conditions and 
	the list of quantified variables, it creates a list of quantifiers

PARAMETERS
	conditions   -  Sinle-linked list of conditions, which may consist of
	                 data atoms, negative data conditions, or predefined
	                 atoms. For local quantifiers, this is a list from the
	                 data restrictions. It can be NULL
	quantList    -  Single-linked list of quantified variables.
	                 Can be list of local or global variables.
	                It can be NULL
RETURNS
	Double-linked list of quantifiers constructed from the local variables
	and conditions.
*/
Quantifier *QuantBuilder::build(TempAtom *conditions, Quantifier *quantList)
{
	TempAtom *minCondition, *atom, *nextAtom;
	float minCost,cost;
	Quantifier *startQuant, *curQuant;
	Expression *curExp, *tmpExp;
	VarNode *minVar, *var;
	int dataModifyFlg;

	// NOTE: During execution of the main loop, the field
	//      'VarNode::tmpQuant' shows whether the variable is
	//       quantified or not. If NULL, variable is not quantified;
	//       otherwise, it is quantified

	    // Initialize variables in conditions and
	    // convert the list of conditions to a doubly linked list
	for(; (atom=conditions) ;)
	{
		    // Set 'conditions' to the next node so that we can
		    //  safely modify the links of the current node:
		conditions=conditions->next;

		    // Append the current node to the new double linked list:
		DLL_APPEND_BEFORE(atom, &conditionList)

		    // Get the expression list from the condition
		    //  (to look for variables).
		switch(atom->type)
		{
			case ATOM_CONDITION:
				curExp=atom->cond.first;
				break;
			case ATOM_DATA:
			case ATOM_DATANEG:
				curExp=atom->data.arg;
				break;
			default:
				fprintf(stderr, "CRITICAL ERROR in "
				   "method QuantBuilder::build\n\n");
				exit(1);
		}

		    // For every variable in the expression from the condition:
		    //  mark is as VAR_FREE and initialize 'tmpQuant' field.
		for(;curExp;curExp=curExp->next)
		{
			    // If there is a sub-expression, need to process it
			if(curExp->token==EXPR_EXPR)
			{
				for(tmpExp=curExp->exp ; tmpExp ;
				    tmpExp=tmpExp->next)
					if(tmpExp->token == EXPR_VAR)
					{
						tmpExp->var->tmpQuant=0;
						tmpExp->var->valIndex=VAR_FREE;
					}
			}
			else
				if(curExp->token==EXPR_VAR)
				{
					curExp->var->tmpQuant=0;
					curExp->var->valIndex=VAR_FREE;
				}
		}
	}

	    // Convert the list of quantified variables to a double-linked list.
	    // The quantifier nodes will now serve to store the information
	    // regarding quantified variables
	for(; (curQuant=quantList) ;)
	{
		    // Go to the next node to be able to modify
		    //  the links of the current:
		quantList=quantList->next;

		    // Append the current node to the new list:
		DLL_APPEND_BEFORE(curQuant, &quantVarList);

		    // Initialize the quantified variable
		    // Variable is already linked to the quantifier,
		    // now, link the quantifier to the variable.
		    // Thus, the variable will be considered quantified
		curQuant->var->tmpQuant=curQuant;
		curQuant->var->valIndex=VAR_FREE;
	}

	// Starting the main loop
	    // Initialize starting quantifier:
	    // Serving as a dummy node for a circular double linked list
	    // of quantifiers it may actually become an empty quantifier
	    // if some conditions will be attached to it.
	startQuant=quantSrc.New();
	startQuant->next=startQuant->prev=startQuant;
	startQuant->token=QUANT_EMPTY;
	startQuant->conditions=0;

	while(1)  // THE MAIN LOOP
	{
		    // Update the costs of quantified variables:
		    // (the data predicate quantifiers that don't have
		    // quantified variables are automatically added to the
		    // list of conditions for the current quantifier)
		updateCosts(&(startQuant->prev->conditions), &curExp);


		    // FIND THE CONDITION WITH THE LEAST COST. Process
		    // the conditions that have costs equal to zero.
		    // NOTE: in the following loop the 'nextAtom' is
		    // right after the atom to be processed
		    // Thus, the loop starts from 'conditionList.next->next'
		    // and finishes when "nextAtom==conditionList->next".
		for(dataModifyFlg=0, minCondition=0, minVar=0,
		    minCost=FLOAT_MAX, nextAtom=conditionList.next->next ;
		    nextAtom->prev!=&conditionList ; nextAtom=nextAtom->next)
		{
			// We are working with the atom right before 'nextAtom':
			atom=nextAtom->prev;

			// Calculate the cost of that condition:
			// If cost is positive, 'var' receives the pointer
			//  to the variable in the condition with the least cost
			cost=conditionCost(atom, &var);

			if(cost<=0)
			{// WE CAN BE HERE IN TWO CASES:
			//  1) 'atom' is a data atom for which all variables
			//     arithmetic expressions are known.
			//     Therefore, this atom can work as a data
			//      predicate quantifier
			//  2) 'atom' is a predefined atom or negative
			//      data atom without quantified variables
			//    Therefore, need to remove this atom, convert
			//      it into conditon, and append to current
			//      quantifier.

				// Remove the atom from the condition list:
				DLL_REMOVE(atom);

				if(atom->type==ATOM_DATA) // Case 1:
				{
					// Append the atom to the data predicate
					//  quantifier list and indicate that
					//  the list have been changed
					dataModifyFlg=1;
					DLL_APPEND_BEFORE(atom, &dataAtomList);
				}
				else // Case 2:
				    createCondition(atom,
				      &(startQuant->prev->conditions), &curExp);
			}
			else   // If the cost is positive,
			       //  we can look for a minimum cost
				if(cost<minCost)
				{
					minCost=cost;      // Save the cost
					minVar=var;  // Save the least variable
					minCondition=atom; // Save the atom
				}
		}

		    // If the data predicate quantifier list was modified,
		    //  the 'minCost', 'minVar', and 'minCondition'
		    //  calculated by the previous loop are no longer
		    //  valid since the costs might have changed.
		    // Therefore, we need to redo that thing again.
		    //  However, there will be no more atoms in the
		    //  conditionList with the zero cost, since all of them
		    //  have been removed by the previous loop (thus,
		    //  'dataAtomList' won't be modified anymore).
		if(dataModifyFlg)
		{
			    // Update the costs of the variables (condition
			    //  list may be also increased by some atoms
			    //  added to 'dataAtomList' in the previous step
			updateCosts(&(startQuant->prev->conditions), &curExp);

			    // Find the condition with the least cost and
			    // variable in that condition with the least cost:
			for(minCondition=0, minCost=FLOAT_MAX,
			    atom=conditionList.next ; atom != &conditionList ;
			    atom=atom->next)
				if((cost=conditionCost(atom, &var)) < minCost)
				{
					minCost=cost;
					minVar=var;
					minCondition=atom;
				}
		}

		    // "minCondition==NULL" means the conditionList is empty:
		if(!minCondition) break;

		    // At this moment, 'minVar' always have to exist
		if(!minVar)
		{
			fprintf(stderr, "CRITICAL ERROR in QuantBuilder::build"
				" : logical error\n\n");
			exit(1);
		}

		    // Remove the best quantifier for the variable with the
		    //  least cost in the condition with the least cost
		    //  and append it to the circular double linked list
		    //  of quantifiers:
		curQuant=removeBestQuantifier(minVar);
		DLL_APPEND_BEFORE(curQuant, startQuant);
	}

	// At this moment, all the conditions are taken care of.
	//  Now, we need to get variable quantifiers for the
	// remaining quantified variables:
	while((curQuant=removeBestVariable())) // Not error
	{
		    // Remove the best quantifier for the best variable
		    //  and append it to the quantifier list:
		DLL_APPEND_BEFORE(curQuant, startQuant);
		updateCosts(&(curQuant->conditions), &curExp);
	}

	// Break up the circular linked list of the quantifiers:
	// Now, it will be just double linked list:
	startQuant->prev->next=0;
	startQuant->prev=0;

	    // If there are no conditions attached to the 'startQuant', we don't
	    //  need an empty quantifier. Therefore, remove the first node in
	    //  the list, deallocate it, and set 'startQuant' to be equal to
	    //  the next node in the list ('startQuant' is to be returned).
	if(!startQuant->conditions)
		if(startQuant->next)
		{
			startQuant=startQuant->next;
			quantSrc.Delete(startQuant->prev);
			startQuant->prev=0;
		}
		else
		{
			quantSrc.Delete(startQuant);
			startQuant=0;
		}

		// Check consistency: make sure the lists 'dataAtomList',
		//  'quantVarList', and 'conditionList' are in fact empty:
	if(dataAtomList.prev!=&dataAtomList || dataAtomList.next!=&dataAtomList
	 || quantVarList.prev!=&quantVarList || quantVarList.next!=&quantVarList
	 || conditionList.prev!=&conditionList
	 || conditionList.next!=&conditionList)
	{
		fprintf(stderr, "CRITICAL ERROR in method QuantBuilder::build\n"
			"LOGICAL ERROR: at least one linked list is left in\n"
			"the inconsistent state\n\n");
		exit(1);
	}
	// this code will move condition containing DIV and MOD operations
	// to the end of the list of conditions
	for (curQuant=startQuant; curQuant; curQuant=curQuant->next)
		moveDivModToEnd(curQuant);
	return(startQuant); // Return the quantifier list
}

/*
void QuantBuilder::moveDivModToEnd(Quantifier *quant)
DESCRIPTION
	This method moves conditions containing operators divide and modulo
	to the end of the list of conditions for a given quantifier. The purpose
	of this operation is to postpone evaluation of conditions involving
	division until other data restrictions on the variable in question
	are evaluated to prevent division by 0 errors

PARAMETERS
	quant   -  the pointer to the quantifier for which the list of conditions
	           will be rearranged
*/
void QuantBuilder::moveDivModToEnd(Quantifier *quant)
{
	Expression *cond,*first=NULL,*last,*movhead=NULL,*movtail=NULL;
	int flag=0;
	for (cond=quant->conditions; cond; cond=cond->next) {
		switch (cond->token) {
		case EXPR_DIV:
		case EXPR_MOD:
			flag=1;
			break;
		// each condition must end with one of the operators given below
		case EXPR_LESS:
		case EXPR_NOTGREATER:
		case EXPR_GREATER:
		case EXPR_NOTLESS:
		case EXPR_EQUAL:
		case EXPR_NOTEQUAL:
		case EXPR_ASSIGN:
			last=cond;
			if (flag) {
				if (movhead) 
					movtail->next=(first) ? first->next : quant->conditions;
				else
					movhead=(first) ? first->next : quant->conditions;
				movtail=last;
				if (first)
					first->next=last->next;
				else
					quant->conditions=last->next;
				flag=0;
			} else
				first=last;
			break;
		default:
			break;
		}
	}
	if (movhead) {
		if (first)
			first->next=movhead;
		else
			quant->conditions=movhead;
		movtail->next=NULL;
	}
}

/*
void QuantBuilder::updateCosts(Expression **firstPtr, Expression **lastPtr)
DESCRIPTION
	This method updates costs of all quantified variables. Also, if some
	data predicate quantifier doesn't have quantified variables, the
	method removes it from the list, creates condition and adds it
	to the condition list given in parameters

PARAMETERS
	firstPtr   -  Points to the pointer to the first element of the list
	                of conditions
	lastPtr    -  Points to the pointer to the last element of the list
	                of condiitons. NOTE, if "*firsPtr==NULL", the value
	                '*lastPtr' doesn't make any difference and will
	                be initialized after the first condition is added
*/
void QuantBuilder::updateCosts(Expression **firstPtr, Expression **lastPtr)
{
	Quantifier *quant;
	TempAtom *atom, *nextAtom;
	Expression *exp;
	int varCount;
	float cost;
	VarNode *var;

		// For every quantifed variable:
		//  costOfVariable = numOfValuesInVariableDataType
		// Best quantifier = NULL.
	for(quant=quantVarList.next;quant!=&quantVarList;quant=quant->next)
	{
		quant->tmpCost=(float)(quant->var->type->valNum);
		quant->tmpBestQuant=0;
	}

	repeatStk.clear(); // Clear the repetition stack (just in case)

		// For every data predicate quantifier:
	for(nextAtom=dataAtomList.next->next;nextAtom->prev!=&dataAtomList;
		nextAtom=nextAtom->next)
	{
		atom=nextAtom->prev;

		    // Calculate the number of DISTINCT quantified variables
		    // Don't look into expressions as parameters, since
		    // they are not supposed to have quantified variables
		for(varCount=0,exp=atom->data.arg;exp;exp=exp->next)
			if(exp->token==EXPR_VAR && exp->var->tmpQuant &&
				exp->var->valIndex==VAR_FREE)
			{
				varCount++;
					// Handle repetitions:
				exp->var->valIndex=VAR_FIRST;
				repeatStk.push(exp->var);
			}

		if(varCount<=0) // If no quantified variables:
		{
			// Remove the atom from the list of data predicate
			//  quantifiers, create condition from this atom and
			//  append it to the list given in parameters
			DLL_REMOVE(atom);
			createCondition(atom, firstPtr, lastPtr);
		}
		else
		{
			// Calculate the cost of the data predicate quantifier
			cost=(float)pow((double)atom->data.dp->valNum,
			                1.0/((double)varCount));

			// Update the costs of all quantified variables:
			while(repeatStk.getNumOfElements())
			{
				// Return the variable attribute
				//  back to VAR_FREE
				var=repeatStk.pop();
				var->valIndex=VAR_FREE;


				if(var->tmpQuant->tmpCost > cost)
				{
					// Update the cost and best quantifier:
					var->tmpQuant->tmpCost=cost;
					var->tmpQuant->tmpBestQuant=atom;
				}
			}
		}
	}
}


/*
float QuantBuilder::conditionCost(TempAtom *atom, VarNode **minVarPtr)
DESCRIPTION
	This method calculates the condition cost and if the cost is positive
	it finds a quantified variable in the condition with the least cost

PARAMETERS
	atom       -  Atom which cost to calculate. Can be a data atom,
	               negative data predicate condition, or predefined atom
	minVarPtr  -  The variable pointed by this parameter will receive
	                the least-cost variable
RETURNS
	The cost of the condition (see documentation for more details on
	how it is calculated)
*/

// This macro is used to count a cost of a variable.
//  It counts only quantified variables and uses stack to count
//  only distinct variables. It also sets a flag once finds
//  a first quantified variable.
#define COUNT_EXP(exp) \
	{if((exp)->token==EXPR_VAR && (exp)->var->tmpQuant && \
		(exp)->var->valIndex==VAR_FREE) \
			{(exp)->var->valIndex=VAR_FIRST;\
			  prodCost*=(exp)->var->tmpQuant->tmpCost;\
			  repeatStk.push((exp)->var); qVarFlg=1;}}

float QuantBuilder::conditionCost(TempAtom *atom, VarNode **minVarPtr)
{
	Expression *exp, *subExp;
	VarNode *var, *minVar;
	short condFlg;    // Set to 1 for predefined atoms
	short expOnlyFlg; // Set to 1 for data atoms
	float mainOpCost, prodCost, minVarCost;
		// Flag indicating whether there are quantified
		// variables in the condition:
	int qVarFlg;

		// Initialize variables:
	condFlg=expOnlyFlg=0;
	exp=0;

		// Analyze the type of the atom, find arithmetic expression
		//  and set the flag based on the type:
	switch(atom->type)
	{
	case ATOM_CONDITION:
		exp=atom->cond.first;
		condFlg=1;
		break;
	case ATOM_DATA:
		expOnlyFlg=1;
	case ATOM_DATANEG:
		exp=atom->data.arg;
		break;
	default:
		fprintf(stderr,"CRITICAL ERROR in method "
			"quantBuilder::conditionCost\n"
			"wrong atom is given as a parameter\n\n");
		exit(1);
	}

	repeatStk.clear();

		// Go over the atoms in the expression:
	for(qVarFlg=0, mainOpCost=0, prodCost=1 ; exp ; exp=exp->next)
	{
		switch(exp->token)
		{
		case EXPR_VAR:
			// If variable, it will contribute to the cost unless
			//  we are dealing with data predicate condition
			if(!expOnlyFlg) COUNT_EXP(exp);
			break;

		case EXPR_EXPR:
			// All the variables in the sub-expression will
			//  always contribute to cost: need to count them
			for(subExp=exp->exp;subExp;subExp=subExp->next)
				COUNT_EXP(subExp)
			break;

		// The rest of the cases will count the cost of
		// the main operator in case of predefined atom
		// (see documentation for more details)
		case EXPR_EQUAL:
			mainOpCost=3;
			break;

		case EXPR_GREATER:
		case EXPR_LESS:
		case EXPR_NOTGREATER:
		case EXPR_NOTLESS:
			mainOpCost=4;
			break;

		case EXPR_ASSIGN:
			mainOpCost=5;
			break;

		case EXPR_NOTEQUAL:
			mainOpCost=6;
			break;
		}
	}

		// For all distinct quantified variables (the ones counted
		// in the cost) restore 'var->valIndex' to its initial state
		// and find the variable with the least cost
	for(minVarCost=FLOAT_MAX, minVar=0 ; repeatStk.getNumOfElements() ;)
	{
		var=repeatStk.pop();
		var->valIndex=VAR_FREE;

		if(var->tmpQuant->tmpCost < minVarCost)
		{
			minVarCost=var->tmpQuant->tmpCost;
			minVar=var;
		}
	}

	if(minVarPtr) *minVarPtr=minVar;

		// No quantified variables => condition cost is zero.
	if(!qVarFlg) return 0;

	if(condFlg)
	{
			// If this is condition, make sure that the main
			//  operator cost is positive
		if(mainOpCost==0)
		{
			fprintf(stderr, "CRITICAL ERROR in the method "
				"QuantBuilder::conditionCost\n"
				"incorrect parameter: condition is given"
				" without the main operator\n\n");
			exit(1);
		}
	}
	else
		// If this is positive or negative data predicate condition,
		//  assign the cost of the main operator manually
		mainOpCost=1;	// Cost of the data quantifier.

		// Return the cost of the condition
	return(mainOpCost*prodCost);
}


/*
Quantifier *QuantBuilder::removeBestVariable()
DESCRIPTION:
	This method finds the variable with the least cost among the quantified
	variables and returns the best quantifier for that variable.
*/
Quantifier *QuantBuilder::removeBestVariable()
{
	Quantifier *quant;
	VarNode *var;
	float minCost;

	// Find the variable with the least cost among quantified variables:
	for(var=0,minCost=FLOAT_MAX,quant=quantVarList.next;
			quant!=&quantVarList;quant=quant->next)
		if(quant->tmpCost<minCost)
			{
				minCost=quant->tmpCost;
				var=quant->var;
			}

		// Remove the best quantifier for that variable:
	return(removeBestQuantifier(var));
}


/*
Quantifier *QuantBuilder::removeBestQuantifier(VarNode *var)
DESCRIPTION:
	For a given variable, the method constructs the best quantifier
	and returns it.
*/
Quantifier *QuantBuilder::removeBestQuantifier(VarNode *var)
{
	Quantifier *quant;
	TempAtom *bestQuant;
	Expression *exp, *subExp;

		// If given parameter is NULL or the variable is not
		//  quantifed, return NULL
	if(!var||!var->tmpQuant) return 0;

		// Find the best quantifier of the given variable
	bestQuant=var->tmpQuant->tmpBestQuant;

		// If the best quantifier is NULL, remove the variable,
		//  create the variable quantifier and return it.
	if(!bestQuant)
	{
		removeVariable(var);

			// Allocate, initialize, and return the quantifier.
		quant=quantSrc.New();
		quant->token=QUANT_VAR;
		quant->next=quant->prev=0;
		quant->conditions=0;
		quant->var=var;
		return(quant);
	}

		// best quantifier is not null: the atom it points to should
		//  always be a data atom.
	if(bestQuant->type!=ATOM_DATA)
	{
		fprintf(stderr, "CRITICAL ERROR: in method QuantBuilder::"
			"removeBestQuantifier\nbest quantifier of a variable "
			"should be a data atom but it is not \n\n");
		exit(1);
	}

		// Remove the atom from the list of data atom quantifiers.
	DLL_REMOVE(bestQuant);

		// Remove all variables quantified by this atom while
		//  marking them as EXPR_VAR_QUANT.
		// Also, make sure that all the variables inside
		// arithmetic expressions are known.
	for(exp=bestQuant->data.arg ; exp ; exp=exp->next)
	{
		if(exp->token==EXPR_EXPR)
		{
				// If data atom contains expressions,
				//  make sure that all the variables inside
				//  them are known at this point.
			for(subExp=exp->exp ; subExp ; subExp=subExp->next)
				if(subExp->token==EXPR_VAR &&
				   subExp->var->tmpQuant)
				{
				    fprintf(stderr, "CRITICAL ERROR: in method"
				        " QuantBuilder::removeBestQuantifier\n"
				        "LOGICAL ERROR: quantifier to be"
				        " output contains an expression with "
				        "unknown variables\n\n");
				    exit(1);
				}
		}
		else
				// If the variable is quantified, remove it
				//  and mark as EXPR_VAR_QUANT.
			if(exp->token==EXPR_VAR && exp->var->tmpQuant)
			{
				exp->token=EXPR_VAR_QUANT;
				removeVariable(exp->var);
			}
	}

		// Create new data predicate quantified and initialize it:
	quant=quantSrc.New();
	quant->token=QUANT_DP;
	quant->next=quant->prev=0;
	quant->conditions=0;
	quant->dp=bestQuant->data.dp;
	quant->arg=bestQuant->data.arg;

		// Release temporary atom node:
	tempSrc.Delete(bestQuant);

	return(quant);
}


/*
void QuantBuilder::removeVariable(VarNode *var)
DESCRIPTION:
	Given quantified variable converts it to non-quantified.
*/
void QuantBuilder::removeVariable(VarNode *var)
{
	Quantifier *quant;

	if(var->tmpQuant)
	{
		quant=var->tmpQuant;
		DLL_REMOVE(quant);
		quantSrc.Delete(quant);
		var->tmpQuant=0;
	}
}


/*
void QuantBuilder::createCondition(TempAtom *tempAtom, Expression **firstPtr,
                                                       Expression **lastPtr)
DESCRIPTION
	The function converts predefined atom, data atom, or negative data atom
	from their temporary representations using TempAtom to permanent
	representation as an expression. Also it appends the result of the
	conversion to the single-linked list given in parameters

PARAMETERS
	tempAtom   -  atom to convert
	firstPtr   -  Points to the pointer to the first element of the list
	                of conditions
	lastPtr    -  Points to the pointer to the last element of the list
	                of condiitons. NOTE, if "*firsPtr==NULL", the value
	                '*lastPtr' doesn't make any difference and will
	                be initialized after the first condition is added
*/
void QuantBuilder::createCondition(TempAtom *tempAtom, Expression **firstPtr,
                                                       Expression **lastPtr)
{
	Expression *first;
	Expression *last;
	Atom *atom;

	if(tempAtom->type==ATOM_DATA||tempAtom->type==ATOM_DATANEG)
	{	// Create positive or negative data predicate condition

			// Allocate and initialize permanent atom node:
		atom=atomSrc.New();
		atom->next=0;
		atom->quant=0;
		atom->weight=0;
		atom->arg=tempAtom->data.arg;
		atom->dp=tempAtom->data.dp;

			// Allocate and initialize expression node:
		first=last=expSrc.New();
		first->next=0;
		first->token=(tempAtom->type==ATOM_DATA)?
		              EXPR_DPOSITIVE:EXPR_DNEGATIVE;
		first->atom=atom;
	}
	else
		if(tempAtom->type==ATOM_CONDITION)
		{	// Create condition from predefined atom:
			//  simply extract expression
			first=tempAtom->cond.first;
			last=tempAtom->cond.last;
		}
		else
		{
				// Listed above are the only atoms that can be
				//  converted to the conditions
			fprintf(stderr, "CRITICAL ERROR in "
			    "QuantBuilder::createCondition\n"
			    "wrong parameter given: atom with the type %d\n\n",
			    tempAtom->type);
			exit(1);
		}

		// Release the temporary atom node
	tempSrc.Delete(tempAtom);

		// If the parameters firstPtr and lastPtr are given, append the
		//  resulting list to the list given by firstPtr and lastPtr.
	if(firstPtr&&lastPtr)
	{
		// If the given list is not empty, then the append procedure
		//  will modify the 'next' for the last node in the list
		// Otherwise, it will modify the pointer
		//  to the head of the list
		if(*firstPtr && *lastPtr) firstPtr=&((*lastPtr)->next);

		// Modify the pointers to append the
		//  resulting list to the given
		*firstPtr=first;
		*lastPtr=last;
	}
}
