/*==========================================================================;
 *
 *  AUTHOR:     Iakhiaev, Mikhail.
 *
 *  CONTENTS:   Template 'Stack' : stack based on ARRAY
 *              Template 'StackL': stack based on LIST OF PAGES
 *
 *  DATE:       March 28, 2004
 *
 *  FILE NAME:  stack.h
 *  VERSION:    2.2
 ***************************************************************************/

/*========================================================================
==========================================================================

DOCUMENTATION:
('type' - type from which template was constructed):

METHODS
  void push(type Element)
  type pop()
  type getElement(int depth) - return element at the distance 'depth'
                                from the top (depth==0 : return top)
  int getNumOfElements()     - return the number of elements in the stack
  void clear()               - remove all the elements from the stack

CONSTRUCTORS
  Stack(int initialNumberOfElements, float reallocationRate)
        initialNumOfElements - OPTIONAL: size of the initial array allocated
                               to store the elements of the stack
                              (SHOULD BE POSITIVE)
        reallocationRate     - OPTIONAL: if the array is full it is reallocated
                               and the size of the new array will be
                               'reallocationRate' the size of the old array
                               (SHOULD BE GRETER THAN ONE)

 StackL(size_t firstPageSize, float growthRatio, size_t maxPageSize)
        firstPageSize - OPTIONAL: size of the first page to be allocated
                        to store elements
        growthRatio   - OPTIONAL: size of every next page will be:
                        growthRatio*sizeOfCurrentPage, unless it is
                        more than 'maxPageSize' in which case every subsequent
                        page will have 'maxPageSize'.
        maxPageSize   - OPTIONAL: size of allocated page will never
                        exceed this number

EXAMPLE OF USAGE:

#include <stdio.h>
#include "stack.h"

int main()
{
  int i;
  Stack<int> stk(6); // Make initially 6 elements

  for(i=1;i<=10;i++) stk.push(i);

  printf("Number of Elements: %d\n", stk.getNumOfElements()); // Prints 10.
  stk.pop();stk.pop();
  printf("%d\n", stk.getElement(3));      // Prints 5.
  printf("%d\n", stk.pop());              // Prints 8.
  stk.pop(); stk.pop(); stk.pop();
  printf("%d\n", stk.pop());              // Prints 4.
  stk.clear();
  printf("%d\n", stk.getNumOfElements()); // Prints 0.
  return 0;
}

=============================================================================
=============================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#ifndef STACK_H
#define STACK_H
 //Constants governing Template Stack.
	//INCREASE - portion of increase stack size during memory
	//  reallocation in case of stack overflow.
	// !!! This should be greter than one !!!
#define STK_REALLOC_DEF         1.5
	//STK_SIZE_DEF - size of stack, allocated by a default constructor
#define STK_SIZE_DEF            100

// Sizes of pieces(pages) of memory allocated by StackL
//  will be divisible by STK_ROUNDOFF.
#define STK_ROUNDOFF            4

// Used in StackL
#define STK_FIRST_DEF           20
#define STK_MAX_DEF             1000
#define STK_MUL_DEF             2



	// STACK TEMPLATE BASED ON ARRAY
template <typename type> class Stack
{
protected:
	   // When the array is full, it will be reallocated and its size will
	   //  be increased that number of times.
	const float STK_REALLOC;
	type *bottom; // Beginning address of an array
	type *top;    // Pointer to the element right after the top of the stack
	int size;     // size of the stack (in elements)

	// This function is called when stack is full.
	//  It reallocates memory for the stack, size of the new piece is
	//   NewSize=OldSize*(1+STK_INCREASE).
	void reallocate()
	{
			// Save the old number of elements in the stack
		int oldSize=top-bottom;
			// Calculate the new size.
		size=(int)((float)size*STK_REALLOC + 4.0);
			// Reallocate the memory:
		bottom=(type *)realloc(bottom, size*sizeof(type));
			// Check for errors:
		if(!bottom || size <= oldSize)
		{
			fprintf(stderr,
			   "CRITICAL ERROR in the method "
			   "Stack::reallocate\n"
			   "Either cannot allocate more memory or value "
			   "of some constant is illegal\n\n");
			exit(1);
		}
			// Re-initialize the 'top'.
		top=bottom+oldSize;
	}

public:
		// Push operation for stack.
		//  If the stack is full, it reallocates memory.
	void push(type element)
	{
		if(top-bottom>=size) reallocate();

		*top=element;
		top++;
	}

		// Pop operation. If the stack is empty, then critical error.
	type pop()
	{
		if(top <= bottom)
		{
			fprintf(stderr,
				"CRITICAL ERROR in the method Stack::pop\n"
				"Popping from the empty stack\n\n");
			exit(1);
		}
			// If there are no errors, return the element
		return(*(--top));
	}

		// This function returns the value of an element
		//  which lies Num elements beneath the top
		//  (function call with Num==0 returns the top).
	type getElement(int num)
	{
		if(top-bottom <= num)
		{
			fprintf(stderr,
		              "CRITICAL ERROR in the method Stack::getElement"
		              "\nTrying to access non-existing element at the"
		              " distance %d from the top.\n"
		              "Since there are only %d elements in the stack,\n"
		              "the maximum parameter to the method can be %d"
		              "\n\n", num, getNumOfElements(),
			       getNumOfElements()-1);
			exit(1);
		}
		return(*(top-1-num));
	}

		// Returns number of elements in the stack.
	int getNumOfElements() { return (top-bottom); }

		// Clears the stack.
	void clear() {top=bottom;}

		// Allocates the stack with a given number of elements.
	Stack(int sizeOfStack=STK_SIZE_DEF,
		float reallocRate=STK_REALLOC_DEF): STK_REALLOC(reallocRate)
	{
		size=sizeOfStack;
		top=bottom = (type *)malloc(size*sizeof(type));
		if(size <= 0 || reallocRate <= 1 || !bottom)
		{
			fprintf(stderr,
				"CRITICAL ERROR in the constructor of the "
				"class \'Stack\'\nEither given incorrect "
				"parameter or the \'malloc\' returns NULL");
			exit(1);
		}
	}
		// Destructor.
	~Stack()
	{
		if(bottom) free(bottom);
		bottom=top=0;
		size=0;
	}
};

	// STACK TEMPLATE BASED ON A LIST OF PAGES
template <typename type> class StackL
{
		// Linked list of these structures contains the
		//  information about allocated pages of memory.
	struct StackPageDescriptor
	{
		type *page;                 // Pointer to the corresponding page
		int size;                   // Size of the page
		StackPageDescriptor *next;  // Next descriptor (PREVIOUS PAGE)
	};

	StackPageDescriptor *pageList;// Linked list of internal pages allocated
	                              //  by an object of the class using malloc

	   // Additional page that is before the pageList.
	   //  NOTE: This page is necessary for allocation/deallocation purposes
	   // The class allocates the next page when it is out of memory;
	   //  however, it doesn't deallocate a page once it is not used (all
	   //   the elements are popped) but rather when the previous page
	   // becomes half empty. This is done to avoid multiple
	   //  allocations/deallocations when one page is almost full or
	   //  another is almost empty.
	   // If a page is neither used nor deallocated, 'pageList' contains
	   //  pointer to the previous page and extraPage contains pointer to
	   //   that page; otherwise, extraPage==NULL.
	StackPageDescriptor *extraPage;

	size_t nextSize;   // Size of the next page to be allocated
	type *curPtr;      // Pointer to the element right after the top
	                   //  of the stack
	int prevSize;      // Total size of pages starting from pageList->next

	const size_t FIRST_PAGE_SIZE;// Size of first page to be allocated
	const float GROWTH_RATIO;  // Growth factor: the size of next page to be
	const size_t MAX_PAGE_SIZE;  // Maximal size of the page to be allocated
	                             // allocated is GROWTH_RATIO as big as the
	                             //  size of the current page.

	void addPage()
	{
		type *maxPtr;       // Temporary pointer
		int actualSize;     // Temporary size

		// NOT ENOUGH MEMORY IN THE CURRENT PAGE: NEED TO
		//  GET ADDITIONAL PAGE

			// At this moment, the size of the current page should
			//  be included in the 'prevSize'
		if(pageList) prevSize+=pageList->size;

			// If we have an extra page, use this page instead of
			//  allocating the new one
		if(extraPage)
		{
				// (at this moment extraPage->next==pageList)
			pageList=extraPage;    // Add extra page to the list:
			extraPage=0;           // No more extra page
			curPtr=pageList->page; // Reset the top of the stack
			return;
		}

		// AT THIS MOMENT, WE NEED TO ALLOCATE A NEW PAGE.

			// Calculate the size of the page to allocate.
			// By default, it is nextSize, but should not be
			//  less then the requested allocation size and
			//  should not be greater than MAX_PAGE_SIZE.
			//  It also should be divisible by STK_ROUNDOFF
		if(MAX_PAGE_SIZE>0&&nextSize>MAX_PAGE_SIZE)
			nextSize=MAX_PAGE_SIZE;

		actualSize=((nextSize*sizeof(type)+
			STK_ROUNDOFF+sizeof(StackPageDescriptor)-1)/
			STK_ROUNDOFF)*STK_ROUNDOFF-sizeof(StackPageDescriptor);

		     // Allocate the new page, test for error, initialize maxPtr
		curPtr=(type *) malloc(actualSize+sizeof(StackPageDescriptor));
		if(!curPtr)
		{
			fprintf(stderr, "CRITICAL ERROR Cannot allocate "
				"memory by the class \'StackL\'\n"
				"\'malloc\' retured NULL after "
				"requesting %d bytes",
				actualSize+sizeof(StackPageDescriptor));
			exit(1);
		}
			// Structure of the allocated piece of memory:
			//  [PAGE DATA] [PAGE DESCRIPTOR].
			// Make 'maxPtr' point to the place for the first byte
			//  byte of page descriptor.
		maxPtr=curPtr+actualSize/sizeof(type);

			// Add the page to the list and initialize the page
			//  descriptor
		((StackPageDescriptor *)maxPtr)->next=pageList;
		pageList=(StackPageDescriptor *)maxPtr;
		pageList->page=curPtr;
		pageList->size=maxPtr-curPtr;

			// Calculate the size of the next page.
		nextSize=(int)((float)nextSize*GROWTH_RATIO);
	}




public:
		// Push operation for stack. If the stack is full,
		//  it adds another page of memory.
	void push(type element)
	{
			// Check whether we have enough space in the current
			//  page. If not, get another page.
		if(curPtr >= pageList->page + pageList->size) addPage();

			// Write the element in the page:
		*curPtr=element;
		curPtr++;
	}

		// Pop operation. If the stack is empty, then critical error.
	type pop()
	{
			// Check whether we need to deallocate memory.
			//  Need to do that if there is an extra page and the
			//  current page (that should be before the extra page)
			//  is at most half full.
		if(extraPage && (curPtr - pageList->page) <= (pageList->size/2))
		{
				// The next page will be allocated with the
				//  same size as the page that was deallocated
			nextSize=extraPage->size;
			free(extraPage->page);   // Deallocate the page
			extraPage=0;             // No more extra page
		}

			// If there is no more elements in the current page:
		if(curPtr <= pageList->page)
		{
			if(pageList->next)    // If there is previous page:
			{
				     // The current page is now an extra page:
				extraPage=pageList;
				     // Go to the previous page:
				pageList=pageList->next;
				     // Adjust 'prevSize': reflect the fact
				     //  that the previous page should no
				     //  longer be counted in this variable
				prevSize-=pageList->size;
				     // Adjust 'curPtr' to point after the
				     //  Last element in the given page:
				curPtr=pageList->page+pageList->size;
			}
			else  // Otherwise, critical error:
			{     // popping from the empty stack
			    fprintf(stderr,
			       "CRITICAL ERROR in the method StackL::pop\n"
			       "Popping from the empty stack\n\n");
			    exit(1);
			}
		}

		return (*(--curPtr));
	}

		// Returns number of elements in the stack.
	int getNumOfElements(){ return (prevSize + (curPtr - pageList->page)); }

		// This function returns the value of an element which lies Num
			// elements beneath the top of the stack
			//(function call with Num==0 returns the top).
	type getElement(int num)
	{
		StackPageDescriptor *target;

			// Checking whether the element can be
			//  found in the current page
		if(curPtr-pageList->page>num) return(*(curPtr-1-num));

			// Checking the previous pages for the element:
		num-=curPtr-pageList->page;
		for(target=pageList->next; target && target->size <= num;
				target=target->next)
			num-=target->size;

			// if the requested number is not too big
			// (i.e. the element for the number is not physically
			//  present in the stack),
			//  print the critical error message:
		if(!target)
		{
		      fprintf(stderr,
		              "CRITICAL ERROR in the method StackL::getElement"
		              "\nTrying to access non-existing element at the"
		              " distance %d from the top.\n"
		              "Since there are only %d elements in the stack,\n"
		              "the maximum parameter to the method can be %d"
		              "\n\n", num, getNumOfElements(),
		              getNumOfElements()-1);
		      exit(1);
		}
			// Return the element
		return(target->page[target->size-1-num]);
	}

	/*
	void clear()
		Clears the stack and releases all memory except the first page.
	*/
	void clear()
	{
		StackPageDescriptor *cur;

			// If there is an extra page, start releasing memory
			//  with this extra page.
		if(extraPage) pageList=extraPage;
		extraPage=0;

			// Release the memory allocated for pages.
		while(pageList&&pageList->next)
		{
			    // Following line makes the size of the next page
			    //  to be the size of the last deallocated page.
			nextSize=pageList->size;

				// Deallocate the page.
			cur=pageList;
			pageList=cur->next;
			free(cur->page);
		}

			// Set curPtr at the beginning of the first page:
		curPtr=pageList->page;

			// Initialize 'prevSize'
		prevSize=0;
	}

	/*
	StackL(size_t firstPageSize, float growthRatio, size_t maxPageSize)

		Constructor. Defines general allocation policies to be
		used by the object of this class.
		The given values CANNOT be changed later.

	OPTIONAL PARAMETERS:
		firstPageSize - Size of the first page of memory to be
		                 allocated. Should be positive number
		                 (its type is the same as the
		                 parameter in the function 'malloc').

		growthRatio   - Size of each new page is the size of the
		                previous page times growthRatio up to the value
		                specified by maxPageSize. The size of actually
		                allocated page may be slightly larger than that
		                calculated by this formula because of
		                STK_ROUNDOFF. GrowthRatio CANNOT BE LESS THAN 1.

		maxPageSize   - Maximal Size of the page. LESS THAN OR EQUAL
		                TO 0 indicates that pages are unlimited.
		                It can't be less than firstPageSize. Page may
		                be larger than the given number because of
		                the following reasons:
		                  1. STK_ROUNDOFF
		                  2. If the program requested a piece of memory
		                        which size is more than maxPageSize
	*/
	StackL(size_t firstPageSize=STK_FIRST_DEF, float growthRatio
		=STK_MUL_DEF, size_t maxPageSize=STK_MAX_DEF):
	    FIRST_PAGE_SIZE(firstPageSize),GROWTH_RATIO(growthRatio),
			MAX_PAGE_SIZE(maxPageSize)
	{
			// Checking for errors (terminate if they occur)
		if(FIRST_PAGE_SIZE<=0)
		{
			fprintf(stderr, "FATAL ERROR while constructing "
				"the class \'StackL\':\n"
				"   size of the initial piece of memory "
				"(first parameter of the constructor)\n"
				"   cannot be %d. It should be greater than 0."
				"\n\n", FIRST_PAGE_SIZE);
			exit(1);
		}
		if(GROWTH_RATIO<1)
		{
			fprintf(stderr, "FATAL ERROR while constructing "
				"the class \'StackL\':\n"
				"    growth factor (second parameter of the "
				"constructor) is %.2f\n"
				"    It should be greater than or equal to 1."
				"\n\n", GROWTH_RATIO);
			exit(1);
		}
		if(MAX_PAGE_SIZE>0&&MAX_PAGE_SIZE<FIRST_PAGE_SIZE)
		{
			fprintf(stderr, "FATAL ERROR while constructing the "
				"class \'StackL\':\n"
				"   size of the maximal piece of memory "
				"(third parameter of the constructor)\n"
				"   is %d, which is less than the size of "
				"the initial piece: %d.\n\n",
				MAX_PAGE_SIZE, FIRST_PAGE_SIZE);
			exit(1);
		}
			// Initialize the variables
		nextSize=FIRST_PAGE_SIZE;
		prevSize=0;
		extraPage=0;
		pageList=0;
		curPtr=0;

			// Allocate the first page:
		addPage();
	}

	~StackL()
	{
			// Release all memory except the first page
		clear();
			// Release the first page:
		if(pageList) free(pageList->page);
		pageList=0;
	}
};
#endif
