/*==========================================================================;
 *
 *  AUTHOR:     Iakhiaev, Mikhail.
 *
 *  FILE NAME:  struct.h
 *
 *  CONTENTS:   Part of PSGRND. Declaration of the most important
 *               structures and constants
 *
 *  DATE:       March 19, 2004
 ***************************************************************************/

#ifndef STRUCTURES_H
#define STRUCTURES_H

//***************************************************************************
// Constants for debugging purposes:

// DEBUG OPTION: see memory.h, MEMORY_H_DEBUG

#define DEBUG_PRINT_RULE_BEFORE_PROCESSING     0

#define DEBUG_PRINT_RULE_INTERMEDIATE          0

#define DEBUG_PRINT_RULES_AFTER                0

#define DEBUG_PRINT_TOKENS                     0

#define DEBUG_DONTFREE                         0

// *************************************************************************

//  ERROR MESSAGE LIMITS:
//   Max. errors in command line parameters
#define MAX_CMDLINE_ERRORS      5

//  Max. lexical, syntax, and semantic errors:
#define MAX_LEXICAL_ERRORS      5
#define MAX_SYNTAX_ERRORS       5
#define MAX_SEMANTIC_ERRORS     10
//  Maximum errors during grounding (division by zero or
//    wrong arguments of a weight function)
#define MAX_GROUNDING_ERRORS    5

//  Maximum messages (errors or warnings) regarding
//   repetitions in data predicates or weight functions
#define MAX_DATA_REPETITION_MSG 10
//  Maximum messages regarding repetitions in cardinality or weight lists
#define MAX_META_REPETITION_MSG 20


//***************************************************************************
// GENERAL CONSTANTS
	// Maximum allowed length of the range
#define RANGE_UPPER_LIMIT       1000000

#define INTEGER_MIN             ((signed int)0x80000000)
#define INTEGER_MAX             0x7FFFFFFF
#define UINTEGER_MAX            0xFFFFFFFF
#define FLOAT_MAX               3.402823466e+38F
#define FLOAT_MIN              -3.402823466e+38F

// ***************************************************************************
// TYPES OF IDENTIFIERS
#define TYPE_STRING_CONST       1
#define TYPE_VAR                2
#define TYPE_DATA_PRED          3
#define TYPE_PROGRAM_PRED       4
#define TYPE_WEIGHT_FUN         5
#define TYPE_NUM_CONST          6

//*********************************************************
// Useful macros:

//  (1) Determines the size of an array
//       NOTE: the declaration of the array (with the size given)
//       should occur before the use of this macro.
#define ARRAY_SIZE(array) ((signed)(sizeof(array)/sizeof((array)[0])))

//  (2) Macros to work with doubly linked lists. The only requirement that the
//       previous and next pointers in the node be named 'prev' and 'next'.
#define DLL_APPEND_AFTER(node, point) {(node)->next=(point)->next; \
    (node)->prev=(point);  (point)->next->prev=(node); (point)->next=(node); }

#define DLL_APPEND_BEFORE(node, point) {(node)->prev=(point)->prev; \
    (node)->next=(point);  (point)->prev->next=(node); (point)->prev=(node); }

#define DLL_REMOVE(node) {(node)->next->prev=(node)->prev; \
    (node)->prev->next=(node)->next; (node)->prev=0; (node)->next=0; }



//***************************************************************************
//**** Long-term Temporary structures:

// Arguments of a data predicate, argument or value of a weight function:
struct Value
{
	int type;  // Either EXPR_CONST or EXPR_NUM
	int num;   // If (type==EXPR_NUM), this is a parameter
	           // If (type==EXPR_CONST), this is a value from symbol table
	char *fileName;     // Name of the file where the value is declared
	int lineNum;        // Number of line where we encountered the value
	struct Value *next; // Next value in the list
};

//*****************************************************************************
//**** PERMANENT STRUCTURES
/*
NOTATION: Any variable that has the letters "tmp" as the first part of its
	   name is valid only before rebuilding of parameter lists into arrays.
	   Any variable which is in the union with this variable is valid
	   only after rebuilding.

	   Word "temp" also indicates that the variable is valid during certain
	   period of time and becomes invalid at the point of grounding, but
	   this period is not the one in the previous case.
*/
struct WFnode;
struct DPnode    // Data Predicate Node
{
	int symbolno;  // Number of identifier in the symbol table.
	int arity;     // Arity of the data predicate.
	int valNum;    // Number of values
	union
	{
		struct Value *tmpArguments; // temporary list of arguments
		int *arguments;             // permanent array of arguments
	};
	WFnode *next;// next Data predicate or weight function node in the list
};


struct WFnode: public DPnode   // Weight Function Node
{
	union
	{
		int *results;// Array of function values of the weight function
		Value *tmpResults;  // The list of function values
	};
};

struct Quantifier;

struct VarNode        // Program Variable
{
	int symbolno; // Permanent Variable: Number in the symbol table
	              // Temporary Variable: Negative number of the variable
	struct DPnode *type; // Variable data type

	union	             // Temporary values: needed only during one stage:
	{
		Quantifier *tmpQuant; // For building quantifiers.
		int tmpFlag;          // For reporting grounding errors
	};

	VarNode *next;       // The next variable in the list.
	int valIndex; // 1) BEFORE GROUNDING: Attribute of the variable.
	              //     can be any constant VAR_..., except the ones
	              //     mentioned in part 2.

	              // 2) AFTER GROUNDING: Index of the current value in the
	              //     array of possible values (in DPnode) or
	              //    VAR_BAD_TYPE - If type checking failed and there
	              //       is no index for the value
};

struct Expression;
struct TempAtom;

struct PPnode           // Program predicate node
{
	int symbolno;   // Number of the symbol in the symbol table
	int arity;      // Arity of the program predicate
	short hornMeta; // The variable is equal PP_HORN if the predicate
	                //   appears in a head of a horn rule,
	                //  PP_META if the predicate appears inside a
	                //   meta-atom in a horn rule, or
	                //   PP_NOTHING if neither is the case.
	                //  Can also be PP_WASTED if PP_HORN and PP_META
	                //     both are true
	short usage;    //This variable indicates whether the predicate was used
	                // Can be: PP_DECLARED, PP_USED, PP_GROUNDED.
	unsigned int numOfAtoms;// The number of ground atoms that can be
	                        //   obtained from this predicate
	DPnode *tail;	// Pointer to the predicate tail or 0 if no tail.
	DPnode **params;// Types of arguments of the program predicate
	PPnode *next;	// Next program predicate.
};

struct Quantifier     // Quantifier node
{
	int token;    // Type of the quantifier
	union         // Object to be quantified:
	{
		struct  // For variable quantifier:
		{
			VarNode *var;           // quantified variable
			TempAtom *tmpBestQuant; // temporary
			float tmpCost;          // temporary cost
		};
		struct  // For data predicate quantifier:
		{
			DPnode *dp;       // Data predicate
			Expression *arg;  // Argument of the data predicate
			int dpValIndex;   // Index of the current set of
			                  //   arguments of the data predicate
		};
	};

	Expression *conditions; // List of conditions for the quantifier

		// Previous and next quantifier in the doubly linked list
	Quantifier *next;
	Quantifier *prev;
};

struct Atom  // Simple atom, existential atom in the rule, existential
             //  atom inside the cardinality atom, weight assignment atom,
             // or data predicate condition.
{
	union                      // Predicate
	{
		struct DPnode *dp; // for data predicate condition
		struct PPnode *pp; // for program atoms
	};
	struct Expression *arg;    // Pointer to the list of arguments
	struct Quantifier *quant;  // Array of quantifiers for existential atom
	                           //   Otherwise zero
	struct Expression *weight; // Pointer to the expression to calculate
	                           // weight for the atom. For atoms outside the
	                           // weight atom, this field is equal to zero.
	struct Atom *next;    // Next atom in the head/body of the same rule
};

struct Expression   // Expression node. Also used in the argument lists for
                    //  data and program atoms
{
	int token;               // Expression type
	union                    // Expression values
	{
		int num;                 // Number, index, or reserved
		struct VarNode *var;     // Variable
		struct WFnode *wFun;     // Weight function call
		struct Atom *atom;       // Atom in data predicate condition
		struct Expression *exp;  // Sub-expression
	};

	struct Expression *next; // Next expression in a list
};

struct MetaAtom     // Cardinality or weight atom
{
	Expression* min;       // minimal cardinality or weight value
	Expression* max;       // maximal cardinality or weight value
	struct Atom *list;     // metalists inside { } are separated by ;
	struct MetaAtom *next; // next cardinality or weight
	                       //   atom in the head/body of the same rule
};

struct RuleNode    // Rule
{
	struct Quantifier *quant;    // Array of global quantifiers
	struct Atom *head;           // List of program atoms in the head
	struct Atom *body;           // List of program atoms in the body
	struct MetaAtom *cardHead;   // List of cardinality atoms in the head
	struct MetaAtom *cardBody;   // List of cardinality atoms in the body
	struct MetaAtom *weightHead; // List of weight atoms in the head
	struct MetaAtom *weightBody; // List of weight atoms in the body
	int lineNum;                 // Line number in the rule file where
	                             //  the rule was parsed from
	int maxGroundRule;           // Maximal number assigned to a ground
	                             //  rule grounded from the given rule
	                             // Minimal number is either zero for the
	                             //  first rule or
	                             //   (MaximalNumberOfThePreviousRule+1)
	int maxAtomNum;              // Estimated maximum number of atom in the
	                             //  ground rule grounded from this rule
	int horn;                    // 1 for Horn rules, 0 for the other rules
	struct RuleNode *next;       // Next rule in the theory
};


//*************************************************************************
// CONSTANTS THAT DEFINE POSSIBLE VALUES OF CERTAIN FIELDS
//  FOR THE STRUCTURES DECLARED ABOVE

// NOTE: Quantifier types:
//     GLOBAL - Different values of variables result in different rules
//     LOCAL  - Different values of variables result in different
//               atoms in the same rule


// [1] "PPnode::hornMeta". Used to enforce condition regarding horn rules
//		The program predicate appears in a head of a horn rule
#define PP_HORN       1
//		The predicate appears in a meta-atom in a horn rule
#define PP_META       2
//		Neither of the above is the case
#define PP_NOTHING    0
//		Program predicate occurrs both in the head of a horn rule and
//		in a meta-atom in a horn rule => ERROR => the type of this
//		predicate becomes PP_WASTED to suppress the other error messages
#define PP_WASTED     3


// [2] "PPnode::usage" Used to output the ground atoms and to check whether the
//     predicate was declared but never used.
//		Declared but never used.
#define PP_DECLARED   0
//		The predicate occurred in some rule but no ground atoms have
//		   been generated from this predicate:
#define PP_USED       1
//		There are some ground atoms that were generated
//		   from this predicate:
#define PP_GROUNDED   2


// [3] "VarNode::valIndex":  Attributes of variables or type checking results
//     (a) Value for grounding: Type checking failed
#define VAR_BAD_TYPE     (-1)

//     (b) Program-scale variable type: variable declared but never used.
#define VAR_NOTUSED      (-3)

//     (c) Rule-scale values:
//         Type of the variable right after it has been reduced by the parser
#define VAR_FREE         (-4)
//         Temporary variable for which there is an assignment condition.
//         This variable cannot be quantified.
#define VAR_DEPENDENT    (-5)

//         Indicates that the variable is to be globally quantified.
#define VAR_GLOBAL       (-6)
//         Indicates that the variable is to be locally quantified.
#define VAR_LOCAL        (-7)

//     (d) Temporary flags:
#define VAR_FIRST        (-8)
#define VAR_SECOND       (-9)
#define VAR_THIRD        (-10)


// [4] "Quantifier::token":  type of the quantifier
#define QUANT_VAR        1
#define	QUANT_DP         2
#define QUANT_EMPTY      3

// [5] "Expression::token":  arithmetic operations or argument types
//     (a) Types of expression arguments
#define EXPR_NUM            1
	// String constant
#define EXPR_STRING         2
	// Variable:
#define EXPR_VAR            3

//	   (b) Additional types of the expressions when used as arguments
//               to the atoms or quantifiers:
//         Numerical argument that is represented by its index in the
//              array for the corresponding type rather than by actual value
#define EXPR_INDEX          4
//         Argument is an expression
#define EXPR_EXPR           5
//         Argument is an underscore
#define EXPR_UNDERSCORE     6

// This can occur inside data predicate quantifier and indicates that the given variable
//  is quantified by the given quantifier.
#define EXPR_VAR_QUANT      7

//     (c) Comparison operations (from predefined atoms)
#define EXPR_LESS           8
#define EXPR_NOTGREATER     9
#define EXPR_GREATER        10
#define EXPR_NOTLESS        11
#define EXPR_EQUAL          12
#define EXPR_NOTEQUAL       13


//     (d) Binary arithmetic operations
#define EXPR_PLUS           14
#define EXPR_MINUS          15
#define EXPR_MUL            16
#define EXPR_DIV            17

//     (e) Binary functions
#define EXPR_MOD            18
#define EXPR_MIN            19
#define EXPR_MAX            20

//     (f) Unary operations
#define EXPR_ABS            21
#define EXPR_UMINUS         22

//	   (g) Weight function
#define EXPR_WEIGHT         23

//     (h) Positive and negative data predicate condition nodes.
#define EXPR_DPOSITIVE      24
#define EXPR_DNEGATIVE      25
//     (i) Assignment
#define EXPR_ASSIGN         26

// Additional definitions for comparison operator:
//  Need to check whether the operand is a number or string constant
//   if types of two operands are different, need to force string
//   constant be greater than a number.
#define EXPR_CMP_SPECIAL    1
//  This flag tells that the left or right operand is always number
//   no matter what value is returned by 'stringConstants::isNumber'
//   should be set if the operand is produced from the expression
//    EXPR_NUM or EXPR_EXPR
#define EXPR_CMP_LEFTNUM    3
#define EXPR_CMP_RIGHTNUM   5

// Check for either of the three previous flags:
#define EXPR_CMP_IS_SPECIAL(num) ((num)&EXPR_CMP_SPECIAL)


//***************************************************************************
// ADDITIONAL TEMPORARY NODES:

// Each such node is used for every expression that appears in
//  the program. It contains additional information about arithmetic expression
//   such as its type, and the type of the variables used in the expression.
//   It also contains 'last' pointer to facilitate the append operation
//   ExpressionHandler node will be released once the expression is put in its place.
struct ExpressionHandler
{
	int type;                 // Type of expression: ET_
	Expression *first;        // First node in the EXPRESSION list
	Expression *last;         // Last node in the EXPRESSION list
	ExpressionHandler *next;  // Next HANDLER node
};


// Every such node is associated with the list of structure
// Value, that store parameters to data predicates and
//  weight functions.
struct VListHandler
{
	int arity;      // Arity of the list of values
	int size;       // Number of different values that the list is
	                //  meant to represent (NOT THE ACTUAL SIZE)
	                // actualSize=size*arity.
	Value *first;   // First node in the list
	Value *last;    // Last node in the list
};

// These nodes form lists of parameters for program predicate
//  declarations. Each node is associated with one parameter of
//  the program predicate. Once the predicate declaration rule
//  is reduced, all nodes used for the predicate are released.
struct TempList
{
	int size;       // Number of elements in the list starting from the
	                //  given node. I.e. number of 'next' links that can
	                //  be consequently dereferenced.
	DPnode *type;   // Data type
	TempList *next; // Next node in the list
};

// Every atom is put inside this structure during parsing.
//  Later it is converted to appropriate structure
struct TempAtom
{
	int type; // Type of the atom: ATOM_...
	union     // Union to store information for every kind of atom
	{
		struct // Program atom (ATOM_PROG)
		{
			struct PPnode *pp;      // Program predicate
			struct Expression *arg; // Argument list

			    // Data restrictions of existential atom:
			    //   before creating local quantifiers
			struct TempAtom *restrictions;

			    // Before quantifiers are created, this array
			    //   contains a list of quantified variables.
			    // After, the list of quantifiers
			struct Quantifier *quant;

			    // Pointer to the weight expression.
			    // For atoms outside the weight atom,
			    //  this field is equal to zero
			struct Expression *weight;
		} prog;

		struct  // Data atom (ATOM_DATA, ATOM_DATANEG)
		{
			struct DPnode *dp;      // Data predicate node
			struct Expression *arg; // Argument list
		} data;

		struct  // Cardinality or weight atom (ATOM_CARD, ATOM_WEIGHT)
		{
				// Limits:
			Expression *min;
			Expression *max;
				// Cardinality or weight list:
			TempAtom *list;
		} meta;

		struct  // Predefined atom (ATOM_CONDITION)
		{
			  // Pointers to the first and last nodes
			  //  in the expression list
			Expression *first;
			Expression *last;
		} cond;
	};
	TempAtom *next;  // Used when atom is in list
	union
	{
		TempAtom *prev;  // Used when atom is in doubly linked list.
		TempAtom *last;  // Used when atom is in singly linked list for
		                 //  certain atoms. Points to the last atom in
		                 // the current list.
	};
};

// Union that unites all temporary structures to ensure the
//  proper size of the node (which is equal to maximal size
//  of a temporary structure).
//  IT SHOULD CONTAIN ALL TEMPORARY STRUCTURES DECLARED ABOVE.
union Handler
{
	ExpressionHandler eh;
	VListHandler vlh;
	TempList tl;
	TempAtom ta;
};


// **************************************************************************
// CONSTANTS FOR ADDITIONAL TEMPORARY NODES:

// [1] ExpressionHandler::type and flags
//     NOTE: THEY CANNOT BE CHANGED
//    (a) Expression types: types combined using '|'.
// Error Type:         0
#define ET_NUM         1
#define ET_VAR         2
// (ET_NUM|ET_VAR)     3
#define ET_WEIGHT      4
// (ET_WEIGHT|ET_NUM)  5

//         String constant:
#define ET_STRING      8

//         String variable: a variable that can have values of string
//         constants only:
#define ET_STRVAR      16

//         This variable can take both numerical and string values:
#define ET_MIXVAR      32

//         Special error type. Stops combining the expression with the
//          other one and indicates an error at the point when expression
//          would be used
#define ET_ERROR       64

	// Array that contains names of the variable types
#define ET_TYPE_NAME_ARRAY \
{/*0-2*/"ERROR", "NUMBER", "EXPRESSION WITH VARIABLES", \
 /*3-4*/ "EXPRESSION WITH VARIABLES", "EXPRESSION WITH WEIGHT FUNCTIONS", \
 /*5-5*/ "EXPRESSION WITH WEIGHT FUNCTIONS",\
 /*6-16*/"","","STRING", "", "", "", "", "", "", "", "STRING VARIABLE", \
 /*17-31*/ "","","","", "", "", "", "", "", "", "", "", "", "", "",\
 /*32-33*/ "MIXED VARIABLE","ERROR EXPRESSION"};

#define EXP_CAN_BE_STRING(type) ((type)&(ET_STRING|ET_STRVAR|ET_MIXVAR))
#define EXP_CAN_BE_NUMBER(type) ((type)&(ET_NUM|ET_VAR|ET_MIXVAR))

//	  (a) Macro that says whether expression is ET_NUM or ET_VAR
#define IS_EXP_VARNUM(type) ((type)<=3)

//	  (d) Macro that tells whether an expression can be at the right side of
//	       a weight assignment.
#define IS_EXP_WEIGHT(type) (!((type)&(~(ET_WEIGHT|ET_NUM))))


//    (e) If IS_EXP_NOT_COMBINE(typeOfLeftExp | typeOfRightExp) is true,
//          the expressions cannot be combined. This includes the case
//           when at least one of the expressions has type ET_ERROR
#define IS_EXP_NOT_COMBINE(type) ((type)>5)

//	  (f) Additional macro. Tells whether at least one of the expressions
//	       has the type ET_ERROR.
#define IS_EXP_SAY_ERROR(type)	(!((type)&ET_ERROR))


// [2] TempAtom::type
#define ATOM_PROG            1
#define ATOM_DATA            2
#define ATOM_CARD            3
#define ATOM_WEIGHT          4
#define ATOM_CONDITION       5
//     Data atom to be transferred to negative condition.
#define ATOM_DATANEG         6

//      Preliminary defined atoms (the parser will give a warning
//      every time these atoms are created):
#define ATOM_TRUE            7
#define ATOM_FALSE           8

//*****************************************************************************
// SEMANTIC ERROR AND WARNING CODES:
//*****************************************************************************

//......................................................................
//  SEMANTIC ERRORS: The theory will be grounded only if no errors
//                    have been detected during parsing
//.......................................................................

#define SE_NO_ERROR                   0
/* Zero is reserved value */

#define SE_DIV_ZERO                   1
/* Division by zero during computing of numerical expression;
	semanticError(SE_DIV_ZERO, Operation);, where
	Operation is EXPR_MOD or EXPR_DIV.
*/

#define SE_EXP_TYPES                  2
/* Error during expression merging. Types of expressions to be
   merged are uncompatible.
	 semanticError(SE_EXP_TYPES, int leftType, int rightType);
*/

#define SE_EXP_UNKNOWN_ID             3
/* This error is generated when an expression is reduced to ID, which should
not be the case. Most probable cause is that some constant from the command
line was not entered

semanticError(SE_EXP_UNKNOWN_ID, identifierName);
*/

#define SE_RANGE_TYPES                4
/* Arithmetic expression(s) representing the minimal and/or maximal values of
    some range don't evaluate to numbers.
         semanticError(SE_RANGE_TYPES, int lowerBoundType, int upperBoundType);
*/

#define SE_RANGE_WRONG                5
/* In range declaration:  number that represents the lower bound is greater
     than the number that represents the upper bound.
		semanticError(SE_RANGE_WRONG, int LowerBound, int UpperBound);
*/

#define SE_RANGE_BIG                  6
/* The given range is too big (greater than the reasonable limit:
        constant RANGE_UPPER_LIMIT)
	semanticError(SE_RANGE_BIG, int LowerBound, int UpperBound);
*/

#define SE_MANY_ARITIES               7
/* Found wrong number of arguments of a data predicate or a weight function
	(different from the previous declaration): "multiple arities in data"
  semanticError(SE_MANY_ARITIES, int SymbolNumber,
				int NewlyEncounteredArity);
*/

#define SE_WEIGHT_EXP_TYPE            8
/* The arithmetic expression which is used in the declaration of weight
    function and serves as a value of the weight function does not have
	type ET_NUM
 semanticError(SE_WEIGHT_EXP_TYPE,int WeightFunctionSymbolNumber,
			ExpressionHandler *TroublesomeExpression);
*/

#define SE_PROBLEM_CONST              9
/*  Problem with assigning numbers to string constants. Program has
	  too many constants and/or uses too small negative numbers
	semanticError(SE_PROBLEM_CONST, int MinimalNumberInTheProgram,
									int NumberOfConstantsInTheProgram);
*/

#define SE_TYPE_NOT_UNARY             10
/*Non-unary data predicate is used as a data type
Program variable declaration:
	semanticError(SE_TYPE_NOT_UNARY, int CausePredSymbNo, VarNode *vars)
ProgramPredicateDeclaration:
    semanticError(SE_TYPE_NOT_UNARY, int CausePredSymbNo, 0)
*/

#define SE_EXPR_DECLARATION           11
/* Any expression in the 'dataSequence' or 'declarationSequence' can contain
	only number as its operands. Otherwise this error will be generated.

semanticError(SE_EXPR_DECLARATION, number, type) - parameters are the same as
in expNew()
*/

#define SE_WEIGHT_IN_DECLARATION      12
/* Expression containing weight function call appears in declaration sequence

 semanticError(SE_WEIGHT_IN_DECLARATION, symbolno);
*/

#define SE_DATA_REPETITION            13
/* Found several definitions of a data predicate or weight functions with the
   same arguments (no checking whether the weight value is the same or different).
   NOTE: the message should be processed separately.

	semanticError(SE_WARN_SAME_DATA, DPnode *Node, int IndexOfRepeatedArgument,
			Value *BadList);
	  where BadList is a list containg information about the
	  line numbers where these declarations can be found (node per declaration).
	  Field LineNum contains the last line the declaration appears
 */

#define SE_PP_BAD_TAIL                14
/* The arity of the tail of the program predicate is not equal to the
    arity of the predicate itself.
semanticError(SE_PP_BAD_TAIL, PPnode *ProgramPredicate);*/

#define SE_MANY_PPVALS                15
/* The number of combinations of parameters for some program predicate is
     larger than the number of all possible integers. Therefore,
	 the grounder doesn't support such program atom (cannot store it in
	 program predicate node).
semanticError(SE_MANY_PPVALS, PPnode *ProgramPredicate)
*/

#define SE_ARGUMENT_WEIGHT            16
/* Argument of some predicate contains a weight function

  semanticError(SE_ARGUMENT_WEIGHT, symbolNumberOfThePredicate, argumentWithWeight);
*/

#define SE_PARAM_NUM                  17
/* Wrong number of arguments for program predicate, data predicate or
weight function

semanticError(SE_PARAM_NUM, symbolno, argList);
*/

#define SE_PP_VAR_TYPE                18
/*
The type of the variable or expression in a parameter of some data predicate
is not the same as the type of the corresponding parameter

semanticError(SE_PP_VAR_TYPE, programPredicateNode, argumentList, indexOfParameter);
*/

#define SE_BAD_HORN_RULE              19
/* Wrong head of the horn rule: most probably FALSE atom in the head. Either
 error occurred during parsing the atom or a number given as a parameter to the
 program predicate in the head was out of range

semanticError(SE_BAD_HORN_RULE, antecedent, consequent);
*/

#define SE_META_BAD_LIMIT             20
/* Wrong expression is used as a limit to a cardinality or weight atom:

semanticError(SE_META_BAD_LIMIT, typeOfAtom, isThisMaximum , expressionHandler);
typeOfAtom: either ATOM_CARD or ATOM_WEIGHT
isThisMaximum : TRUE if we are dealing with maximum limit (FALSE otherwise).
expressionHadler: Expression in that limit.
*/

#define SE_WEIGHTEXP_BAD              21
/* An expression in a weight assignment is not of an appropriate type:

semanticError(SE_WEIGHTEXP_BAD, simpleAtom, expressionHandler);
*/

#define SE_PATOM_WRONGTYPES           22
/* Predefined atom contains weight functions in either
	its left or right operand.

semanticError(SE_PATOM_WRONGTYPES, HLeft, HRight);
*/


#define SE_HORN_ATOM_CONFLICT         23
/* A horn predicate appears inside a cardinality or weight atom
NOTE: the message should be processed separately.

semanticError(SE_HORN_ATOM_CONFLICT, programPredicate);
*/


#define SE_GLOBAL_LOCAL_MIX           24
/*
Variable that has to be global according to one atom, appears to be
local for another atom.


semanticError(SE_GLOBAL_LOCAL_MIX, temporaryAtom, variable);
variable appears in local variable list of the 'temporaryAtom'.
*/



// ---------------------------------------------------------------------------
//  SEMANTIC WARNINGS:

#define SW_NUM_OUT_RANGE              33
/* The number that occurred in the argument list is out of range because it
     is less than the minimal number that occurred in the declaration sequence.

semanticError(SW_NUM_OUT_RANGE, temp->first->num);
*/


#define SW_PP_OUT_RANGE               34
/*
Constant argument of the program predicate is not contained in the range of
the type of this argument.

semanticError(SW_PP_OUT_RANGE, programPredicate, argument, indexOfParameter);
*/

#define SW_META_NO_LIMITS             35
/*
A cardinality or weight atom is given without cardinalities.

semanticError(SW_META_NO_LIMITS,atomType,cList);
*/

#define SW_META_MAX_LESS_MIN          36
/*
A cardinality or weight atom is always false because minimum is less than
maximum.

semanticError(SW_META_MAX_LESS_MIN, atomType, minNum, maxNum);
*/

#define SW_META_TRUE                  37
/* Meta atom is always true.

semanticError(SW_META_TRUE, atomType);
*/

#define SW_META_NEGATIVE              38
/* Negative cardinality limit. In case of minimum, it doesn't make sense
to have a negative limit. In case of maximum, negative limit will
make the atom always to be false.

semanticError(SE_META_NOSENSE, atomType, isMaximum, minNum);
*/

#define SW_LOCALVAR_UNUSED            39
/*
A variable in the local variable list doesn't occur in the argument list of
the program predicate; therefore, it will be ignored

semanticError(SW_LOCALVAR_UNUSED, atom, var);
*/

#define SW_LOCALVAR_REPEAT            40
/*
Repetition in the local variable list. Repeated variables are ignored.

semanticError(SW_LOCALVAR_REPEAT, atom, var);
*/

#define SW_USELESS_DATA_ATOM          41
/* Found data atom: dataType(variableOfThatDataType) which doesn't make any
sense because it is always true

semanticError(SW_USELESS_DATA_ATOM, DataPredicate, Variable)
*/


// The following codes are used to report some event:
#define SM_STOP_BEFORE_RULES          48
/* The parser is stopped before parsing rule sequence because there are
errors in declaration sequence */
#define SM_INTERRUPTED                49
/* Notify the user that the parser was interrupted because of
errors */
#define SM_SUPPRESS_REPEAT_MSG        50
/* Too many messages regarding repetitions in data or weight.
All subsequent messages will be suppressed. */


	// Check whether the code is an error
#define SE_IS_ERROR(code) ((code) < 32)
	// Check whether the code is a warning
#define SE_IS_WARNING(code) ((code)>=32 && (code) <48)


//**************** END: SEMANTIC ERROR AND WARNING CODES **********************************

#endif
